@@ -1032,7 +1032,7 @@ F2x_halfgcd_i(GEN a, GEN b)
       gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
     }
   }
-  return gerepilecopy(av, mkmat2(mkcol2(u,u1), mkcol2(v,v1)));
+  return gerepilecopy(av, mkmat22(u,v,u1,v1));
 }
 
 GEN
@@ -2199,6 +2199,35 @@ F2xqX_sqr(GEN x, GEN T)
   return FlxX_renormalize(z, l);
 }
 
+static GEN
+_F2xqX_mul(void *data,GEN a,GEN b) { return F2xqX_mul(a,b, (GEN) data); }
+static GEN
+_F2xqX_sqr(void *data,GEN a) { return F2xqX_sqr(a, (GEN) data); }
+GEN
+F2xqX_powu(GEN x, ulong n, GEN T)
+{ return gen_powu(x, n, (void*)T, &_F2xqX_sqr, &_F2xqX_mul); }
+
+GEN
+F2xqXV_prod(GEN V, GEN T)
+{
+  return gen_product(V, (void*)T, &_F2xqX_mul);
+}
+
+static GEN
+F2xqV_roots_to_deg1(GEN x, GEN T, long v)
+{
+  long sv = get_Flx_var(T);
+  pari_APPLY_same(deg1pol_shallow(pol1_F2x(sv),gel(x,i), v))
+}
+
+GEN
+F2xqV_roots_to_pol(GEN V, GEN T, long v)
+{
+  pari_sp ltop = avma;
+  GEN W = F2xqV_roots_to_deg1(V, T, v);
+  return gerepileupto(ltop, F2xqXV_prod(W, T));
+}
+
 static GEN
 F2xqX_divrem_basecase(GEN x, GEN y, GEN T, GEN *pr)
 {
@@ -2266,7 +2295,7 @@ F2xqX_divrem_basecase(GEN x, GEN y, GEN T, GEN *pr)
     return gc_const((pari_sp)rem, z-2);
   }
   lr=i+3; rem -= lr;
-  rem[0] = evaltyp(t_POL) | evallg(lr);
+  rem[0] = evaltyp(t_POL) | _evallg(lr);
   rem[1] = z[-1];
   p1 = gerepile((pari_sp)rem,tetpil,p1);
   rem += 2; gel(rem,i) = p1;
@@ -2534,35 +2563,15 @@ F2xqX_rem(GEN x, GEN S, GEN T)
   }
 }
 
-static GEN
-F2xqX_halfgcd_basecase(GEN a, GEN b, GEN T)
-{
-  pari_sp av=avma;
-  GEN u,u1,v,v1;
-  long vx = varn(a);
-  long n = lgpol(a)>>1;
-  u1 = v = pol_0(vx);
-  u = v1 = pol1_F2xX(vx, get_F2x_var(T));
-  while (lgpol(b)>n)
-  {
-    GEN r, q = F2xqX_divrem(a,b, T, &r);
-    a = b; b = r; swap(u,u1); swap(v,v1);
-    u1 = F2xX_add(u1, F2xqX_mul(u, q, T));
-    v1 = F2xX_add(v1, F2xqX_mul(v, q ,T));
-    if (gc_needed(av,2))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"F2xqX_halfgcd (d = %ld)",degpol(b));
-      gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
-    }
-  }
-  return gerepilecopy(av, mkmat2(mkcol2(u,u1), mkcol2(v,v1)));
-}
 static GEN
 F2xqX_addmulmul(GEN u, GEN v, GEN x, GEN y, GEN T)
 {
   return F2xX_add(F2xqX_mul(u, x, T),F2xqX_mul(v, y, T));
 }
 
+INLINE GEN
+F2xXn_red(GEN a, long n) { return FlxXn_red(a, n); }
+
 static GEN
 F2xqXM_F2xqX_mul2(GEN M, GEN x, GEN y, GEN T)
 {
@@ -2586,75 +2595,202 @@ F2xqXM_mul2(GEN A, GEN B, GEN T)
   GEN M7 = F2xqX_mul(F2xX_add(A12,A22), F2xX_add(B21,B22), T);
   GEN T1 = F2xX_add(M1,M4), T2 = F2xX_add(M7,M5);
   GEN T3 = F2xX_add(M1,M2), T4 = F2xX_add(M3,M6);
-  retmkmat2(mkcol2(F2xX_add(T1,T2), F2xX_add(M2,M4)),
-            mkcol2(F2xX_add(M3,M5), F2xX_add(T3,T4)));
+  retmkmat22(F2xX_add(T1,T2), F2xX_add(M3,M5),
+             F2xX_add(M2,M4), F2xX_add(T3,T4));
 }
 
 /* Return [0,1;1,-q]*M */
 static GEN
 F2xqX_F2xqXM_qmul(GEN q, GEN M, GEN T)
 {
-  GEN u, v, res = cgetg(3, t_MAT);
-  u = F2xX_add(gcoeff(M,1,1), F2xqX_mul(gcoeff(M,2,1), q, T));
-  gel(res,1) = mkcol2(gcoeff(M,2,1), u);
-  v = F2xX_add(gcoeff(M,1,2), F2xqX_mul(gcoeff(M,2,2), q, T));
-  gel(res,2) = mkcol2(gcoeff(M,2,2), v);
-  return res;
+  GEN u = F2xqX_mul(gcoeff(M,2,1), q, T);
+  GEN v = F2xqX_mul(gcoeff(M,2,2), q, T);
+  retmkmat22(gcoeff(M,2,1), gcoeff(M,2,2),
+    F2xX_add(gcoeff(M,1,1), u), F2xX_add(gcoeff(M,1,2), v));
 }
 
 static GEN
 matid2_F2xXM(long v, long sv)
+{ retmkmat22(pol1_F2xX(v, sv),pol_0(v),pol_0(v),pol1_F2xX(v, sv)); }
+
+static GEN
+matJ2_F2xXM(long v, long sv)
+{ retmkmat22(pol_0(v),pol1_F2xX(v, sv),pol1_F2xX(v, sv),pol_0(v)); }
+
+struct F2xqX_res
 {
-  retmkmat2(mkcol2(pol1_F2xX(v, sv),pol_0(v)),
-            mkcol2(pol_0(v),pol1_F2xX(v, sv)));
+   GEN res, lc;
+   long deg0, deg1;
+};
+
+INLINE void
+F2xqX_halfres_update(long da, long db, long dr, GEN T, struct F2xqX_res *res)
+{
+  if (dr >= 0)
+  {
+    if (!F2x_equal1(res->lc))
+    {
+      res->lc  = F2xq_powu(res->lc, da - dr, T);
+      res->res = F2xq_mul(res->res, res->lc, T);
+    }
+  } else
+  {
+    if (db == 0)
+    {
+      if (!F2x_equal1(res->lc))
+      {
+          res->lc  = F2xq_powu(res->lc, da, T);
+          res->res = F2xq_mul(res->res, res->lc, T);
+      }
+    } else
+      res->res = pol0_F2x(get_F2x_var(T));
+  }
 }
 
 static GEN
-F2xqX_halfgcd_split(GEN x, GEN y, GEN T)
+F2xqX_halfres_basecase(GEN a, GEN b, GEN T, GEN *pa, GEN *pb, struct F2xqX_res *res)
 {
   pari_sp av=avma;
-  GEN R, S, V;
-  GEN y1, r, q;
-  long l = lgpol(x), n = l>>1, k;
-  if (lgpol(y)<=n) return matid2_F2xXM(varn(x),get_F2x_var(T));
-  R = F2xqX_halfgcd(RgX_shift_shallow(x,-n),RgX_shift_shallow(y,-n), T);
-  V = F2xqXM_F2xqX_mul2(R,x,y, T); y1 = gel(V,2);
-  if (lgpol(y1)<=n) return gerepilecopy(av, R);
-  q = F2xqX_divrem(gel(V,1), y1, T, &r);
-  k = 2*n-degpol(y1);
-  S = F2xqX_halfgcd(RgX_shift_shallow(y1,-k), RgX_shift_shallow(r,-k), T);
-  return gerepileupto(av, F2xqXM_mul2(S,F2xqX_F2xqXM_qmul(q,R, T), T));
+  GEN u,u1,v,v1, M;
+  long vx = varn(a), vT = get_F2x_var(T), n = lgpol(a)>>1;
+  u1 = v = pol_0(vx);
+  u = v1 = pol1_F2xX(vx, vT);
+  while (lgpol(b)>n)
+  {
+    GEN r, q;
+    q = F2xqX_divrem(a,b, T, &r);
+    if (res)
+    {
+      long da = degpol(a), db = degpol(b), dr = degpol(r);
+      res->lc = gel(b,db+2);
+      if (dr >= n)
+        F2xqX_halfres_update(da, db, dr, T, res);
+      else
+      {
+        res->deg0 = da;
+        res->deg1 = db;
+      }
+    }
+    a = b; b = r; swap(u,u1); swap(v,v1);
+    u1 = F2xX_add(u1, F2xqX_mul(u, q, T));
+    v1 = F2xX_add(v1, F2xqX_mul(v, q, T));
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"F2xqX_halfgcd (d = %ld)",degpol(b));
+      gerepileall(av,res ? 8: 6, &a,&b,&u1,&v1,&u,&v,&res->res,&res->lc);
+    }
+  }
+  M = mkmat22(u,v,u1,v1); *pa = a; *pb = b;
+  return gc_all(av, res ? 5: 3, &M, pa, pb, &res->res, &res->lc);
 }
 
-/* Return M in GL_2(Fp[X]) such that:
+static GEN F2xqX_halfres_i(GEN x, GEN y, GEN T, GEN *a, GEN *b, struct F2xqX_res *res);
+
+static GEN
+F2xqX_halfres_split(GEN x, GEN y, GEN T, GEN *a, GEN *b, struct F2xqX_res *res)
+{
+  pari_sp av = avma;
+  GEN Q, R, S, V1, V2;
+  GEN x1, y1, r, q;
+  long l = lgpol(x), n = l>>1, k, vT = get_F2x_var(T);
+  if (lgpol(y) <= n)
+    { *a = RgX_copy(x); *b = RgX_copy(y); return matid2_F2xXM(varn(x),vT); }
+  if (res)
+  {
+     res->lc = leading_coeff(y);
+     res->deg0 -= n;
+     res->deg1 -= n;
+  }
+  R = F2xqX_halfres_i(F2xX_shift(x,-n, vT),F2xX_shift(y,-n, vT), T, a, b, res);
+  if (res)
+  {
+    res->deg0 += n;
+    res->deg1 += n;
+  }
+  V1 = F2xqXM_F2xqX_mul2(R, Flxn_red(x,n), Flxn_red(y,n), T);
+  x1 = F2xX_add(F2xX_shift(*a,n,vT), gel(V1,1));
+  y1 = F2xX_add(F2xX_shift(*b,n,vT), gel(V1,2));
+  if (lgpol(y1) <= n)
+  {
+    *a = x1; *b = y1;
+    return gc_all(av, res ? 5: 3, &R, a, b, &res->res, &res->lc);
+  }
+  k = 2*n-degpol(y1);
+  q = F2xqX_divrem(x1, y1, T, &r);
+  if (res)
+  {
+    long dx1 = degpol(x1), dy1 = degpol(y1), dr = degpol(r);
+    if (dy1 < degpol(y))
+      F2xqX_halfres_update(res->deg0, res->deg1, dy1, T, res);
+    res->lc = leading_coeff(y1);
+    res->deg0 = dx1;
+    res->deg1 = dy1;
+    if (dr >= n)
+    {
+      F2xqX_halfres_update(dx1, dy1, dr, T, res);
+      res->deg0 = dy1;
+      res->deg1 = dr;
+    }
+    res->deg0 -= k;
+    res->deg1 -= k;
+  }
+  S = F2xqX_halfres_i(F2xX_shift(y1,-k, vT), F2xX_shift(r,-k, vT), T, a, b, res);
+  if (res)
+  {
+    res->deg0 += k;
+    res->deg1 += k;
+  }
+  Q = F2xqXM_mul2(S,F2xqX_F2xqXM_qmul(q, R, T), T);
+  V2 = F2xqXM_F2xqX_mul2(S, F2xXn_red(y1,k), F2xXn_red(r,k), T);
+  *a = F2xX_add(F2xX_shift(*a,k,vT), gel(V2,1));
+  *b = F2xX_add(F2xX_shift(*b,k,vT), gel(V2,2));
+  return gc_all(av, res ? 5: 3, &Q, a, b, &res->res, &res->lc);
+}
+
+static GEN
+F2xqX_halfres_i(GEN x, GEN y, GEN T, GEN *a, GEN *b, struct F2xqX_res *res)
+{
+  if (lgpol(x) < F2xqX_HALFGCD_LIMIT)
+    return F2xqX_halfres_basecase(x, y, T, a, b, res);
+  return F2xqX_halfres_split(x, y, T, a, b, res);
+}
+
+static GEN
+F2xqX_halfgcd_all_i(GEN x, GEN y, GEN T, GEN *pa, GEN *pb)
+{
+  GEN a, b;
+  GEN R = F2xqX_halfres_i(x, y, T, &a, &b, NULL);
+  if (pa) *pa = a;
+  if (pb) *pb = b;
+  return R;
+}
+
+/* Return M in GL_2(F_2[X]/(T)[Y]) such that:
 if [a',b']~=M*[a,b]~ then degpol(a')>= (lgpol(a)>>1) >degpol(b')
 */
 
-static GEN
-F2xqX_halfgcd_i(GEN x, GEN y, GEN T)
+GEN
+F2xqX_halfgcd_all(GEN x, GEN y, GEN T, GEN *a, GEN *b)
 {
-  if (lg(x)<=F2xqX_HALFGCD_LIMIT) return F2xqX_halfgcd_basecase(x, y, T);
-  return F2xqX_halfgcd_split(x, y, T);
+  pari_sp av = avma;
+  GEN R,q,r;
+  if (!signe(x))
+  {
+    if (a) *a = RgX_copy(y);
+    if (b) *b = RgX_copy(x);
+    return matJ2_F2xXM(varn(x), get_F2x_var(T));
+  }
+  if (degpol(y)<degpol(x)) return F2xqX_halfgcd_all_i(x, y, T, a, b);
+  q = F2xqX_divrem(y, x, T, &r);
+  R = F2xqX_halfgcd_all_i(x, r, T, a, b);
+  gcoeff(R,1,1) = F2xX_add(gcoeff(R,1,1), F2xqX_mul(q, gcoeff(R,1,2), T));
+  gcoeff(R,2,1) = F2xX_add(gcoeff(R,2,1), F2xqX_mul(q, gcoeff(R,2,2), T));
+  return !a && b ? gc_all(av, 2, &R, b): gc_all(av, 1+!!a+!!b, &R, a, b);
 }
 
 GEN
 F2xqX_halfgcd(GEN x, GEN y, GEN T)
-{
-  pari_sp av = avma;
-  GEN M,q,r;
-  if (!signe(x))
-  {
-    long v = varn(x), vT = get_F2x_var(T);
-    retmkmat2(mkcol2(pol_0(v),pol1_F2xX(v,vT)),
-        mkcol2(pol1_F2xX(v,vT),pol_0(v)));
-  }
-  if (degpol(y)<degpol(x)) return F2xqX_halfgcd_i(x, y, T);
-  q = F2xqX_divrem(y, x, T, &r);
-  M = F2xqX_halfgcd_i(x, r, T);
-  gcoeff(M,1,1) = F2xX_add(gcoeff(M,1,1), F2xqX_mul(q, gcoeff(M,1,2), T));
-  gcoeff(M,2,1) = F2xX_add(gcoeff(M,2,1), F2xqX_mul(q, gcoeff(M,2,2), T));
-  return gerepilecopy(av, M);
-}
+{ return F2xqX_halfgcd_all(x, y, T, NULL, NULL); }
 
 static GEN
 F2xqX_gcd_basecase(GEN a, GEN b, GEN T)
@@ -2680,17 +2816,19 @@ F2xqX_gcd(GEN x, GEN y, GEN T)
   x = F2xqX_red(x, T);
   y = F2xqX_red(y, T);
   if (!signe(x)) return gerepileupto(av, y);
-  while (lg(y)>F2xqX_GCD_LIMIT)
+  while (lgpol(y)>=F2xqX_GCD_LIMIT)
   {
-    GEN c;
     if (lgpol(y)<=(lgpol(x)>>1))
     {
       GEN r = F2xqX_rem(x, y, T);
       x = y; y = r;
     }
-    c = F2xqXM_F2xqX_mul2(F2xqX_halfgcd(x,y, T), x, y, T);
-    x = gel(c,1); y = gel(c,2);
-    gerepileall(av,2,&x,&y);
+    (void) F2xqX_halfgcd_all(x, y, T, &x, &y);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"F2xqX_gcd (y = %ld)",degpol(y));
+      gerepileall(av,2,&x,&y);
+    }
   }
   return gerepileupto(av, F2xqX_gcd_basecase(x, y, T));
 }
@@ -2722,26 +2860,33 @@ F2xqX_extgcd_basecase(GEN a, GEN b, GEN T,  GEN *ptu, GEN *ptv)
 static GEN
 F2xqX_extgcd_halfgcd(GEN x, GEN y, GEN T,  GEN *ptu, GEN *ptv)
 {
-  pari_sp av=avma;
-  GEN u,v,R = matid2_F2xXM(varn(x), get_F2x_var(T));
-  while (lg(y)>F2xqX_EXTGCD_LIMIT)
+  GEN u,v;
+  GEN V = cgetg(expu(lgpol(y))+2,t_VEC);
+  long i, n = 0, vs = varn(x), vT = get_F2x_var(T);
+  while (lgpol(y)>=F2xqX_EXTGCD_LIMIT)
   {
-    GEN M, c;
     if (lgpol(y)<=(lgpol(x)>>1))
     {
       GEN r, q = F2xqX_divrem(x, y, T, &r);
       x = y; y = r;
-      R = F2xqX_F2xqXM_qmul(q, R, T);
-    }
-    M = F2xqX_halfgcd(x,y, T);
-    c = F2xqXM_F2xqX_mul2(M, x,y, T);
-    R = F2xqXM_mul2(M, R, T);
-    x = gel(c,1); y = gel(c,2);
-    gerepileall(av,3,&x,&y,&R);
+      gel(V,++n) = mkmat22(pol_0(vs),pol1_F2xX(vs,vT),pol1_F2xX(vs,vT),RgX_copy(q));
+    } else
+      gel(V,++n) = F2xqX_halfgcd_all(x, y, T, &x, &y);
   }
   y = F2xqX_extgcd_basecase(x,y, T, &u,&v);
-  if (ptu) *ptu = F2xqX_addmulmul(u,v,gcoeff(R,1,1),gcoeff(R,2,1), T);
-  *ptv = F2xqX_addmulmul(u,v,gcoeff(R,1,2),gcoeff(R,2,2), T);
+  for (i = n; i>1; i--)
+  {
+    GEN R = gel(V,i);
+    GEN u1 = F2xqX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), T);
+    GEN v1 = F2xqX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), T);
+    u = u1; v = v1;
+  }
+  {
+    GEN R = gel(V,1);
+    if (ptu)
+      *ptu = F2xqX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), T);
+    *ptv   = F2xqX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), T);
+  }
   return y;
 }
 
@@ -2754,7 +2899,7 @@ F2xqX_extgcd(GEN x, GEN y, GEN T,  GEN *ptu, GEN *ptv)
   GEN d;
   x = F2xqX_red(x, T);
   y = F2xqX_red(y, T);
-  if (lg(y)>F2xqX_EXTGCD_LIMIT)
+  if (lgpol(y)>=F2xqX_EXTGCD_LIMIT)
     d = F2xqX_extgcd_halfgcd(x, y, T, ptu, ptv);
   else
     d = F2xqX_extgcd_basecase(x, y, T, ptu, ptv);
@@ -2762,20 +2907,30 @@ F2xqX_extgcd(GEN x, GEN y, GEN T,  GEN *ptu, GEN *ptv)
 }
 
 static GEN
-_F2xqX_mul(void *data,GEN a,GEN b) { return F2xqX_mul(a,b, (GEN) data); }
-static GEN
-_F2xqX_sqr(void *data,GEN a) { return F2xqX_sqr(a, (GEN) data); }
-GEN
-F2xqX_powu(GEN x, ulong n, GEN T)
-{ return gen_powu(x, n, (void*)T, &_F2xqX_sqr, &_F2xqX_mul); }
+F2xqX_halfres(GEN x, GEN y, GEN T, GEN *a, GEN *b, GEN *r)
+{
+  struct F2xqX_res res;
+  GEN V;
+  long dB;
+
+  res.res  = *r;
+  res.lc   = leading_coeff(y);
+  res.deg0 = degpol(x);
+  res.deg1 = degpol(y);
+  V = F2xqX_halfres_i(x, y, T, a, b, &res);
+  dB = degpol(*b);
+  if (dB < degpol(y))
+    F2xqX_halfres_update(res.deg0, res.deg1, dB, T, &res);
+  *r = res.res;
+  return V;
+}
 
 /* Res(A,B) = Res(B,R) * lc(B)^(a-r) * (-1)^(ab), with R=A%B, a=deg(A) ...*/
-GEN
-F2xqX_resultant(GEN a, GEN b, GEN T)
+static GEN
+F2xqX_resultant_basecase(GEN a, GEN b, GEN T)
 {
-  long vT = get_F2x_var(T);
-  long da,db,dc;
-  pari_sp av;
+  pari_sp av = avma;
+  long vT = get_F2x_var(T), da,db,dc;
   GEN c,lb, res = pol1_F2x(vT);
 
   if (!signe(a) || !signe(b)) return pol0_F2x(vT);
@@ -2785,7 +2940,6 @@ F2xqX_resultant(GEN a, GEN b, GEN T)
   if (db > da)
     swapspec(a,b, da,db);
   if (!da) return pol1_F2x(vT); /* = res * a[2] ^ db, since 0 <= db <= da = 0 */
-  av = avma;
   while (db)
   {
     lb = gel(b,db+2);
@@ -2806,6 +2960,39 @@ F2xqX_resultant(GEN a, GEN b, GEN T)
   return gerepileupto(av, res);
 }
 
+/* Res(A,B) = Res(B,R) * lc(B)^(a-r) * (-1)^(ab), with R=A%B, a=deg(A) ...*/
+GEN
+F2xqX_resultant(GEN x, GEN y, GEN T)
+{
+  pari_sp av = avma;
+  long dx, dy, vT = get_F2x_var(T);
+  GEN res = pol1_F2x(vT);
+  if (!signe(x) || !signe(y)) return pol0_F2x(vT);
+  dx = degpol(x); dy = degpol(y);
+  if (dx < dy)
+    swap(x,y);
+  while (lgpol(y) >= F2xqX_GCD_LIMIT)
+  {
+    if (lgpol(y)<=(lgpol(x)>>1))
+    {
+      GEN r = F2xqX_rem(x, y, T);
+      long dx = degpol(x), dy = degpol(y), dr = degpol(r);
+      GEN ly = gel(y,dy+2);
+      if (!F2x_equal1(ly))
+        res = F2xq_mul(res, F2xq_powu(ly, dx - dr, T), T);
+      x = y; y = r;
+    }
+    (void) F2xqX_halfres(x, y, T, &x, &y, &res);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"F2xqX_resultant (y = %ld)",degpol(y));
+      gerepileall(av,3,&x,&y,&res);
+    }
+  }
+  res = F2xq_mul(res, F2xqX_resultant_basecase(x, y, T), T);
+  return gerepileupto(av, res);
+}
+
 /* disc P = (-1)^(n(n-1)/2) lc(P)^(n - deg P' - 2) Res(P,P'), n = deg P */
 GEN
 F2xqX_disc(GEN P, GEN T)
