@@ -796,6 +796,28 @@ GEN
 FlxqX_normalize(GEN z, GEN T, ulong p)
 { return FlxqX_normalize_pre(z, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
 
+struct _FlxqX {ulong p, pi; GEN T;};
+static GEN _FlxqX_mul(void *data,GEN a,GEN b)
+{
+  struct _FlxqX *d=(struct _FlxqX*)data;
+  return FlxqX_mul_pre(a,b,d->T,d->p,d->pi);
+}
+static GEN _FlxqX_sqr(void *data,GEN a)
+{
+  struct _FlxqX *d=(struct _FlxqX*)data;
+  return FlxqX_sqr_pre(a,d->T,d->p,d->pi);
+}
+
+GEN
+FlxqX_powu_pre(GEN V, ulong n, GEN T, ulong p, ulong pi)
+{
+  struct _FlxqX d; d.p = p; d.pi = pi; d.T = T;
+  return gen_powu(V, n, (void*)&d, &_FlxqX_sqr, &_FlxqX_mul);
+}
+GEN
+FlxqX_powu(GEN V, ulong n, GEN T, ulong p)
+{ return FlxqX_powu_pre(V, n, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
+
 /* x and y in Z[Y][X]. Assume T irreducible mod p */
 static GEN
 FlxqX_divrem_basecase(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *pr)
@@ -865,7 +887,7 @@ FlxqX_divrem_basecase(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *pr)
     return gc_const((pari_sp)rem, z-2);
   }
   lr=i+3; rem -= lr;
-  rem[0] = evaltyp(t_POL) | evallg(lr);
+  rem[0] = evaltyp(t_POL) | _evallg(lr);
   rem[1] = z[-1];
   p1 = gerepile((pari_sp)rem,tetpil,p1);
   rem += 2; gel(rem,i) = p1;
@@ -1140,29 +1162,39 @@ GEN
 FlxqX_rem(GEN x, GEN S, GEN T, ulong p)
 { return FlxqX_rem_pre(x, S, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
 
-static GEN
-FlxqX_halfgcd_basecase(GEN a, GEN b, GEN T, ulong p, ulong pi)
+/* x + y*z mod p */
+INLINE GEN
+Flxq_addmul_pre(GEN x, GEN y, GEN z, GEN T, ulong p, ulong pi)
 {
-  pari_sp av=avma;
-  GEN u,u1,v,v1;
-  long vx = varn(a);
-  long n = lgpol(a)>>1;
-  u1 = v = pol_0(vx);
-  u = v1 = pol1_FlxX(vx, get_Flx_var(T));
-  while (lgpol(b)>n)
-  {
-    GEN r, q = FlxqX_divrem_pre(a,b, T, p, pi, &r);
-    a = b; b = r; swap(u,u1); swap(v,v1);
-    u1 = FlxX_sub(u1, FlxqX_mul_pre(u, q, T, p, pi), p);
-    v1 = FlxX_sub(v1, FlxqX_mul_pre(v, q ,T, p, pi), p);
-    if (gc_needed(av,2))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"FlxqX_halfgcd (d = %ld)",degpol(b));
-      gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
-    }
-  }
-  return gerepilecopy(av, mkmat2(mkcol2(u,u1), mkcol2(v,v1)));
+  pari_sp av;
+  if (!lgpol(y) || !lgpol(z)) return Flx_rem_pre(x, T, p, pi);
+  if (!lgpol(x)) return Flxq_mul_pre(z, y, T, p, pi);
+  av = avma;
+  return gerepileupto(av, Flx_add(x, Flxq_mul_pre(y, z, T, p, pi), p));
 }
+
+GEN
+FlxqX_div_by_X_x_pre(GEN a, GEN x, GEN T, ulong p, ulong pi, GEN *r)
+{
+  long l = lg(a), i;
+  GEN z;
+  if (l <= 3)
+  {
+    if (r) *r = l == 2? pol0_Flx(get_Flx_var(T)): Flx_copy(gel(a,2));
+    return pol_0(varn(a));
+  }
+  l--; z = cgetg(l, t_POL); z[1] = a[1];
+  gel(z, l-1) = gel(a,l);
+  for (i=l-2; i>1; i--) /* z[i] = a[i+1] + x*z[i+1] */
+    gel(z, i) = Flxq_addmul_pre(gel(a,i+1), x, gel(z,i+1), T, p, pi);
+  if (r) *r = Flxq_addmul_pre(gel(a,2), x, gel(z,2), T, p, pi);
+  return z;
+}
+
+GEN
+FlxqX_div_by_X_x(GEN a, GEN x, GEN T, ulong p, GEN *r)
+{ return FlxqX_div_by_X_x_pre(a, x, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p), r); }
+
 static GEN
 FlxqX_addmulmul(GEN u, GEN v, GEN x, GEN y, GEN T, ulong p, ulong pi)
 {
@@ -1193,78 +1225,214 @@ FlxqXM_mul2(GEN A, GEN B, GEN T, ulong p, ulong pi)
   GEN M7 = FlxqX_mul_pre(FlxX_sub(A12,A22, p), FlxX_add(B21,B22, p), T, p, pi);
   GEN T1 = FlxX_add(M1,M4, p), T2 = FlxX_sub(M7,M5, p);
   GEN T3 = FlxX_sub(M1,M2, p), T4 = FlxX_add(M3,M6, p);
-  retmkmat2(mkcol2(FlxX_add(T1,T2, p), FlxX_add(M2,M4, p)),
-            mkcol2(FlxX_add(M3,M5, p), FlxX_add(T3,T4, p)));
+  retmkmat22(FlxX_add(T1,T2, p), FlxX_add(M3,M5, p),
+             FlxX_add(M2,M4, p), FlxX_add(T3,T4, p));
 }
 
 /* Return [0,1;1,-q]*M */
 static GEN
 FlxqX_FlxqXM_qmul(GEN q, GEN M, GEN T, ulong p, ulong pi)
 {
-  GEN u, v, res = cgetg(3, t_MAT);
-  u = FlxX_sub(gcoeff(M,1,1), FlxqX_mul_pre(gcoeff(M,2,1), q, T,p,pi), p);
-  gel(res,1) = mkcol2(gcoeff(M,2,1), u);
-  v = FlxX_sub(gcoeff(M,1,2), FlxqX_mul_pre(gcoeff(M,2,2), q, T,p,pi), p);
-  gel(res,2) = mkcol2(gcoeff(M,2,2), v);
-  return res;
+  GEN u = FlxqX_mul_pre(gcoeff(M,2,1), q, T,p,pi);
+  GEN v = FlxqX_mul_pre(gcoeff(M,2,2), q, T,p,pi);
+  retmkmat22(gcoeff(M,2,1), gcoeff(M,2,2),
+    FlxX_sub(gcoeff(M,1,1), u, p), FlxX_sub(gcoeff(M,1,2), v, p));
 }
 
 static GEN
 matid2_FlxXM(long v, long sv)
+{ retmkmat22(pol1_FlxX(v, sv),pol_0(v),pol_0(v),pol1_FlxX(v, sv)); }
+
+static GEN
+matJ2_FlxXM(long v, long sv)
+{ retmkmat22(pol_0(v),pol1_FlxX(v, sv),pol1_FlxX(v, sv),pol_0(v)); }
+
+struct FlxqX_res
 {
-  retmkmat2(mkcol2(pol1_FlxX(v, sv),pol_0(v)),
-            mkcol2(pol_0(v),pol1_FlxX(v, sv)));
+   GEN res, lc;
+   long deg0, deg1, off;
+};
+
+INLINE void
+FlxqX_halfres_update(long da, long db, long dr, GEN T, ulong p, ulong pi, struct FlxqX_res *res)
+{
+  if (dr >= 0)
+  {
+    if (!Flx_equal1(res->lc))
+    {
+      res->lc  = Flxq_powu_pre(res->lc, da - dr, T, p, pi);
+      res->res = Flxq_mul_pre(res->res, res->lc, T, p, pi);
+    }
+    if (both_odd(da + res->off, db + res->off))
+      res->res = Flx_neg(res->res, p);
+  } else
+  {
+    if (db == 0)
+    {
+      if (!Flx_equal1(res->lc))
+      {
+          res->lc  = Flxq_powu_pre(res->lc, da, T, p, pi);
+          res->res = Flxq_mul_pre(res->res, res->lc, T, p, pi);
+      }
+    } else
+      res->res = pol0_Flx(get_Flx_var(T));
+  }
 }
 
 static GEN
-FlxqX_halfgcd_split(GEN x, GEN y, GEN T, ulong p, ulong pi)
+FlxqX_halfres_basecase(GEN a, GEN b, GEN T, ulong p, ulong pi, GEN *pa, GEN *pb, struct FlxqX_res *res)
 {
   pari_sp av=avma;
-  GEN R, S, V;
-  GEN y1, r, q;
-  long l = lgpol(x), n = l>>1, k;
-  long vT = get_Flx_var(T);
-  if (lgpol(y)<=n) return matid2_FlxXM(varn(x),vT);
-  R = FlxqX_halfgcd_pre(FlxX_shift(x,-n,vT),FlxX_shift(y,-n,vT), T, p, pi);
-  V = FlxqXM_FlxqX_mul2(R,x,y, T, p, pi); y1 = gel(V,2);
-  if (lgpol(y1)<=n) return gerepilecopy(av, R);
-  q = FlxqX_divrem_pre(gel(V,1), y1, T, p, pi, &r);
-  k = 2*n-degpol(y1);
-  S = FlxqX_halfgcd_pre(FlxX_shift(y1,-k,vT), FlxX_shift(r,-k,vT), T, p, pi);
-  return gerepileupto(av, FlxqXM_mul2(S,FlxqX_FlxqXM_qmul(q,R,T,p,pi), T,p,pi));
+  GEN u,u1,v,v1, M;
+  long vx = varn(a), vT = get_Flx_var(T), n = lgpol(a)>>1;
+  u1 = v = pol_0(vx);
+  u = v1 = pol1_FlxX(vx, vT);
+  while (lgpol(b)>n)
+  {
+    GEN r, q;
+    q = FlxqX_divrem(a,b, T, p, &r);
+    if (res)
+    {
+      long da = degpol(a), db = degpol(b), dr = degpol(r);
+      res->lc = gel(b,db+2);
+      if (dr >= n)
+        FlxqX_halfres_update(da, db, dr, T, p, pi, res);
+      else
+      {
+        res->deg0 = da;
+        res->deg1 = db;
+      }
+    }
+    a = b; b = r; swap(u,u1); swap(v,v1);
+    u1 = FlxX_sub(u1, FlxqX_mul_pre(u, q, T, p, pi), p);
+    v1 = FlxX_sub(v1, FlxqX_mul_pre(v, q, T, p, pi), p);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FlxqX_halfgcd (d = %ld)",degpol(b));
+      gerepileall(av,res ? 8: 6, &a,&b,&u1,&v1,&u,&v,&res->res,&res->lc);
+    }
+  }
+  M = mkmat22(u,v,u1,v1); *pa = a; *pb = b;
+  return gc_all(av, res ? 5: 3, &M, pa, pb, &res->res, &res->lc);
 }
 
-/* Return M in GL_2(Fp[X]) such that:
+static GEN FlxqX_halfres_i(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *a, GEN *b, struct FlxqX_res *res);
+
+static GEN
+FlxqX_halfres_split(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *a, GEN *b, struct FlxqX_res *res)
+{
+  pari_sp av = avma;
+  GEN Q, R, S, V1, V2;
+  GEN x1, y1, r, q;
+  long l = lgpol(x), n = l>>1, k, vT = get_Flx_var(T);
+  if (lgpol(y) <= n)
+    { *a = RgX_copy(x); *b = RgX_copy(y); return matid2_FlxXM(varn(x), vT); }
+  if (res)
+  {
+     res->lc = leading_coeff(y);
+     res->deg0 -= n;
+     res->deg1 -= n;
+     res->off += n;
+  }
+  R = FlxqX_halfres_i(FlxX_shift(x,-n, vT),FlxX_shift(y,-n, vT), T, p, pi, a, b, res);
+  if (res)
+  {
+    res->off -= n;
+    res->deg0 += n;
+    res->deg1 += n;
+  }
+  V1 = FlxqXM_FlxqX_mul2(R, Flxn_red(x,n), Flxn_red(y,n), T, p, pi);
+  x1 = FlxX_add(FlxX_shift(*a,n,vT), gel(V1,1), p);
+  y1 = FlxX_add(FlxX_shift(*b,n,vT), gel(V1,2), p);
+  if (lgpol(y1) <= n)
+  {
+    *a = x1; *b = y1;
+    return gc_all(av, res ? 5: 3, &R, a, b, &res->res, &res->lc);
+  }
+  k = 2*n-degpol(y1);
+  q = FlxqX_divrem(x1, y1, T, p, &r);
+  if (res)
+  {
+    long dx1 = degpol(x1), dy1 = degpol(y1), dr = degpol(r);
+    if (dy1 < degpol(y))
+      FlxqX_halfres_update(res->deg0, res->deg1, dy1, T, p, pi, res);
+    res->lc = leading_coeff(y1);
+    res->deg0 = dx1;
+    res->deg1 = dy1;
+    if (dr >= n)
+    {
+      FlxqX_halfres_update(dx1, dy1, dr, T, p, pi, res);
+      res->deg0 = dy1;
+      res->deg1 = dr;
+    }
+    res->deg0 -= k;
+    res->deg1 -= k;
+    res->off += k;
+  }
+  S = FlxqX_halfres_i(FlxX_shift(y1,-k, vT), FlxX_shift(r,-k, vT), T, p, pi, a, b, res);
+  if (res)
+  {
+    res->deg0 += k;
+    res->deg1 += k;
+    res->off -= k;
+  }
+  Q = FlxqXM_mul2(S, FlxqX_FlxqXM_qmul(q, R, T, p, pi), T, p, pi);
+  V2 = FlxqXM_FlxqX_mul2(S, FlxXn_red(y1,k), FlxXn_red(r,k), T, p, pi);
+  *a = FlxX_add(FlxX_shift(*a,k,vT), gel(V2,1), p);
+  *b = FlxX_add(FlxX_shift(*b,k,vT), gel(V2,2), p);
+  return gc_all(av, res ? 5: 3, &Q, a, b, &res->res, &res->lc);
+}
+
+static GEN
+FlxqX_halfres_i(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *a, GEN *b, struct FlxqX_res *res)
+{
+  if (lgpol(x) < FlxqX_HALFGCD_LIMIT)
+    return FlxqX_halfres_basecase(x, y, T, p, pi, a, b, res);
+  return FlxqX_halfres_split(x, y, T, p, pi, a, b, res);
+}
+
+static GEN
+FlxqX_halfgcd_all_i(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *pa, GEN *pb)
+{
+  GEN a, b;
+  GEN R = FlxqX_halfres_i(x, y, T, p, pi, &a, &b, NULL);
+  if (pa) *pa = a;
+  if (pb) *pb = b;
+  return R;
+}
+
+/* Return M in GL_2(Fp[X]/(T)[Y]) such that:
 if [a',b']~=M*[a,b]~ then degpol(a')>= (lgpol(a)>>1) >degpol(b')
 */
 
-static GEN
-FlxqX_halfgcd_i(GEN x, GEN y, GEN T, ulong p, ulong pi)
+GEN
+FlxqX_halfgcd_all_pre(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *a, GEN *b)
 {
-  if (lg(x)<=FlxqX_HALFGCD_LIMIT) return FlxqX_halfgcd_basecase(x, y, T, p, pi);
-  return FlxqX_halfgcd_split(x, y, T, p, pi);
+  pari_sp av = avma;
+  GEN R,q,r;
+  if (!signe(x))
+  {
+    if (a) *a = RgX_copy(y);
+    if (b) *b = RgX_copy(x);
+    return matJ2_FlxXM(varn(x),get_Flx_var(T));
+  }
+  if (degpol(y)<degpol(x)) return FlxqX_halfgcd_all_i(x, y, T, p, pi, a, b);
+  q = FlxqX_divrem_pre(y, x, T, p, pi, &r);
+  R = FlxqX_halfgcd_all_i(x, r, T, p, pi, a, b);
+  gcoeff(R,1,1) = FlxX_sub(gcoeff(R,1,1),
+                           FlxqX_mul_pre(q, gcoeff(R,1,2), T, p, pi), p);
+  gcoeff(R,2,1) = FlxX_sub(gcoeff(R,2,1),
+                           FlxqX_mul_pre(q, gcoeff(R,2,2), T, p, pi), p);
+  return !a && b ? gc_all(av, 2, &R, b): gc_all(av, 1+!!a+!!b, &R, a, b);
 }
+GEN
+FlxqX_halfgcd_all(GEN x, GEN y, GEN T, ulong p, GEN *a, GEN *b)
+{ return FlxqX_halfgcd_all_pre(x, y, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p), a, b); }
 
 GEN
 FlxqX_halfgcd_pre(GEN x, GEN y, GEN T, ulong p, ulong pi)
-{
-  pari_sp av = avma;
-  GEN M,q,r;
-  if (!signe(x))
-  {
-    long v = varn(x), vT = get_Flx_var(T);
-    retmkmat2(mkcol2(pol_0(v),pol1_FlxX(v,vT)),
-              mkcol2(pol1_FlxX(v,vT),pol_0(v)));
-  }
-  if (degpol(y)<degpol(x)) return FlxqX_halfgcd_i(x, y, T, p, pi);
-  q = FlxqX_divrem_pre(y, x, T, p, pi, &r);
-  M = FlxqX_halfgcd_i(x, r, T, p, pi);
-  gcoeff(M,1,1) = FlxX_sub(gcoeff(M,1,1),
-                           FlxqX_mul_pre(q, gcoeff(M,1,2), T, p, pi), p);
-  gcoeff(M,2,1) = FlxX_sub(gcoeff(M,2,1),
-                           FlxqX_mul_pre(q, gcoeff(M,2,2), T, p, pi), p);
-  return gerepilecopy(av, M);
-}
+{ return FlxqX_halfgcd_all_pre(x, y, T, p, pi, NULL, NULL); }
+
 GEN
 FlxqX_halfgcd(GEN x, GEN y, GEN T, ulong p)
 { return FlxqX_halfgcd_pre(x, y, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
@@ -1293,17 +1461,19 @@ FlxqX_gcd_pre(GEN x, GEN y, GEN T, ulong p, ulong pi)
   x = FlxqX_red_pre(x, T, p, pi);
   y = FlxqX_red_pre(y, T, p, pi);
   if (!signe(x)) return gerepileupto(av, y);
-  while (lg(y)>FlxqX_GCD_LIMIT)
+  while (lgpol(y)>=FlxqX_GCD_LIMIT)
   {
-    GEN c;
     if (lgpol(y)<=(lgpol(x)>>1))
     {
       GEN r = FlxqX_rem_pre(x, y, T, p, pi);
       x = y; y = r;
     }
-    c = FlxqXM_FlxqX_mul2(FlxqX_halfgcd_pre(x,y, T, p, pi), x, y, T, p, pi);
-    x = gel(c,1); y = gel(c,2);
-    gerepileall(av,2,&x,&y);
+    (void) FlxqX_halfgcd_all_pre(x,y, T, p, pi, &x, &y);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FlxqX_gcd (y = %ld)",degpol(y));
+      gerepileall(av,2,&x,&y);
+    }
   }
   return gerepileupto(av, FlxqX_gcd_basecase(x, y, T, p, pi));
 }
@@ -1339,26 +1509,33 @@ FlxqX_extgcd_basecase(GEN a, GEN b, GEN T, ulong p,ulong pi, GEN *ptu, GEN *ptv)
 static GEN
 FlxqX_extgcd_halfgcd(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *ptu, GEN *ptv)
 {
-  pari_sp av=avma;
-  GEN u,v,R = matid2_FlxXM(varn(x), get_Flx_var(T));
-  while (lg(y)>FlxqX_EXTGCD_LIMIT)
+  GEN u,v;
+  GEN V = cgetg(expu(lgpol(y))+2,t_VEC);
+  long i, n = 0, vs = varn(x), vT = get_Flx_var(T);
+  while (lgpol(y) >= FlxqX_EXTGCD_LIMIT)
   {
-    GEN M, c;
     if (lgpol(y)<=(lgpol(x)>>1))
     {
       GEN r, q = FlxqX_divrem_pre(x, y, T, p, pi, &r);
       x = y; y = r;
-      R = FlxqX_FlxqXM_qmul(q, R, T, p, pi);
-    }
-    M = FlxqX_halfgcd_pre(x,y, T, p, pi);
-    c = FlxqXM_FlxqX_mul2(M, x,y, T, p, pi);
-    R = FlxqXM_mul2(M, R, T, p, pi);
-    x = gel(c,1); y = gel(c,2);
-    gerepileall(av,3,&x,&y,&R);
+      gel(V,++n) = mkmat22(pol_0(vs),pol1_FlxX(vs,vT),pol1_FlxX(vs,vT),FlxX_neg(q,p));
+    } else
+      gel(V,++n) = FlxqX_halfgcd_all_pre(x, y, T, p, pi, &x, &y);
   }
   y = FlxqX_extgcd_basecase(x,y, T, p, pi, &u,&v);
-  if (ptu) *ptu = FlxqX_addmulmul(u,v,gcoeff(R,1,1),gcoeff(R,2,1), T, p, pi);
-  *ptv = FlxqX_addmulmul(u,v,gcoeff(R,1,2),gcoeff(R,2,2), T, p, pi);
+  for (i = n; i>1; i--)
+  {
+    GEN R = gel(V,i);
+    GEN u1 = FlxqX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), T, p, pi);
+    GEN v1 = FlxqX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), T, p, pi);
+    u = u1; v = v1;
+  }
+  {
+    GEN R = gel(V,1);
+    if (ptu)
+      *ptu = FlxqX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), T, p, pi);
+    *ptv   = FlxqX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), T, p, pi);
+  }
   return y;
 }
 
@@ -1371,7 +1548,7 @@ FlxqX_extgcd_pre(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *ptu, GEN *ptv)
   GEN d;
   x = FlxqX_red_pre(x, T, p, pi);
   y = FlxqX_red_pre(y, T, p, pi);
-  if (lg(y)>FlxqX_EXTGCD_LIMIT)
+  if (lgpol(y)>=FlxqX_EXTGCD_LIMIT)
     d = FlxqX_extgcd_halfgcd(x, y, T, p, pi, ptu, ptv);
   else
     d = FlxqX_extgcd_basecase(x, y, T, p, pi, ptu, ptv);
@@ -1393,6 +1570,31 @@ FlxqX_saferem(GEN P, GEN Q, GEN T, ulong p, ulong pi)
   return FlxqX_rem_pre(P,Q,T,p,pi);
 }
 
+GEN
+FlxqX_eval(GEN x, GEN y, GEN T, ulong p)
+{
+  pari_sp av;
+  GEN p1, r;
+  long j, i=lg(x)-1;
+  if (i<=2)
+    return (i==2)? gcopy(gel(x,2)): pol0_Flx(get_Flx_var(T));
+  av=avma; p1=gel(x,i);
+  /* specific attention to sparse polynomials (see poleval)*/
+  /*You've guessed it! It's a copy-paste(tm)*/
+  for (i--; i>=2; i=j-1)
+  {
+    for (j=i; lg(gel(x,j))==1; j--)
+      if (j==2)
+      {
+        if (i!=j) y = Flxq_powu(y, i-j+1, T, p);
+        return gerepileupto(av, Flxq_mul(p1,y, T, p));
+      }
+    r = (i==j)? y: Flxq_powu(y, i-j+1, T, p);
+    p1 = Flx_add(Flxq_mul(p1,r,T,p), gel(x,j), p);
+  }
+  return gerepileupto(av, p1);
+}
+
 GEN
 FlxqX_safegcd(GEN P, GEN Q, GEN T, ulong p)
 {
@@ -1421,28 +1623,6 @@ FlxqX_safegcd(GEN P, GEN Q, GEN T, ulong p)
   return gerepileupto(av, Q);
 }
 
-struct _FlxqX {ulong p, pi; GEN T;};
-static GEN _FlxqX_mul(void *data,GEN a,GEN b)
-{
-  struct _FlxqX *d=(struct _FlxqX*)data;
-  return FlxqX_mul_pre(a,b,d->T,d->p,d->pi);
-}
-static GEN _FlxqX_sqr(void *data,GEN a)
-{
-  struct _FlxqX *d=(struct _FlxqX*)data;
-  return FlxqX_sqr_pre(a,d->T,d->p,d->pi);
-}
-
-GEN
-FlxqX_powu_pre(GEN V, ulong n, GEN T, ulong p, ulong pi)
-{
-  struct _FlxqX d; d.p = p; d.pi = pi; d.T = T;
-  return gen_powu(V, n, (void*)&d, &_FlxqX_sqr, &_FlxqX_mul);
-}
-GEN
-FlxqX_powu(GEN V, ulong n, GEN T, ulong p)
-{ return FlxqX_powu_pre(V, n, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
-
 /* Res(A,B) = Res(B,R) * lc(B)^(a-r) * (-1)^(ab), with R=A%B, a=deg(A) ...*/
 GEN
 FlxqX_saferesultant(GEN a, GEN b, GEN T, ulong p)
@@ -1486,13 +1666,31 @@ FlxqX_saferesultant(GEN a, GEN b, GEN T, ulong p)
   return gerepileupto(av, res);
 }
 
-/* Res(A,B) = Res(B,R) * lc(B)^(a-r) * (-1)^(ab), with R=A%B, a=deg(A) ...*/
-GEN
-FlxqX_resultant(GEN a, GEN b, GEN T, ulong p)
+static GEN
+FlxqX_halfres(GEN x, GEN y, GEN T, ulong p, ulong pi, GEN *a, GEN *b, GEN *r)
 {
+  struct FlxqX_res res;
+  GEN V;
+  long dB;
+
+  res.res  = *r;
+  res.lc   = leading_coeff(y);
+  res.deg0 = degpol(x);
+  res.deg1 = degpol(y);
+  res.off = 0;
+  V = FlxqX_halfres_i(x, y, T, p, pi, a, b, &res);
+  dB = degpol(*b);
+  if (dB < degpol(y))
+    FlxqX_halfres_update(res.deg0, res.deg1, dB, T, p, pi, &res);
+  *r = res.res;
+  return V;
+}
+
+static GEN
+FlxqX_resultant_basecase(GEN a, GEN b, GEN T, ulong p, ulong pi)
+{
+  pari_sp av = avma;
   long vT = get_Flx_var(T), da,db,dc;
-  ulong pi;
-  pari_sp av;
   GEN c,lb, res = pol1_Flx(vT);
 
   if (!signe(a) || !signe(b)) return pol0_Flx(vT);
@@ -1505,7 +1703,6 @@ FlxqX_resultant(GEN a, GEN b, GEN T, ulong p)
     if (both_odd(da,db)) res = Flx_neg(res, p);
   }
   if (!da) return pol1_Flx(vT); /* = res * a[2] ^ db, since 0 <= db <= da = 0 */
-  pi = SMALL_ULONG(p)? 0: get_Fl_red(p); av = avma;
   while (db)
   {
     lb = gel(b,db+2);
@@ -1528,6 +1725,48 @@ FlxqX_resultant(GEN a, GEN b, GEN T, ulong p)
   return gerepileupto(av, res);
 }
 
+/* Res(A,B) = Res(B,R) * lc(B)^(a-r) * (-1)^(ab), with R=A%B, a=deg(A) ...*/
+GEN
+FlxqX_resultant_pre(GEN x, GEN y, GEN T, ulong p, ulong pi)
+{
+  pari_sp av = avma;
+  long dx, dy, vT = get_Flx_var(T);
+  GEN res = pol1_Flx(vT);
+  if (!signe(x) || !signe(y)) return pol0_Flx(vT);
+  dx = degpol(x); dy = degpol(y);
+  if (dx < dy)
+  {
+    swap(x,y);
+    if (both_odd(dx, dy))
+      res = Flx_neg(res, p);
+  }
+  while (lgpol(y) >= FlxqX_GCD_LIMIT)
+  {
+    if (lgpol(y)<=(lgpol(x)>>1))
+    {
+      GEN r = FlxqX_rem_pre(x, y, T, p, pi);
+      long dx = degpol(x), dy = degpol(y), dr = degpol(r);
+      GEN ly = gel(y,dy+2);
+      if (!Flx_equal1(ly))
+        res = Flxq_mul_pre(res, Flxq_powu_pre(ly, dx - dr, T, p, pi), T, p, pi);
+      if (both_odd(dx, dy))
+        res = Flx_neg(res, p);
+      x = y; y = r;
+    }
+    (void) FlxqX_halfres(x, y, T, p, pi, &x, &y, &res);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FlxqX_resultant (y = %ld)",degpol(y));
+      gerepileall(av,3,&x,&y,&res);
+    }
+  }
+  res = Flxq_mul_pre(res, FlxqX_resultant_basecase(x, y, T, p, pi), T, p, pi);
+  return gerepileupto(av, res);
+}
+GEN
+FlxqX_resultant(GEN x, GEN y, GEN T, ulong p)
+{ return FlxqX_resultant_pre(x, y, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
+
 /* disc P = (-1)^(n(n-1)/2) lc(P)^(n - deg P' - 2) Res(P,P'), n = deg P */
 GEN
 FlxqX_disc(GEN P, GEN T, ulong p)
@@ -1579,9 +1818,10 @@ GEN
 FlxqX_fromNewton(GEN P, GEN T, ulong p)
 { return FlxqX_fromNewton_pre(P, T, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
 
-static GEN
+GEN
 FlxqX_composedsum(GEN P, GEN Q, GEN T, ulong p)
 {
+  pari_sp av = avma;
   ulong pi = SMALL_ULONG(p)? 0: get_Fl_red(p);
   long n = 1+ degpol(P)*degpol(Q);
   GEN Pl = FlxX_invLaplace(FlxqX_Newton_pre(P,n, T,p,pi), p);
@@ -1591,13 +1831,9 @@ FlxqX_composedsum(GEN P, GEN Q, GEN T, ulong p)
   GEN lead = Flxq_mul_pre(Flxq_powu_pre(leading_coeff(P),degpol(Q), T,p,pi),
                           Flxq_powu_pre(leading_coeff(Q),degpol(P), T,p,pi),
                           T, p, pi);
-  return FlxqX_Flxq_mul_pre(R, lead, T, p, pi);
+  return gerepileupto(av, FlxqX_Flxq_mul_pre(R, lead, T, p, pi));
 }
 
-GEN
-FlxqX_direct_compositum(GEN P, GEN Q, GEN T, ulong p)
-{ return FlxqX_composedsum(P, Q, T, p); }
-
 GEN
 FlxqXV_prod(GEN V, GEN T, ulong p)
 {
