@@ -501,7 +501,7 @@ FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
     return gc_const((pari_sp)rem, z-2);
   }
   lr=i+3; rem -= lr;
-  rem[0] = evaltyp(t_POL) | evallg(lr);
+  rem[0] = evaltyp(t_POL) | _evallg(lr);
   rem[1] = z[-1];
   p1 = gerepileuptoint((pari_sp)rem, p1);
   rem += 2; gel(rem,i) = p1;
@@ -585,29 +585,6 @@ FpX_valrem(GEN x, GEN t, GEN p, GEN *py)
   return k;
 }
 
-static GEN
-FpX_halfgcd_basecase(GEN a, GEN b, GEN p)
-{
-  pari_sp av=avma;
-  GEN u,u1,v,v1;
-  long vx = varn(a);
-  long n = lgpol(a)>>1;
-  u1 = v = pol_0(vx);
-  u = v1 = pol_1(vx);
-  while (lgpol(b)>n)
-  {
-    GEN r, q = FpX_divrem(a,b,p, &r);
-    a = b; b = r; swap(u,u1); swap(v,v1);
-    u1 = FpX_sub(u1, FpX_mul(u, q, p), p);
-    v1 = FpX_sub(v1, FpX_mul(v, q ,p), p);
-    if (gc_needed(av,2))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_halfgcd (d = %ld)",degpol(b));
-      gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
-    }
-  }
-  return gerepilecopy(av, mkmat2(mkcol2(u,u1), mkcol2(v,v1)));
-}
 static GEN
 FpX_addmulmul(GEN u, GEN v, GEN x, GEN y, GEN p)
 {
@@ -637,86 +614,225 @@ FpXM_mul2(GEN A, GEN B, GEN p)
   GEN M7 = FpX_mul(FpX_sub(A12,A22, p), FpX_add(B21,B22, p), p);
   GEN T1 = FpX_add(M1,M4, p), T2 = FpX_sub(M7,M5, p);
   GEN T3 = FpX_sub(M1,M2, p), T4 = FpX_add(M3,M6, p);
-  retmkmat2(mkcol2(FpX_add(T1,T2, p), FpX_add(M2,M4, p)),
-            mkcol2(FpX_add(M3,M5, p), FpX_add(T3,T4, p)));
+  retmkmat22(FpX_add(T1,T2, p), FpX_add(M3,M5, p),
+             FpX_add(M2,M4, p), FpX_add(T3,T4, p));
 }
 
 /* Return [0,1;1,-q]*M */
 static GEN
 FpX_FpXM_qmul(GEN q, GEN M, GEN p)
 {
-  GEN u, v, res = cgetg(3, t_MAT);
-  u = FpX_sub(gcoeff(M,1,1), FpX_mul(gcoeff(M,2,1), q, p), p);
-  gel(res,1) = mkcol2(gcoeff(M,2,1), u);
-  v = FpX_sub(gcoeff(M,1,2), FpX_mul(gcoeff(M,2,2), q, p), p);
-  gel(res,2) = mkcol2(gcoeff(M,2,2), v);
-  return res;
+  GEN u = FpX_mul(gcoeff(M,2,1), q, p);
+  GEN v = FpX_mul(gcoeff(M,2,2), q, p);
+  retmkmat22(gcoeff(M,2,1), gcoeff(M,2,2),
+    FpX_sub(gcoeff(M,1,1), u, p), FpX_sub(gcoeff(M,1,2), v, p));
 }
 
 static GEN
 matid2_FpXM(long v)
+{ retmkmat22(pol_1(v), pol_0(v), pol_0(v), pol_1(v)); }
+
+static GEN
+matJ2_FpXM(long v)
+{ retmkmat22(pol_0(v), pol_1(v), pol_1(v), pol_0(v)); }
+
+INLINE GEN
+FpX_shift(GEN a, long n) { return RgX_shift_shallow(a, n); }
+
+INLINE GEN
+FpXn_red(GEN a, long n) { return RgXn_red_shallow(a, n); }
+
+/* Fast resultant formula from William Hart in Flint <http://flintlib.org/> */
+
+struct FpX_res
 {
-  retmkmat2(mkcol2(pol_1(v),pol_0(v)),
-            mkcol2(pol_0(v),pol_1(v)));
+   GEN res, lc;
+   long deg0, deg1, off;
+};
+
+INLINE void
+FpX_halfres_update(long da, long db, long dr, GEN p, struct FpX_res *res)
+{
+  if (dr >= 0)
+  {
+    if (!equali1(res->lc))
+    {
+      res->lc  = Fp_powu(res->lc, da - dr, p);
+      res->res = Fp_mul(res->res, res->lc, p);
+    }
+    if (both_odd(da + res->off, db + res->off))
+      res->res = Fp_neg(res->res, p);
+  } else
+  {
+    if (db == 0)
+    {
+      if (!equali1(res->lc))
+      {
+          res->lc  = Fp_powu(res->lc, da, p);
+          res->res = Fp_mul(res->res, res->lc, p);
+      }
+    } else
+      res->res = gen_0;
+  }
 }
 
 static GEN
-FpX_shift(GEN a, long n) { return RgX_shift_shallow(a, n); }
-
-static GEN
-FpX_halfgcd_split(GEN x, GEN y, GEN p)
+FpX_halfres_basecase(GEN a, GEN b, GEN p, GEN *pa, GEN *pb, struct FpX_res *res)
 {
   pari_sp av=avma;
-  GEN R, S, V;
-  GEN y1, r, q;
+  GEN u,u1,v,v1, M;
+  long vx = varn(a), n = lgpol(a)>>1;
+  u1 = v = pol_0(vx);
+  u = v1 = pol_1(vx);
+  while (lgpol(b)>n)
+  {
+    GEN r, q;
+    q = FpX_divrem(a,b,p, &r);
+    if (res)
+    {
+      long da = degpol(a), db=degpol(b), dr = degpol(r);
+      res->lc = leading_coeff(b);
+      if (dr >= n)
+        FpX_halfres_update(da,db,dr,p,res);
+      else
+      {
+        res->deg0 = da;
+        res->deg1 = db;
+      }
+    }
+    a = b; b = r; swap(u,u1); swap(v,v1);
+    u1 = FpX_sub(u1, FpX_mul(u, q, p), p);
+    v1 = FpX_sub(v1, FpX_mul(v, q, p), p);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_halfgcd (d = %ld)",degpol(b));
+      gerepileall(av,res ? 8: 6, &a,&b,&u1,&v1,&u,&v,&res->res,&res->lc);
+    }
+  }
+  M = mkmat22(u,v,u1,v1); *pa = a; *pb = b;
+  return gc_all(av, res ? 5: 3, &M, pa, pb, &res->res, &res->lc);
+}
+
+static GEN FpX_halfres_i(GEN x, GEN y, GEN p, GEN *a, GEN *b, struct FpX_res *res);
+
+static GEN
+FpX_halfres_split(GEN x, GEN y, GEN p, GEN *a, GEN *b, struct FpX_res *res)
+{
+  pari_sp av = avma;
+  GEN R, S, T, V1, V2;
+  GEN x1, y1, r, q;
   long l = lgpol(x), n = l>>1, k;
-  if (lgpol(y)<=n) return matid2_FpXM(varn(x));
-  R = FpX_halfgcd(FpX_shift(x,-n), FpX_shift(y,-n), p);
-  V = FpXM_FpX_mul2(R,x,y,p); y1 = gel(V,2);
-  if (lgpol(y1)<=n) return gerepilecopy(av, R);
-  q = FpX_divrem(gel(V,1), y1, p, &r);
+  if (lgpol(y) <= n)
+    { *a = RgX_copy(x); *b = RgX_copy(y); return matid2_FpXM(varn(x)); }
+  if (res)
+  {
+     res->lc = leading_coeff(y);
+     res->deg0 -= n;
+     res->deg1 -= n;
+     res->off += n;
+  }
+  R = FpX_halfres_i(FpX_shift(x,-n), FpX_shift(y,-n), p, a, b, res);
+  if (res)
+  {
+    res->off -= n;
+    res->deg0 += n;
+    res->deg1 += n;
+  }
+  V1 = FpXM_FpX_mul2(R, FpXn_red(x,n), FpXn_red(y,n), p);
+  x1 = FpX_add(FpX_shift(*a,n), gel(V1,1), p);
+  y1 = FpX_add(FpX_shift(*b,n), gel(V1,2), p);
+  if (lgpol(y1) <= n)
+  {
+    *a = x1; *b = y1;
+    return gc_all(av, res ? 5: 3, &R, a, b, &res->res, &res->lc);
+  }
   k = 2*n-degpol(y1);
-  S = FpX_halfgcd(FpX_shift(y1,-k), FpX_shift(r,-k), p);
-  return gerepileupto(av, FpXM_mul2(S,FpX_FpXM_qmul(q,R,p),p));
+  q = FpX_divrem(x1, y1, p, &r);
+  if (res)
+  {
+    long dx1 = degpol(x1), dy1 = degpol(y1), dr = degpol(r);
+    if (dy1 < degpol(y))
+      FpX_halfres_update(res->deg0, res->deg1, dy1, p,res);
+    res->lc = gel(y1, dy1+2);
+    res->deg0 = dx1;
+    res->deg1 = dy1;
+    if (dr >= n)
+    {
+      FpX_halfres_update(dx1, dy1, dr, p,res);
+      res->deg0 = dy1;
+      res->deg1 = dr;
+    }
+    res->deg0 -= k;
+    res->deg1 -= k;
+    res->off += k;
+  }
+  S = FpX_halfres_i(FpX_shift(y1,-k), FpX_shift(r,-k), p, a, b, res);
+  if (res)
+  {
+    res->deg0 += k;
+    res->deg1 += k;
+    res->off -= k;
+  }
+  T = FpXM_mul2(S, FpX_FpXM_qmul(q, R, p), p);
+  V2 = FpXM_FpX_mul2(S, FpXn_red(y1,k), FpXn_red(r,k), p);
+  *a = FpX_add(FpX_shift(*a,k), gel(V2,1), p);
+  *b = FpX_add(FpX_shift(*b,k), gel(V2,2), p);
+  return gc_all(av, res ? 5: 3, &T, a, b, &res->res, &res->lc);
+}
+
+static GEN
+FpX_halfres_i(GEN x, GEN y, GEN p, GEN *a, GEN *b, struct FpX_res *res)
+{
+  if (lgpol(x) < FpX_HALFGCD_LIMIT)
+    return FpX_halfres_basecase(x, y, p, a, b, res);
+  return FpX_halfres_split(x, y, p, a, b, res);
+}
+
+static GEN
+FpX_halfgcd_all_i(GEN x, GEN y, GEN p, GEN *pa, GEN *pb)
+{
+  GEN a, b;
+  GEN R = FpX_halfres_i(x, y, p, &a, &b, NULL);
+  if (pa) *pa = a;
+  if (pb) *pb = b;
+  return R;
 }
 
 /* Return M in GL_2(Fp[X]) such that:
 if [a',b']~=M*[a,b]~ then degpol(a')>= (lgpol(a)>>1) >degpol(b')
 */
 
-static GEN
-FpX_halfgcd_i(GEN x, GEN y, GEN p)
+GEN
+FpX_halfgcd_all(GEN x, GEN y, GEN p, GEN *a, GEN *b)
 {
-  if (lg(x)<=FpX_HALFGCD_LIMIT) return FpX_halfgcd_basecase(x,y,p);
-  return FpX_halfgcd_split(x,y,p);
+  pari_sp av = avma;
+  GEN R, q, r;
+  if (lgefint(p)==3)
+  {
+    ulong pp = to_Flx(&x, &y, p);
+    R = Flx_halfgcd_all(x, y, pp, a, b);
+    R = FlxM_to_ZXM(R);
+    if (a) *a = Flx_to_ZX(*a);
+    if (b) *b = Flx_to_ZX(*b);
+    return !a && b ? gc_all(av, 2, &R, b): gc_all(av, 1+!!a+!!b, &R, a, b);
+  }
+  if (!signe(x))
+  {
+    if (a) *a = RgX_copy(y);
+    if (b) *b = RgX_copy(x);
+    return matJ2_FpXM(varn(x));
+  }
+  if (degpol(y)<degpol(x)) return FpX_halfgcd_all_i(x, y, p, a, b);
+  q = FpX_divrem(y,x,p,&r);
+  R = FpX_halfgcd_all_i(x, r, p, a, b);
+  gcoeff(R,1,1) = FpX_sub(gcoeff(R,1,1), FpX_mul(q, gcoeff(R,1,2), p), p);
+  gcoeff(R,2,1) = FpX_sub(gcoeff(R,2,1), FpX_mul(q, gcoeff(R,2,2), p), p);
+  return !a && b ? gc_all(av, 2, &R, b): gc_all(av, 1+!!a+!!b, &R, a, b);
 }
 
 GEN
 FpX_halfgcd(GEN x, GEN y, GEN p)
-{
-  pari_sp av = avma;
-  GEN M,q,r;
-  if (lgefint(p)==3)
-  {
-    ulong pp = to_Flx(&x, &y, p);
-    M = FlxM_to_ZXM(Flx_halfgcd(x, y, pp));
-  }
-  else
-  {
-    if (!signe(x))
-    {
-      long v = varn(x);
-      retmkmat2(mkcol2(pol_0(v),pol_1(v)),
-                mkcol2(pol_1(v),pol_0(v)));
-    }
-    if (degpol(y)<degpol(x)) return FpX_halfgcd_i(x,y,p);
-    q = FpX_divrem(y,x,p,&r);
-    M = FpX_halfgcd_i(x,r,p);
-    gcoeff(M,1,1) = FpX_sub(gcoeff(M,1,1), FpX_mul(q, gcoeff(M,1,2), p), p);
-    gcoeff(M,2,1) = FpX_sub(gcoeff(M,2,1), FpX_mul(q, gcoeff(M,2,2), p), p);
-  }
-  return gerepilecopy(av, M);
-}
+{ return FpX_halfgcd_all(x, y, p, NULL, NULL); }
 
 static GEN
 FpX_gcd_basecase(GEN a, GEN b, GEN p)
@@ -750,17 +866,19 @@ FpX_gcd(GEN x, GEN y, GEN p)
   x = FpX_red(x, p);
   y = FpX_red(y, p);
   if (!signe(x)) return gerepileupto(av, y);
-  while (lg(y)>FpX_GCD_LIMIT)
+  while (lgpol(y) >= FpX_GCD_LIMIT)
   {
-    GEN c;
     if (lgpol(y)<=(lgpol(x)>>1))
     {
       GEN r = FpX_rem(x, y, p);
       x = y; y = r;
     }
-    c = FpXM_FpX_mul2(FpX_halfgcd(x,y, p), x, y, p);
-    x = gel(c,1); y = gel(c,2);
-    gerepileall(av,2,&x,&y);
+    (void) FpX_halfgcd_all(x, y, p, &x, &y);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_gcd (y = %ld)",degpol(y));
+      gerepileall(av,2,&x,&y);
+    }
   }
   return gerepileupto(av, FpX_gcd_basecase(x,y,p));
 }
@@ -793,49 +911,63 @@ static GEN
 FpX_extgcd_basecase(GEN a, GEN b, GEN p, GEN *ptu, GEN *ptv)
 {
   pari_sp av=avma;
-  GEN u,v,d,d1,v1;
+  GEN v,v1, A = a, B = b;
   long vx = varn(a);
-  d = a; d1 = b;
-  v = pol_0(vx); v1 = pol_1(vx);
-  while (signe(d1))
+  if (!lgpol(b))
   {
-    GEN r, q = FpX_divrem(d,d1,p, &r);
-    v = FpX_sub(v,FpX_mul(q,v1,p),p);
-    u=v; v=v1; v1=u;
-    u=r; d=d1; d1=u;
+    if (ptu) *ptu = pol_1(vx);
+    *ptv = pol_0(vx);
+    return RgX_copy(a);
+  }
+  v = pol_0(vx); v1 = pol_1(vx);
+  while (1)
+  {
+    GEN r, q = FpX_divrem(a,b,p, &r);
+    a = b; b = r;
+    swap(v,v1);
+    if (!lgpol(b)) break;
+    v1 = FpX_sub(v1, FpX_mul(v, q, p), p);
     if (gc_needed(av,2))
     {
-      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_extgcd (d = %ld)",degpol(d));
-      gerepileall(av,5, &d,&d1,&u,&v,&v1);
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_extgcd (d = %ld)",degpol(a));
+      gerepileall(av,4,&a,&b,&v,&v1);
     }
   }
-  if (ptu) *ptu = FpX_div(FpX_sub(d,FpX_mul(b,v,p),p),a,p);
-  *ptv = v; return d;
+  if (ptu) *ptu = FpX_div(FpX_sub(a,FpX_mul(B,v,p),p),A,p);
+  *ptv = v;
+  return a;
 }
 
 static GEN
 FpX_extgcd_halfgcd(GEN x, GEN y, GEN p, GEN *ptu, GEN *ptv)
 {
-  pari_sp av=avma;
-  GEN u,v,R = matid2_FpXM(varn(x));
-  while (lg(y)>FpX_EXTGCD_LIMIT)
+  GEN u, v;
+  GEN V = cgetg(expu(lgpol(y))+2,t_VEC);
+  long i, n = 0, vs = varn(x);
+  while (lgpol(y) >= FpX_EXTGCD_LIMIT)
   {
-    GEN M, c;
     if (lgpol(y)<=(lgpol(x)>>1))
     {
       GEN r, q = FpX_divrem(x, y, p, &r);
       x = y; y = r;
-      R = FpX_FpXM_qmul(q, R, p);
-    }
-    M = FpX_halfgcd(x,y, p);
-    c = FpXM_FpX_mul2(M, x,y, p);
-    R = FpXM_mul2(M, R, p);
-    x = gel(c,1); y = gel(c,2);
-    gerepileall(av,3,&x,&y,&R);
+      gel(V,++n) = mkmat22(pol_0(vs),pol_1(vs),pol_1(vs),FpX_neg(q,p));
+    } else
+      gel(V,++n) = FpX_halfgcd_all(x, y, p, &x, &y);
+  }
+  y = FpX_extgcd_basecase(x, y, p, &u, &v);
+  for (i = n; i>1; i--)
+  {
+    GEN R = gel(V,i);
+    GEN u1 = FpX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), p);
+    GEN v1 = FpX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), p);
+    u = u1; v = v1;
+  }
+  {
+    GEN R = gel(V,1);
+    if (ptu)
+      *ptu = FpX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), p);
+    *ptv   = FpX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), p);
   }
-  y = FpX_extgcd_basecase(x,y,p,&u,&v);
-  if (ptu) *ptu = FpX_addmulmul(u,v,gcoeff(R,1,1),gcoeff(R,2,1),p);
-  *ptv = FpX_addmulmul(u,v,gcoeff(R,1,2),gcoeff(R,2,2),p);
   return y;
 }
 
@@ -851,14 +983,14 @@ FpX_extgcd(GEN x, GEN y, GEN p, GEN *ptu, GEN *ptv)
     ulong pp = to_Flx(&x, &y, p);
     d = Flx_extgcd(x,y, pp, ptu,ptv);
     d = Flx_to_ZX(d);
-    if (ptu) *ptu=Flx_to_ZX(*ptu);
-    *ptv=Flx_to_ZX(*ptv);
+    if (ptu) *ptu = Flx_to_ZX(*ptu);
+    *ptv = Flx_to_ZX(*ptv);
   }
   else
   {
     x = FpX_red(x, p);
     y = FpX_red(y, p);
-    if (lg(y)>FpX_EXTGCD_LIMIT)
+    if (lgpol(y) >= FpX_EXTGCD_LIMIT)
       d = FpX_extgcd_halfgcd(x, y, p, ptu, ptv);
     else
       d = FpX_extgcd_basecase(x, y, p, ptu, ptv);
@@ -866,6 +998,195 @@ FpX_extgcd(GEN x, GEN y, GEN p, GEN *ptu, GEN *ptv)
   return gc_all(av, ptu?3:2, &d, ptv, ptu);
 }
 
+static GEN
+FpX_halfres(GEN x, GEN y, GEN p, GEN *a, GEN *b, GEN *r)
+{
+  struct FpX_res res;
+  GEN V;
+  long dB;
+
+  res.res  = *r;
+  res.lc   = leading_coeff(y);
+  res.deg0 = degpol(x);
+  res.deg1 = degpol(y);
+  res.off = 0;
+  V = FpX_halfres_i(x, y, p, a, b, &res);
+  dB = degpol(*b);
+  if (dB < degpol(y))
+    FpX_halfres_update(res.deg0, res.deg1, dB, p, &res);
+  *r = res.res;
+  return V;
+}
+
+static GEN
+FpX_resultant_basecase(GEN a, GEN b, GEN p)
+{
+  pari_sp av = avma;
+  long da,db,dc;
+  GEN c, lb, res = gen_1;
+
+  if (!signe(a) || !signe(b)) return pol_0(varn(a));
+
+  da = degpol(a);
+  db = degpol(b);
+  if (db > da)
+  {
+    swapspec(a,b, da,db);
+    if (both_odd(da,db)) res = subii(p, res);
+  }
+  if (!da) return gc_const(av, gen_1); /* = res * a[2] ^ db, since 0 <= db <= da = 0 */
+  while (db)
+  {
+    lb = gel(b,db+2);
+    c = FpX_rem(a,b, p);
+    a = b; b = c; dc = degpol(c);
+    if (dc < 0) return gc_const(av, gen_0);
+
+    if (both_odd(da,db)) res = subii(p, res);
+    if (!equali1(lb)) res = Fp_mul(res, Fp_powu(lb, da - dc, p), p);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_resultant (da = %ld)",da);
+      gerepileall(av,3, &a,&b,&res);
+    }
+    da = db; /* = degpol(a) */
+    db = dc; /* = degpol(b) */
+  }
+  return gerepileuptoint(av, Fp_mul(res, Fp_powu(gel(b,2), da, p), p));
+}
+
+GEN
+FpX_resultant(GEN x, GEN y, GEN p)
+{
+  pari_sp av = avma;
+  long dx, dy;
+  GEN res = gen_1;
+  if (!signe(x) || !signe(y)) return gen_0;
+  if (lgefint(p) == 3)
+  {
+    pari_sp av = avma;
+    ulong pp = to_Flx(&x, &y, p);
+    ulong res = Flx_resultant(x, y, pp);
+    return gc_utoi(av, res);
+  }
+  dx = degpol(x); dy = degpol(y);
+  if (dx < dy)
+  {
+    swap(x,y);
+    if (both_odd(dx, dy))
+      res = Fp_neg(res, p);
+  }
+  while (lgpol(y) >= FpX_GCD_LIMIT)
+  {
+    if (lgpol(y)<=(lgpol(x)>>1))
+    {
+      GEN r = FpX_rem(x, y, p);
+      long dx = degpol(x), dy = degpol(y), dr = degpol(r);
+      GEN ly = gel(y,dy+2);
+      if (!equali1(ly)) res = Fp_mul(res, Fp_powu(ly, dx - dr, p), p);
+      if (both_odd(dx, dy))
+        res = Fp_neg(res, p);
+      x = y; y = r;
+    }
+    (void) FpX_halfres(x, y, p, &x, &y, &res);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_res (y = %ld)",degpol(y));
+      gerepileall(av,3,&x,&y,&res);
+    }
+  }
+  return gerepileuptoint(av, Fp_mul(res, FpX_resultant_basecase(x, y, p), p));
+}
+
+/* If resultant is 0, *ptU and *ptV are not set */
+static GEN
+FpX_extresultant_basecase(GEN a, GEN b, GEN p, GEN *ptU, GEN *ptV)
+{
+  pari_sp av = avma;
+  GEN z,q,u,v, x = a, y = b;
+  GEN lb, res = gen_1;
+  long dx, dy, dz;
+  long vs = varn(a);
+
+  u = pol_0(vs);
+  v = pol_1(vs); /* v = 1 */
+  dx = degpol(x);
+  dy = degpol(y);
+  while (dy)
+  { /* b u = x (a), b v = y (a) */
+    lb = gel(y,dy+2);
+    q = FpX_divrem(x,y, p, &z);
+    x = y; y = z; /* (x,y) = (y, x - q y) */
+    dz = degpol(z); if (dz < 0) return gc_const(av,gen_0);
+    z = FpX_sub(u, FpX_mul(q,v, p), p);
+    u = v; v = z; /* (u,v) = (v, u - q v) */
+
+    if (both_odd(dx,dy)) res = Fp_neg(res, p);
+    if (!equali1(lb)) res = Fp_mul(res, Fp_powu(lb, dx-dz, p), p);
+    dx = dy; /* = degpol(x) */
+    dy = dz; /* = degpol(y) */
+  }
+  res = Fp_mul(res, Fp_powu(gel(y,2), dx, p), p);
+  lb = Fp_mul(res, Fp_inv(gel(y,2),p), p);
+  v = FpX_Fp_mul(v, lb, p);
+  u = Fp_FpX_sub(res, FpX_mul(b,v,p), p);
+  u = FpX_div(u,a,p); /* = (res - b v) / a */
+  *ptU = u;
+  *ptV = v;
+  return res;
+}
+
+GEN
+FpX_extresultant(GEN x, GEN y, GEN p, GEN *ptU, GEN *ptV)
+{
+  pari_sp av=avma;
+  GEN u, v, R;
+  GEN res = gen_1, res1;
+  long dx = degpol(x), dy = degpol(y);
+  if (lgefint(p) == 3)
+  {
+    pari_sp av = avma;
+    ulong pp = to_Flx(&x, &y, p);
+    ulong resp = Flx_extresultant(x, y, pp, &u, &v);
+    if (!resp) return gc_const(av, gen_0);
+    res = utoi(resp);
+    *ptU = Flx_to_ZX(u); *ptV = Flx_to_ZX(v);
+    return gc_all(av, 3, &res, ptU, ptV);
+  }
+  if (dy > dx)
+  {
+    swap(x,y); lswap(dx,dy);
+    if (both_odd(dx,dy)) res = Fp_neg(res,p);
+    R = matJ2_FpXM(x[1]);
+  } else R = matid2_FpXM(x[1]);
+  if (dy < 0) return gen_0;
+  while (lgpol(y) >= FpX_EXTGCD_LIMIT)
+  {
+    GEN M;
+    if (lgpol(y)<=(lgpol(x)>>1))
+    {
+      GEN r, q = FpX_divrem(x, y, p, &r);
+      long dx = degpol(x), dy = degpol(y), dr = degpol(r);
+      GEN ly = gel(y,dy+2);
+      if (!equali1(ly)) res = Fp_mul(res, Fp_powu(ly, dx - dr, p), p);
+      if (both_odd(dx, dy))
+        res = Fp_neg(res, p);
+      x = y; y = r;
+      R = FpX_FpXM_qmul(q, R, p);
+    }
+    M = FpX_halfres(x, y, p, &x, &y, &res);
+    if (!signe(res)) return gc_const(av, gen_0);
+    R = FpXM_mul2(M, R, p);
+    gerepileall(av,4,&x,&y,&R,&res);
+  }
+  res1 = FpX_extresultant_basecase(x,y,p,&u,&v);
+  if (!signe(res1)) return gc_const(av, gen_0);
+  *ptU = FpX_Fp_mul(FpX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), p), res, p);
+  *ptV = FpX_Fp_mul(FpX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), p), res, p);
+  res = Fp_mul(res1,res,p);
+  return gc_all(av, 3, &res, ptU, ptV);
+}
+
 GEN
 FpX_rescale(GEN P, GEN h, GEN p)
 {
@@ -1077,54 +1398,6 @@ FpX_chinese_coprime(GEN x,GEN y,GEN Tx,GEN Ty,GEN Tz,GEN p)
   return gerepileupto(av,p1);
 }
 
-/* Res(A,B) = Res(B,R) * lc(B)^(a-r) * (-1)^(ab), with R=A%B, a=deg(A) ...*/
-GEN
-FpX_resultant(GEN a, GEN b, GEN p)
-{
-  long da,db,dc;
-  pari_sp av;
-  GEN c,lb, res = gen_1;
-
-  if (!signe(a) || !signe(b)) return gen_0;
-  if (lgefint(p) == 3)
-  {
-    pari_sp av = avma;
-    ulong pp = to_Flx(&a, &b, p);
-    long r = Flx_resultant(a, b, pp);
-    set_avma(av);
-    return utoi(r);
-  }
-
-  da = degpol(a);
-  db = degpol(b);
-  if (db > da)
-  {
-    swapspec(a,b, da,db);
-    if (both_odd(da,db)) res = subii(p, res);
-  }
-  if (!da) return gen_1; /* = res * a[2] ^ db, since 0 <= db <= da = 0 */
-  av = avma;
-  while (db)
-  {
-    lb = gel(b,db+2);
-    c = FpX_rem(a,b, p);
-    a = b; b = c; dc = degpol(c);
-    if (dc < 0) return gc_const(av, gen_0);
-
-    if (both_odd(da,db)) res = subii(p, res);
-    if (!equali1(lb)) res = Fp_mul(res, Fp_powu(lb, da - dc, p), p);
-    if (gc_needed(av,2))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"FpX_resultant (da = %ld)",da);
-      gerepileall(av,3, &a,&b,&res);
-    }
-    da = db; /* = degpol(a) */
-    db = dc; /* = degpol(b) */
-  }
-  res = Fp_mul(res, Fp_powu(gel(b,2), da, p), p);
-  return gerepileuptoint(av, res);
-}
-
 /* disc P = (-1)^(n(n-1)/2) lc(P)^(n - deg P' - 2) Res(P,P'), n = deg P */
 GEN
 FpX_disc(GEN P, GEN p)
@@ -2420,10 +2693,71 @@ FpXQ_sqrtn(GEN a, GEN n, GEN T, GEN p, GEN *zeta)
   return gc_all(av, 2, &z,zeta);
 }
 
-GEN
-FpXQ_sqrt(GEN a, GEN T, GEN p)
+static GEN
+Fp2_norm(GEN x, GEN D, GEN p)
 {
-  return FpXQ_sqrtn(a, gen_2, T, p, NULL);
+  GEN a = gel(x,1), b = gel(x,2);
+  if (signe(b)==0) return Fp_sqr(a,p);
+  return Fp_sub(sqri(a), mulii(D, Fp_sqr(b, p)), p);
+}
+
+static GEN
+Fp2_sqrt(GEN z, GEN D, GEN p)
+{
+  GEN a = gel(z,1), b = gel(z,2), as2, u, v, s;
+  GEN y = Fp_2gener_i(D, p);
+  if (signe(b)==0)
+    return kronecker(a, p)==1 ? mkvec2(Fp_sqrt_i(a, y, p), gen_0)
+                              : mkvec2(gen_0,Fp_sqrt_i(Fp_div(a, D, p), y, p));
+  s = Fp_sqrt_i(Fp2_norm(z, D, p), y, p);
+  if(!s) return NULL;
+  as2 = Fp_halve(Fp_add(a, s, p), p);
+  if (kronecker(as2, p)==-1) as2 = Fp_sub(as2,s,p);
+  u = Fp_sqrt_i(as2, y, p);
+  v = Fp_div(b, Fp_double(u, p), p);
+  return mkvec2(u,v);
+}
+
+GEN
+FpXQ_sqrt(GEN z, GEN T, GEN p)
+{
+   pari_sp av = avma;
+  long d = get_FpX_degree(T);
+  if (lgefint(p)==3)
+  {
+    if (uel(p,2) == 2)
+    {
+      GEN r = F2xq_sqrt(ZX_to_F2x(z), ZX_to_F2x(get_FpX_mod(T)));
+      return gerepileupto(av, F2x_to_ZX(r));
+    } else
+    {
+      ulong pp = to_Flxq(&z, &T, p);
+      z = Flxq_sqrt(z, T, pp);
+      if (!z) return NULL;
+      return gerepileupto(av, Flx_to_ZX(z));
+    }
+  }
+  if (d==2)
+  {
+    GEN P = get_FpX_mod(T);
+    GEN c = gel(P,2), b = gel(P,3), a = gel(P,4), b2 = Fp_halve(b, p);
+    GEN t = Fp_div(b2, a, p);
+    GEN D = Fp_sub(Fp_sqr(b2, p), Fp_mul(a, c, p), p);
+    GEN x = degpol(z)<1 ? constant_coeff(z): Fp_sub(gel(z,2), Fp_mul(gel(z,3), t, p), p);
+    GEN y = degpol(z)<1 ? gen_0: gel(z,3);
+    GEN r = Fp2_sqrt(mkvec2(x, y), D, p), s;
+    if (!r) return gc_NULL(av);
+    s = deg1pol_shallow(gel(r,2),Fp_add(gel(r,1), Fp_mul(gel(r,2),t,p), p), varn(P));
+    return gerepilecopy(av, s);
+  }
+  if (lgpol(z)<=1 && odd(d))
+  {
+    pari_sp av = avma;
+    GEN s = Fp_sqrt(constant_coeff(z), p);
+    if (!s) return gc_NULL(av);
+    return gerepilecopy(av, scalarpol_shallow(s, get_FpX_var(T)));
+  }
+  return FpXQ_sqrtn(z, gen_2, T, p, NULL);
 }
 
 GEN
@@ -2690,10 +3024,6 @@ gener_FpXQ_local(GEN T, GEN p, GEN L)
 /**                                                                   **/
 /***********************************************************************/
 
-INLINE GEN
-FpXn_red(GEN a, long n)
-{ return RgXn_red_shallow(a, n); }
-
 GEN
 FpXn_mul(GEN a, GEN b, long n, GEN p)
 {
