@@ -24,6 +24,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /*******************************************************************/
 /*Polynomials whose coefficients are either polynomials or integers*/
 
+static GEN
+to_ZX(GEN a, long v) { return typ(a)==t_INT? scalarpol_shallow(a,v): a; }
+
 static ulong
 to_FlxqX(GEN P, GEN Q, GEN T, GEN p, GEN *pt_P, GEN *pt_Q, GEN *pt_T)
 {
@@ -283,12 +286,17 @@ to_intmod(GEN x, GEN p) { retmkintmod(modii(x, p), p); }
 GEN
 FpXQX_to_mod(GEN z, GEN T, GEN p)
 {
-  long i,l = lg(z);
-  GEN x = cgetg(l, t_POL);
-  x[1] = z[1];
-  if (l == 2) return x;
-  p = icopy(p);
-  T = FpX_to_mod_raw(T, p);
+  long i, l = lg(z);
+  GEN x;
+  if (l == 2)
+  {
+    x = cgetg(3, t_POL); x[1] = z[1];
+    p = icopy(p); T = FpX_to_mod_raw(T, p);
+    gel(x,2) = mkpolmod(mkintmod(gen_0, p), T);
+    return x;
+  }
+  x = cgetg(l, t_POL); x[1] = z[1];
+  p = icopy(p); T = FpX_to_mod_raw(T, p);
   for (i=2; i<l; i++)
   {
     GEN zi = gel(z,i);
@@ -301,10 +309,17 @@ FpXQX_to_mod(GEN z, GEN T, GEN p)
 static GEN
 FpXQX_to_mod_raw(GEN z, GEN T, GEN p)
 {
-  long i,l = lg(z);
-  GEN x = cgetg(l, t_POL);
-  x[1] = z[1];
-  if (l == 2) return x;
+  long i, l = lg(z);
+  GEN x;
+
+  if (l == 2)
+  {
+    x = cgetg(3, t_POL); x[1] = z[1];
+    p = icopy(p);
+    gel(x,2) = mkpolmod(mkintmod(gen_0, p), T);
+    return x;
+  }
+  x = cgetg(l, t_POL); x[1] = z[1];
   for (i=2; i<l; i++)
   {
     GEN zi = gel(z,i);
@@ -515,7 +530,7 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
     return gc_const((pari_sp)rem, z-2);
   }
   lr=i+3; rem -= lr;
-  rem[0] = evaltyp(t_POL) | evallg(lr);
+  rem[0] = evaltyp(t_POL) | _evallg(lr);
   rem[1] = z[-1];
   p1 = gerepile((pari_sp)rem,tetpil,p1);
   rem += 2; gel(rem,i) = p1;
@@ -533,29 +548,6 @@ FpXQX_divrem_basecase(GEN x, GEN y, GEN T, GEN p, GEN *pr)
   *pr = rem; return z-2;
 }
 
-static GEN
-FpXQX_halfgcd_basecase(GEN a, GEN b, GEN T, GEN p)
-{
-  pari_sp av=avma;
-  GEN u,u1,v,v1;
-  long vx = varn(a);
-  long n = lgpol(a)>>1;
-  u1 = v = pol_0(vx);
-  u = v1 = pol_1(vx);
-  while (lgpol(b)>n)
-  {
-    GEN r, q = FpXQX_divrem(a,b, T, p, &r);
-    a = b; b = r; swap(u,u1); swap(v,v1);
-    u1 = FpXX_sub(u1, FpXQX_mul(u, q, T, p), p);
-    v1 = FpXX_sub(v1, FpXQX_mul(v, q ,T, p), p);
-    if (gc_needed(av,2))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_halfgcd (d = %ld)",degpol(b));
-      gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
-    }
-  }
-  return gerepilecopy(av, mkmat2(mkcol2(u,u1), mkcol2(v,v1)));
-}
 static GEN
 FpXQX_addmulmul(GEN u, GEN v, GEN x, GEN y, GEN T, GEN p)
 {
@@ -585,85 +577,225 @@ FpXQXM_mul2(GEN A, GEN B, GEN T, GEN p)
   GEN M7 = FpXQX_mul(FpXX_sub(A12,A22, p), FpXX_add(B21,B22, p), T, p);
   GEN T1 = FpXX_add(M1,M4, p), T2 = FpXX_sub(M7,M5, p);
   GEN T3 = FpXX_sub(M1,M2, p), T4 = FpXX_add(M3,M6, p);
-  retmkmat2(mkcol2(FpXX_add(T1,T2, p), FpXX_add(M2,M4, p)),
-            mkcol2(FpXX_add(M3,M5, p), FpXX_add(T3,T4, p)));
+  retmkmat22(FpXX_add(T1,T2, p), FpXX_add(M3,M5, p),
+             FpXX_add(M2,M4, p), FpXX_add(T3,T4, p));
 }
 /* Return [0,1;1,-q]*M */
 static GEN
 FpXQX_FpXQXM_qmul(GEN q, GEN M, GEN T, GEN p)
 {
-  GEN u, v, res = cgetg(3, t_MAT);
-  u = FpXX_sub(gcoeff(M,1,1), FpXQX_mul(gcoeff(M,2,1), q, T, p), p);
-  gel(res,1) = mkcol2(gcoeff(M,2,1), u);
-  v = FpXX_sub(gcoeff(M,1,2), FpXQX_mul(gcoeff(M,2,2), q, T, p), p);
-  gel(res,2) = mkcol2(gcoeff(M,2,2), v);
-  return res;
+  GEN u = FpXQX_mul(gcoeff(M,2,1), q, T, p);
+  GEN v = FpXQX_mul(gcoeff(M,2,2), q, T, p);
+  retmkmat22(gcoeff(M,2,1), gcoeff(M,2,2),
+    FpXX_sub(gcoeff(M,1,1), u, p), FpXX_sub(gcoeff(M,1,2), v, p));
 }
 
 static GEN
 matid2_FpXQXM(long v)
-{
-  retmkmat2(mkcol2(pol_1(v),pol_0(v)),
-            mkcol2(pol_0(v),pol_1(v)));
-}
+{ retmkmat22(pol_1(v),pol_0(v),pol_0(v),pol_1(v)); }
+
+static GEN
+matJ2_FpXQXM(long v)
+{ retmkmat22(pol_0(v),pol_1(v),pol_1(v),pol_0(v)); }
 
 static GEN
 FpXX_shift(GEN a, long n) { return RgX_shift_shallow(a, n); }
 
-static GEN
-FpXQX_halfgcd_split(GEN x, GEN y, GEN T, GEN p)
+INLINE GEN
+FpXXn_red(GEN a, long n) { return RgXn_red_shallow(a, n); }
+
+/* Fast resultant formula from William Hart in Flint <http://flintlib.org/> */
+
+struct FpXQX_res
 {
-  pari_sp av=avma;
-  GEN R, S, V;
-  GEN y1, r, q;
-  long l = lgpol(x), n = l>>1, k;
-  if (lgpol(y)<=n) return matid2_FpXQXM(varn(x));
-  R = FpXQX_halfgcd(FpXX_shift(x,-n),FpXX_shift(y,-n), T, p);
-  V = FpXQXM_FpXQX_mul2(R,x,y, T, p); y1 = gel(V,2);
-  if (lgpol(y1)<=n) return gerepilecopy(av, R);
-  q = FpXQX_divrem(gel(V,1), y1, T, p, &r);
-  k = 2*n-degpol(y1);
-  S = FpXQX_halfgcd(FpXX_shift(y1,-k), FpXX_shift(r,-k), T, p);
-  return gerepileupto(av, FpXQXM_mul2(S,FpXQX_FpXQXM_qmul(q,R, T, p), T, p));
+   GEN res, lc;
+   long deg0, deg1, off;
+};
+
+INLINE void
+FpXQX_halfres_update(long da, long db, long dr, GEN T, GEN p, struct FpXQX_res *res)
+{
+  if (dr >= 0)
+  {
+    if (!ZX_equal1(res->lc))
+    {
+      res->lc  = FpXQ_powu(res->lc, da - dr, T, p);
+      res->res = FpXQ_mul(res->res, res->lc, T, p);
+    }
+    if (both_odd(da + res->off, db + res->off))
+      res->res = FpX_neg(res->res, p);
+  } else
+  {
+    if (db == 0)
+    {
+      if (!ZX_equal1(res->lc))
+      {
+          res->lc  = FpXQ_powu(res->lc, da, T, p);
+          res->res = FpXQ_mul(res->res, res->lc, T, p);
+      }
+    } else
+      res->res = pol_0(get_FpX_var(T));
+  }
 }
 
-/* Return M in GL_2(Fp[X]) such that:
+static GEN
+FpXQX_halfres_basecase(GEN a, GEN b, GEN T, GEN p, GEN *pa, GEN *pb, struct FpXQX_res *res)
+{
+  pari_sp av=avma;
+  GEN u,u1,v,v1, M;
+  long vx = varn(a), vT = get_FpX_var(T), n = lgpol(a)>>1;
+  u1 = v = pol_0(vx);
+  u = v1 = pol_1(vx);
+  while (lgpol(b)>n)
+  {
+    GEN r, q;
+    q = FpXQX_divrem(a,b, T, p, &r);
+    if (res)
+    {
+      long da = degpol(a), db=degpol(b), dr = degpol(r);
+      res->lc = to_ZX(gel(b,db+2),vT);
+      if (dr >= n)
+        FpXQX_halfres_update(da, db, dr, T, p, res);
+      else
+      {
+        res->deg0 = da;
+        res->deg1 = db;
+      }
+    }
+    a = b; b = r; swap(u,u1); swap(v,v1);
+    u1 = FpXX_sub(u1, FpXQX_mul(u, q, T, p), p);
+    v1 = FpXX_sub(v1, FpXQX_mul(v, q, T, p), p);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_halfgcd (d = %ld)",degpol(b));
+      gerepileall(av,res ? 8: 6, &a,&b,&u1,&v1,&u,&v,&res->res,&res->lc);
+    }
+  }
+  M = mkmat22(u,v,u1,v1); *pa = a; *pb = b;
+  return gc_all(av, res ? 5: 3, &M, pa, pb, &res->res, &res->lc);
+}
+
+static GEN FpXQX_halfres_i(GEN x, GEN y, GEN T, GEN p, GEN *a, GEN *b, struct FpXQX_res *res);
+
+static GEN
+FpXQX_halfres_split(GEN x, GEN y, GEN T, GEN p, GEN *a, GEN *b, struct FpXQX_res *res)
+{
+  pari_sp av = avma;
+  GEN Q, R, S, V1, V2;
+  GEN x1, y1, r, q;
+  long l = lgpol(x), n = l>>1, k, vT = get_FpX_var(T);
+  if (lgpol(y) <= n)
+    { *a = RgX_copy(x); *b = RgX_copy(y); return matid2_FpXQXM(varn(x)); }
+  if (res)
+  {
+     res->lc = to_ZX(leading_coeff(y), vT);
+     res->deg0 -= n;
+     res->deg1 -= n;
+     res->off += n;
+  }
+  R = FpXQX_halfres_i(FpXX_shift(x,-n),FpXX_shift(y,-n), T, p, a, b, res);
+  if (res)
+  {
+    res->off -= n;
+    res->deg0 += n;
+    res->deg1 += n;
+  }
+  V1 = FpXQXM_FpXQX_mul2(R, Flxn_red(x,n), Flxn_red(y,n), T, p);
+  x1 = FpXX_add(FpXX_shift(*a,n), gel(V1,1), p);
+  y1 = FpXX_add(FpXX_shift(*b,n), gel(V1,2), p);
+  if (lgpol(y1) <= n)
+  {
+    *a = x1; *b = y1;
+    return gc_all(av, res ? 5: 3, &R, a, b, &res->res, &res->lc);
+  }
+  k = 2*n-degpol(y1);
+  q = FpXQX_divrem(x1, y1, T, p, &r);
+  if (res)
+  {
+    long dx1 = degpol(x1), dy1 = degpol(y1), dr = degpol(r);
+    if (dy1 < degpol(y))
+      FpXQX_halfres_update(res->deg0, res->deg1, dy1, T, p, res);
+    res->lc = to_ZX(leading_coeff(y1), vT);
+    res->deg0 = dx1;
+    res->deg1 = dy1;
+    if (dr >= n)
+    {
+      FpXQX_halfres_update(dx1, dy1, dr, T, p, res);
+      res->deg0 = dy1;
+      res->deg1 = dr;
+    }
+    res->deg0 -= k;
+    res->deg1 -= k;
+    res->off += k;
+  }
+  S = FpXQX_halfres_i(FpXX_shift(y1,-k), FpXX_shift(r,-k), T, p, a, b, res);
+  if (res)
+  {
+    res->deg0 += k;
+    res->deg1 += k;
+    res->off -= k;
+  }
+  Q = FpXQXM_mul2(S,FpXQX_FpXQXM_qmul(q, R, T, p), T, p);
+  V2 = FpXQXM_FpXQX_mul2(S, FpXXn_red(y1,k), FpXXn_red(r,k), T, p);
+  *a = FpXX_add(FpXX_shift(*a,k), gel(V2,1), p);
+  *b = FpXX_add(FpXX_shift(*b,k), gel(V2,2), p);
+  return gc_all(av, res ? 5: 3, &Q, a, b, &res->res, &res->lc);
+}
+
+static GEN
+FpXQX_halfres_i(GEN x, GEN y, GEN T, GEN p, GEN *a, GEN *b, struct FpXQX_res *res)
+{
+  if (lgpol(x) < FpXQX_HALFGCD_LIMIT)
+    return FpXQX_halfres_basecase(x, y, T, p, a, b, res);
+  return FpXQX_halfres_split(x, y, T, p, a, b, res);
+}
+
+static GEN
+FpXQX_halfgcd_all_i(GEN x, GEN y, GEN T, GEN p, GEN *pa, GEN *pb)
+{
+  GEN a, b;
+  GEN R = FpXQX_halfres_i(x, y, T, p, &a, &b, NULL);
+  if (pa) *pa = a;
+  if (pb) *pb = b;
+  return R;
+}
+
+/* Return M in GL_2(Fp[X]/(T)[Y]) such that:
 if [a',b']~=M*[a,b]~ then degpol(a')>= (lgpol(a)>>1) >degpol(b')
 */
 
-static GEN
-FpXQX_halfgcd_i(GEN x, GEN y, GEN T, GEN p)
+GEN
+FpXQX_halfgcd_all(GEN x, GEN y, GEN T, GEN p, GEN *a, GEN *b)
 {
-  if (lg(x)<=FpXQX_HALFGCD_LIMIT) return FpXQX_halfgcd_basecase(x, y, T, p);
-  return FpXQX_halfgcd_split(x, y, T, p);
+  pari_sp av = avma;
+  GEN R,q,r;
+  if (lgefint(p)==3)
+  {
+    ulong pp = to_FlxqX(x, y, T, p, &x, &y, &T);
+    R = FlxXM_to_ZXXM(FlxqX_halfgcd(x, y, T, pp));
+    if (a) *a = Flx_to_ZX(*a);
+    if (b) *b = Flx_to_ZX(*b);
+    return !a && b ? gc_all(av, 2, &R, b): gc_all(av, 1+!!a+!!b, &R, a, b);
+  }
+  if (!signe(x))
+  {
+    if (a) *a = RgX_copy(y);
+    if (b) *b = RgX_copy(x);
+    return matJ2_FpXQXM(varn(x));
+  }
+  if (degpol(y)<degpol(x)) return FpXQX_halfgcd_all_i(x, y, T, p, a, b);
+  q = FpXQX_divrem(y, x, T, p, &r);
+  R = FpXQX_halfgcd_all_i(x, r, T, p, a, b);
+  gcoeff(R,1,1) = FpXX_sub(gcoeff(R,1,1),
+                           FpXQX_mul(q, gcoeff(R,1,2), T, p), p);
+  gcoeff(R,2,1) = FpXX_sub(gcoeff(R,2,1),
+                           FpXQX_mul(q, gcoeff(R,2,2), T, p), p);
+  return !a && b ? gc_all(av, 2, &R, b): gc_all(av, 1+!!a+!!b, &R, a, b);
 }
 
 GEN
 FpXQX_halfgcd(GEN x, GEN y, GEN T, GEN p)
-{
-  pari_sp av = avma;
-  GEN M,q,r;
-  if (lgefint(p)==3)
-  {
-    ulong pp = to_FlxqX(x, y, T, p, &x, &y, &T);
-    M = FlxXM_to_ZXXM(FlxqX_halfgcd(x, y, T, pp));
-  }
-  else
-  {
-    if (!signe(x))
-    {
-      long v = varn(x);
-      retmkmat2(mkcol2(pol_0(v),pol_1(v)),
-                mkcol2(pol_1(v),pol_0(v)));
-    }
-    if (degpol(y)<degpol(x)) return FpXQX_halfgcd_i(x, y, T, p);
-    q = FpXQX_divrem(y, x, T, p, &r);
-    M = FpXQX_halfgcd_i(x, r, T, p);
-    gcoeff(M,1,1) = FpXX_sub(gcoeff(M,1,1), FpXQX_mul(q, gcoeff(M,1,2), T, p), p);
-    gcoeff(M,2,1) = FpXX_sub(gcoeff(M,2,1), FpXQX_mul(q, gcoeff(M,2,2), T, p), p);
-  }
-  return gerepilecopy(av, M);
-}
+{ return FpXQX_halfgcd_all(x, y, T, p, NULL, NULL); }
 
 static GEN
 FpXQX_gcd_basecase(GEN a, GEN b, GEN T, GEN p)
@@ -696,17 +828,19 @@ FpXQX_gcd(GEN x, GEN y, GEN T, GEN p)
   x = FpXQX_red(x, T, p);
   y = FpXQX_red(y, T, p);
   if (!signe(x)) return gerepileupto(av, y);
-  while (lg(y)>FpXQX_GCD_LIMIT)
+  while (lgpol(y)>=FpXQX_GCD_LIMIT)
   {
-    GEN c;
     if (lgpol(y)<=(lgpol(x)>>1))
     {
       GEN r = FpXQX_rem(x, y, T, p);
       x = y; y = r;
     }
-    c = FpXQXM_FpXQX_mul2(FpXQX_halfgcd(x,y, T, p), x, y, T, p);
-    x = gel(c,1); y = gel(c,2);
-    gerepileall(av,2,&x,&y);
+    (void) FpXQX_halfgcd_all(x,y, T, p, &x, &y);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_gcd (y = %ld)",degpol(y));
+      gerepileall(av,2,&x,&y);
+    }
   }
   return gerepileupto(av, FpXQX_gcd_basecase(x, y, T, p));
 }
@@ -738,26 +872,33 @@ FpXQX_extgcd_basecase(GEN a, GEN b, GEN T, GEN p, GEN *ptu, GEN *ptv)
 static GEN
 FpXQX_extgcd_halfgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
 {
-  pari_sp av=avma;
-  GEN u,v,R = matid2_FpXQXM(varn(x));
-  while (lg(y)>FpXQX_EXTGCD_LIMIT)
+  GEN u,v;
+  GEN V = cgetg(expu(lgpol(y))+2,t_VEC);
+  long i, n = 0, vs = varn(x);
+  while (lgpol(y) >= FpXQX_EXTGCD_LIMIT)
   {
-    GEN M, c;
     if (lgpol(y)<=(lgpol(x)>>1))
     {
       GEN r, q = FpXQX_divrem(x, y, T, p, &r);
       x = y; y = r;
-      R = FpXQX_FpXQXM_qmul(q, R, T, p);
-    }
-    M = FpXQX_halfgcd(x,y, T, p);
-    c = FpXQXM_FpXQX_mul2(M, x,y, T, p);
-    R = FpXQXM_mul2(M, R, T, p);
-    x = gel(c,1); y = gel(c,2);
-    gerepileall(av,3,&x,&y,&R);
+      gel(V,++n) = mkmat22(pol_0(vs),pol_1(vs),pol_1(vs),FpXX_neg(q,p));
+    } else
+      gel(V,++n) = FpXQX_halfgcd_all(x, y, T, p, &x, &y);
   }
   y = FpXQX_extgcd_basecase(x,y, T, p, &u,&v);
-  if (ptu) *ptu = FpXQX_addmulmul(u,v,gcoeff(R,1,1),gcoeff(R,2,1), T, p);
-  *ptv = FpXQX_addmulmul(u,v,gcoeff(R,1,2),gcoeff(R,2,2), T, p);
+  for (i = n; i>1; i--)
+  {
+    GEN R = gel(V,i);
+    GEN u1 = FpXQX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), T, p);
+    GEN v1 = FpXQX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), T, p);
+    u = u1; v = v1;
+  }
+  {
+    GEN R = gel(V,1);
+    if (ptu)
+      *ptu = FpXQX_addmulmul(u, v, gcoeff(R,1,1), gcoeff(R,2,1), T, p);
+    *ptv   = FpXQX_addmulmul(u, v, gcoeff(R,1,2), gcoeff(R,2,2), T, p);
+  }
   return y;
 }
 
@@ -781,7 +922,7 @@ FpXQX_extgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
   {
     x = FpXQX_red(x, T, p);
     y = FpXQX_red(y, T, p);
-    if (lg(y)>FpXQX_EXTGCD_LIMIT)
+    if (lgpol(y)>=FpXQX_EXTGCD_LIMIT)
       d = FpXQX_extgcd_halfgcd(x, y, T, p, ptu, ptv);
     else
       d = FpXQX_extgcd_basecase(x, y, T, p, ptu, ptv);
@@ -789,36 +930,35 @@ FpXQX_extgcd(GEN x, GEN y, GEN T, GEN p, GEN *ptu, GEN *ptv)
   return gc_all(av, ptu?3:2, &d, ptv, ptu);
 }
 
-GEN
-FpXQX_dotproduct(GEN x, GEN y, GEN T, GEN p)
+static GEN
+FpXQX_halfres(GEN x, GEN y, GEN T, GEN p, GEN *a, GEN *b, GEN *r)
 {
-  long i, l = minss(lg(x), lg(y));
-  pari_sp av;
-  GEN c;
-  if (l == 2) return gen_0;
-  av = avma; c = gmul(gel(x,2),gel(y,2));
-  for (i=3; i<l; i++) c = gadd(c, gmul(gel(x,i),gel(y,i)));
-  return gerepileupto(av, Fq_red(c,T,p));
+  struct FpXQX_res res;
+  GEN V;
+  long dB, vT = get_FpX_var(T);
+
+  res.res  = *r;
+  res.lc   = to_ZX(leading_coeff(y),vT);
+  res.deg0 = degpol(x);
+  res.deg1 = degpol(y);
+  res.off = 0;
+  V = FpXQX_halfres_i(x, y, T, p, a, b, &res);
+  dB = degpol(*b);
+  if (dB < degpol(y))
+    FpXQX_halfres_update(res.deg0, res.deg1, dB, T, p, &res);
+  *r = res.res;
+  return V;
 }
 
 /* Res(A,B) = Res(B,R) * lc(B)^(a-r) * (-1)^(ab), with R=A%B, a=deg(A) ...*/
-GEN
-FpXQX_resultant(GEN a, GEN b, GEN T, GEN p)
+static GEN
+FpXQX_resultant_basecase(GEN a, GEN b, GEN T, GEN p)
 {
-  long da,db,dc;
-  pari_sp av;
-  long vT = get_FpX_var(T);
+  pari_sp av = avma;
+  long vT = get_FpX_var(T), da,db,dc;
   GEN c,lb, res = pol_1(vT);
 
   if (!signe(a) || !signe(b)) return pol_0(vT);
-  if (lgefint(p) == 3)
-  {
-    pari_sp av = avma;
-    GEN Pl, Ql, Tl, R;
-    ulong pp = to_FlxqX(a, b, T, p, &Pl, &Ql, &Tl);
-    R = FlxqX_resultant(Pl, Ql, Tl, pp);
-    return gerepileupto(av, Flx_to_ZX(R));
-  }
 
   da = degpol(a);
   db = degpol(b);
@@ -828,16 +968,15 @@ FpXQX_resultant(GEN a, GEN b, GEN T, GEN p)
     if (both_odd(da,db)) res = FpX_neg(res, p);
   }
   if (!da) return pol_1(vT); /* = res * a[2] ^ db, since 0 <= db <= da = 0 */
-  av = avma;
   while (db)
   {
-    lb = gel(b,db+2);
+    lb = to_ZX(gel(b,db+2),vT);
     c = FpXQX_rem(a,b, T,p);
     a = b; b = c; dc = degpol(c);
     if (dc < 0) { set_avma(av); return pol_0(vT); }
 
     if (both_odd(da,db)) res = FpX_neg(res, p);
-    if (!equali1(lb)) res = FpXQ_mul(res, FpXQ_powu(lb, da - dc, T, p), T, p);
+    if (!ZX_equal1(lb)) res = FpXQ_mul(res, FpXQ_powu(lb, da - dc, T, p), T, p);
     if (gc_needed(av,2))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_resultant (da = %ld)",da);
@@ -850,6 +989,51 @@ FpXQX_resultant(GEN a, GEN b, GEN T, GEN p)
   return gerepileupto(av, res);
 }
 
+GEN
+FpXQX_resultant(GEN x, GEN y, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  long dx, dy, vT = get_FpX_var(T);
+  GEN res = pol_1(vT);
+  if (!signe(x) || !signe(y)) return pol_0(vT);
+  if (lgefint(p) == 3)
+  {
+    pari_sp av = avma;
+    GEN Pl, Ql, Tl, R;
+    ulong pp = to_FlxqX(x, y, T, p, &Pl, &Ql, &Tl);
+    R = FlxqX_resultant(Pl, Ql, Tl, pp);
+    return gerepileupto(av, Flx_to_ZX(R));
+  }
+
+  dx = degpol(x); dy = degpol(y);
+  if (dx < dy)
+  {
+    swap(x,y);
+    if (both_odd(dx, dy))
+      res = Fp_neg(res, p);
+  }
+  while (lgpol(y) >= FpXQX_GCD_LIMIT)
+  {
+    if (lgpol(y)<=(lgpol(x)>>1))
+    {
+      GEN r = FpXQX_rem(x, y, T, p);
+      long dx = degpol(x), dy = degpol(y), dr = degpol(r);
+      GEN ly = FpX_red(gel(y,dy+2),p);
+      if (!ZX_equal1(ly)) res = FpXQ_mul(res, FpXQ_powu(ly, dx - dr, T, p), T, p);
+      if (both_odd(dx, dy))
+        res = Fp_neg(res, p);
+      x = y; y = r;
+    }
+    (void) FpXQX_halfres(x, y, T, p, &x, &y, &res);
+    if (gc_needed(av,2))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"FpXQX_resultant (y = %ld)",degpol(y));
+      gerepileall(av,3,&x,&y,&res);
+    }
+  }
+  return gerepileupto(av, FpXQ_mul(res, FpXQX_resultant_basecase(x, y, T, p), T, p));
+}
+
 /* disc P = (-1)^(n(n-1)/2) lc(P)^(n - deg P' - 2) Res(P,P'), n = deg P */
 GEN
 FpXQX_disc(GEN P, GEN T, GEN p)
@@ -866,6 +1050,18 @@ FpXQX_disc(GEN P, GEN T, GEN p)
   return gerepileupto(av, D);
 }
 
+GEN
+FpXQX_dotproduct(GEN x, GEN y, GEN T, GEN p)
+{
+  long i, l = minss(lg(x), lg(y));
+  pari_sp av;
+  GEN c;
+  if (l == 2) return gen_0;
+  av = avma; c = gmul(gel(x,2),gel(y,2));
+  for (i=3; i<l; i++) c = gadd(c, gmul(gel(x,i),gel(y,i)));
+  return gerepileupto(av, Fq_red(c,T,p));
+}
+
 /***********************************************************************/
 /**                                                                   **/
 /**                       Barrett reduction                           **/
@@ -1149,6 +1345,17 @@ FpXQX_div_by_X_x(GEN a, GEN x, GEN T, GEN p, GEN *r)
 {
   long l = lg(a), i;
   GEN z;
+  if (lgefint(p)==3)
+  {
+    pari_sp av = avma;
+    GEN ap, xp, t, z;
+    ulong pp = to_FlxqX(a, NULL, T, p, &ap, NULL, &t);
+    xp = ZX_to_Flx(to_ZX(x, get_FpX_var(T)), pp);
+    z = FlxX_to_ZXX(FlxqX_div_by_X_x(ap, xp, t, pp, r));
+    if (!r) return gerepileupto(av, z);
+    *r = Flx_to_ZX(*r);
+    return gc_all(av, 2, &z, r);
+  }
   if (l <= 3)
   {
     if (r) *r = l == 2? gen_0: gcopy(gel(a,2));
@@ -1778,10 +1985,6 @@ FpXQXn_sqr(GEN x, long n, GEN T, GEN p)
   return gerepileupto(av, z);
 }
 
-INLINE GEN
-FpXXn_red(GEN a, long n)
-{ return RgXn_red_shallow(a, n); }
-
 /* (f*g) \/ x^n */
 static GEN
 FpXQX_mulhigh_i(GEN f, GEN g, long n, GEN T, GEN p)
