@@ -159,19 +159,22 @@ RgXV_RgV_eval(GEN Q, GEN x)
 GEN
 RgX_homogenous_evalpow(GEN P, GEN A, GEN B)
 {
-  pari_sp av = avma;
-  long i, d = degpol(P);
+  pari_sp av = avma, btop;
+  long i, d = degpol(P), o;
   GEN s;
   if (signe(P)==0) return pol_0(varn(P));
   s = gel(P, d+2);
   if (d == 0) return gcopy(s);
-  for (i = d-1; i >= 0; i--)
+  o = RgX_deflate_order(P);
+  if (o > 1) A = gpowgs(A, o);
+  btop = avma;
+  for (i = d-o; i >= 0; i-=o)
   {
     s = gadd(gmul(s, A), gmul(gel(B,d+1-i), gel(P,i+2)));
-    if (gc_needed(av,1))
+    if (gc_needed(btop,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgX_homogenous_eval(%ld)",i);
-      s = gerepileupto(av, s);
+      s = gerepileupto(btop, s);
     }
   }
   return gerepileupto(av, s);
@@ -413,7 +416,7 @@ ZX_unscale_divpow(GEN P, GEN h, long k)
 GEN
 RgXV_unscale(GEN x, GEN h)
 {
-  if (!h || isint1(h)) return x;
+  if (isint1(h)) return gcopy(x);
   pari_APPLY_same(RgX_unscale(gel(x,i), h));
 }
 
@@ -433,6 +436,13 @@ RgX_rescale(GEN P, GEN h)
   Q[1] = P[1]; return Q;
 }
 
+GEN
+RgXV_rescale(GEN x, GEN h)
+{
+  if (isint1(h)) return RgX_copy(x);
+  pari_APPLY_same(RgX_rescale(gel(x,i), h));
+}
+
 /* A(X^d) --> A(X) */
 GEN
 RgX_deflate(GEN x0, long d)
@@ -1867,7 +1877,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
   lr=i+3; rem -= lr;
   if (avma==av1) { set_avma((pari_sp)rem); p1 = gcopy(p1); }
   else p1 = gerepileupto((pari_sp)rem,p1);
-  rem[0] = evaltyp(t_POL) | evallg(lr);
+  rem[0] = evaltyp(t_POL) | _evallg(lr);
   rem[1] = z[-1];
   rem += 2;
   gel(rem,i) = p1;
@@ -1963,7 +1973,7 @@ RgXQX_divrem(GEN x, GEN y, GEN T, GEN *pr)
     return gc_const((pari_sp)rem, z-2);
   }
   lr=i+3; rem -= lr;
-  rem[0] = evaltyp(t_POL) | evallg(lr);
+  rem[0] = evaltyp(t_POL) | _evallg(lr);
   rem[1] = z[-1];
   p1 = gerepile((pari_sp)rem,tetpil,p1);
   rem += 2; gel(rem,i) = p1;
@@ -2097,7 +2107,7 @@ RgXQX_pseudorem(GEN x, GEN y, GEN T)
   }
   if (dx < 0) return pol_0(vx);
   lx = dx+3; x -= 2;
-  x[0] = evaltyp(t_POL) | evallg(lx);
+  x[0] = evaltyp(t_POL) | _evallg(lx);
   x[1] = evalsigne(1) | evalvarn(vx);
   x = RgX_recip_i(x);
   if (p)
@@ -2183,7 +2193,7 @@ RgXQX_pseudodivrem(GEN x, GEN y, GEN T, GEN *ptr)
     {
       GEN X = x-2;
       if(DEBUGMEM>1) pari_warn(warnmem,"RgX_pseudodivrem dx=%ld >= %ld",dx,dy);
-      X[0] = evaltyp(t_POL)|evallg(dx+3); X[1] = z[1]; /* hack */
+      X[0] = evaltyp(t_POL)|_evallg(dx+3); X[1] = z[1]; /* hack */
       gerepileall(av2,2, &X, &z); x = X+2;
     }
   }
@@ -2193,7 +2203,7 @@ RgXQX_pseudodivrem(GEN x, GEN y, GEN T, GEN *ptr)
   else
   {
     lx = dx+3; x -= 2;
-    x[0] = evaltyp(t_POL) | evallg(lx);
+    x[0] = evaltyp(t_POL) | _evallg(lx);
     x[1] = evalsigne(1) | evalvarn(vx);
     x = RgX_recip_i(x);
   }
