@@ -69,7 +69,8 @@ idealtyp(GEN *ideal, GEN *arch)
       t = id_MAT;
       break;
 
-    case t_VEC: if (lg(x)!=6) pari_err_TYPE("idealtyp",x);
+    case t_VEC:
+      if (!checkprid_i(x)) pari_err_TYPE("idealtyp [fake prime ideal]",x);
       t = id_PRIME; break;
 
     case t_POL: case t_POLMOD: case t_COL:
@@ -121,16 +122,34 @@ idealhnf_principal(GEN nf, GEN x)
   return cx? ZM_Q_mul(x,cx): x;
 }
 
-/* x integral ideal in t_MAT form, nx columns */
+/* true nf; x integral Z_K-module as t_MAT generated by its columns.
+ * Return square hnf representation */
 static GEN
-vec_mulid(GEN nf, GEN x, long nx, long N)
+vec_mulid(GEN nf, GEN x)
 {
-  GEN m = cgetg(nx*N + 1, t_MAT);
-  long i, j, k;
-  for (i=k=1; i<=nx; i++)
-    for (j=1; j<=N; j++) gel(m, k++) = zk_ei_mul(nf, gel(x,i),j);
-  return m;
+  long i, j, l = lg(x);
+  GEN D = NULL, v;
+  v = cgetg(l, t_VEC);
+  for (i = j = 1; i < l; i++)
+  {
+    GEN m, d;
+    if (D && ZV_Z_dvd(gel(x,i), D)) l--;
+    gel(v,j++) = m = zk_multable(nf, gel(x,i));
+    d = zkmultable_capZ(m);
+    D = D? gcdii(D, d): d;
+    if (is_pm1(D)) return matid(lg(m)-1);
+  }
+  setlg(v, l); if (l == 1) return cgetg(1, t_MAT);
+  return ZM_hnfmodid(shallowconcat1(v), D);
 }
+
+static GEN
+nfV_idealhnf(GEN nf, GEN v, GEN *pden)
+{
+  GEN H = ZM_hnf(Q_remove_denom(matalgtobasis(nf, v), pden));
+  return vec_mulid(nf, H);
+}
+
 /* true nf */
 GEN
 idealhnf_shallow(GEN nf, GEN x)
@@ -139,7 +158,11 @@ idealhnf_shallow(GEN nf, GEN x)
 
   /* cannot use idealtyp because here we allow nonsquare matrices */
   if (tx == t_VEC && lx == 3) { x = gel(x,1); tx = typ(x); lx = lg(x); }
-  if (tx == t_VEC && lx == 6) return pr_hnf(nf,x); /* PRIME */
+  if (tx == t_VEC && lx == 6)
+  {
+    if (!checkprid_i(x)) pari_err_TYPE("idealhnf [fake prime ideal]",x);
+    return pr_hnf(nf,x); /* PRIME */
+  }
   switch(tx)
   {
     case t_MAT:
@@ -153,8 +176,10 @@ idealhnf_shallow(GEN nf, GEN x)
 
       if (nx == N && RgM_is_ZM(x) && ZM_ishnf(x)) return x;
       x = Q_primitive_part(x, &cx);
-      if (nx < N) x = vec_mulid(nf, x, nx, N);
-      x = ZM_hnfmod(x, ZM_detmult(x));
+      if (nx < N)
+        x = vec_mulid(nf, x); /* build ZK-module generated from cols */
+      else
+        x = ZM_hnfmod(x, ZM_detmult(x)); /* assume Z-span cols is ZK-module */
       return cx? ZM_Q_mul(x,cx): x;
     }
     case t_QFB:
@@ -189,6 +214,43 @@ idealhnf(GEN nf, GEN x)
   return (avma == av)? gcopy(y): gerepileupto(av, y);
 }
 
+static GEN
+nfV_eltembed(GEN nf, GEN x, long prec)
+{ pari_APPLY_type(t_VEC, nfeltembed(nf, gel(x,i), NULL, prec)) }
+
+/* true nf */
+static GEN
+nfweilheight_i(GEN nf, GEN v, long prec)
+{
+  long i, j, r1, r2, u, N, l = lg(v);
+  GEN den, h = gen_1, id = nfV_idealhnf(nf, v, &den);
+  GEN V = nfV_eltembed(nf, v, prec);
+
+  nf_get_sign(nf, &r1, &r2); u = r1 + r2; N = u + r2;
+  for (i = 1; i <= r1; i++)
+    for (j = 1; j < l; j++) gmael(V,j,i) = gabs(gmael(V,j,i), prec);
+  for (     ; i <= u; i++)
+    for (j = 1; j < l; j++) gmael(V,j,i) = gnorm(gmael(V,j,i));
+  for (i = 1; i <= u; i++)
+  {
+    long j0 = 1;
+    for (j = 2; j < l; j++)
+      if (gcmp(gmael(V,j,i), gmael(V,j0,i)) > 0) j0 = j;
+    h = gmul(h, gmael(V,j0,i));
+  }
+  if (den) h = gmul(h, powiu(den, N));
+  return divru(glog(gdiv(h, idealnorm(nf, id)), prec), N);
+}
+
+GEN
+nfweilheight(GEN nf, GEN v, long prec)
+{
+  pari_sp av = avma;
+  nf = checknf(nf);
+  if (!is_vec_t(typ(v)) || lg(v) < 2) pari_err_TYPE("nfweilheight",v);
+  return gerepileupto(av, nfweilheight_i(nf, v, prec));
+}
+
 /* GP functions */
 
 GEN
@@ -1148,6 +1210,20 @@ GEN
 famat_div_shallow(GEN f, GEN g)
 { return famat_mul_shallow(f, famat_inv_shallow(g)); }
 
+GEN
+Z_to_famat(GEN x)
+{
+  long k;
+  if (equali1(x)) return trivial_fact();
+  k = Z_isanypower(x, &x) ;
+  return to_famat_shallow(x, k? utoi(k): gen_1);
+}
+GEN
+Q_to_famat(GEN x)
+{
+  if (typ(x) == t_INT) return Z_to_famat(x);
+  return famat_div(Z_to_famat(gel(x,1)), Z_to_famat(gel(x,2)));
+}
 GEN
 to_famat(GEN x, GEN y) { retmkmat2(mkcolcopy(x), mkcolcopy(y)); }
 GEN
@@ -1690,28 +1766,35 @@ famat_to_nf_moddivisor(GEN nf, GEN g, GEN e, GEN bid)
 GEN
 vecmul(GEN x, GEN y)
 {
-  if (is_scalar_t(typ(x))) return gmul(x,y);
+  if (!is_vec_t(typ(x))) return gmul(x,y);
   pari_APPLY_same(vecmul(gel(x,i), gel(y,i)))
 }
 
+GEN
+vecsqr(GEN x)
+{
+  if (!is_vec_t(typ(x))) return gsqr(x);
+  pari_APPLY_same(vecsqr(gel(x,i)))
+}
+
 GEN
 vecinv(GEN x)
 {
-  if (is_scalar_t(typ(x))) return ginv(x);
+  if (!is_vec_t(typ(x))) return ginv(x);
   pari_APPLY_same(vecinv(gel(x,i)))
 }
 
 GEN
 vecpow(GEN x, GEN n)
 {
-  if (is_scalar_t(typ(x))) return powgi(x,n);
+  if (!is_vec_t(typ(x))) return powgi(x,n);
   pari_APPLY_same(vecpow(gel(x,i), n))
 }
 
 GEN
 vecdiv(GEN x, GEN y)
 {
-  if (is_scalar_t(typ(x))) return gdiv(x,y);
+  if (!is_vec_t(typ(x))) return gdiv(x,y);
   pari_APPLY_same(vecdiv(gel(x,i), gel(y,i)))
 }
 
@@ -1989,6 +2072,7 @@ idealismaximal_i(GEN nf, GEN x)
   x = Q_primitive_part(x, &c);
   if (c) return NULL;
   p = zkmultable_capZ(zk_multable(nf, x));
+  if (!BPSW_psp(p)) return NULL;
   L = idealprimedec(nf, p); l = lg(L); pr = NULL;
   for (i = 1; i < l; i++)
   {
@@ -2559,7 +2643,7 @@ GEN
 idealred0(GEN nf, GEN I, GEN vdir)
 {
   pari_sp av = avma;
-  GEN G, aI, IZ, J, y, my, yi, c1 = NULL;
+  GEN G, aI, IZ, J, y, my, dyi, yi, c1 = NULL;
   long N;
 
   nf = checknf(nf);
@@ -2596,7 +2680,7 @@ idealred0(GEN nf, GEN I, GEN vdir)
   }
   /* now I integral, HNF; and J = (I\cap Z) I^(-1), integral */
   y = idealpseudomin(J, G); /* small elt in (I\cap Z)I^(-1), integral */
-  if (equalii(ZV_content(y), IZ))
+  if (ZV_isscalar(y))
   { /* already reduced */
     if (!aI) return gerepilecopy(av, I);
     goto END;
@@ -2607,10 +2691,21 @@ idealred0(GEN nf, GEN I, GEN vdir)
   c1 = mul_content(c1, IZ);
   if (equali1(c1)) c1 = NULL; /* can be simplified with IZ */
   yi = ZM_gauss(my, col_ei(N,1)); /* y^-1 */
-  I = hnfmodid(I, Q_denom(yi)); /* denom(yi) generates (y) \cap Z */
+  dyi = Q_denom(yi); /* generates (y) \cap Z */
+  I = hnfmodid(I, dyi);
   if (!aI) return gerepileupto(av, I);
-  if (typ(aI) == t_MAT) /* yi is not integral and usually larger than y */
-    aI = famat_div(aI, y);
+  if (typ(aI) == t_MAT)
+  {
+    GEN nyi = Q_muli_to_int(yi, dyi);
+    if (gexpo(nyi) >= gexpo(y))
+      aI = famat_div(aI, y); /* yi "larger" than y, keep the latter */
+    else
+    { /* use yi */
+      aI = famat_mul(aI, nyi);
+      c1 = div_content(c1, dyi);
+    }
+    if (c1) { aI = famat_mul(aI, Q_to_famat(c1)); c1 = NULL; }
+  }
   else
     c1 = c1? RgC_Rg_mul(yi, c1): yi;
 END:
@@ -2763,11 +2858,8 @@ static GEN
 Z_cba_rec(GEN L, GEN a, GEN b)
 {
   GEN g;
-  if (lg(L) > 10)
-  { /* a few naive steps before switching to dcba */
-    Z_dcba_rec(L, a, b);
-    return gel(L, lg(L)-1);
-  }
+  /* a few naive steps before switching to dcba */
+  if (lg(L) > 10) { Z_dcba_rec(L, a, b); return veclast(L); }
   if (is_pm1(a)) return b;
   g = gcdii(a,b);
   if (is_pm1(g)) { vectrunc_append(L, a); return b; }
