@@ -86,7 +86,7 @@ bernset(GEN *y, long m, long n)
   t = cgetg(r+1, t_VEC); b = int2n(p); /* fixed point */
   for (j = 3; j <= r; j += 2)
   {
-    GEN z = cgeti(prec);
+    GEN z = cgeti(nbits2lg(p));
     pari_sp av2 = avma;
     affii(divii(b, powuu(j, N)), z);
     gel(t,j) = z; set_avma(av2);
@@ -184,7 +184,7 @@ mpbern(long n, long prec) { (void)prec; constbern(n); }
 static GEN
 bernreal_using_zeta(long n, long prec)
 {
-  GEN pi2 = Pi2n(1, prec+EXTRAPRECWORD);
+  GEN pi2 = Pi2n(1, prec+EXTRAPREC64);
   GEN iz = inv_szeta_euler(n, prec);
   GEN z = divrr(mpfactr(n, prec), mulrr(powru(pi2, n), iz));
   shiftr_inplace(z, 1); /* 2 * n! * zeta(n) / (2Pi)^n */
@@ -250,6 +250,18 @@ bernpol(long k, long v)
   if (k < 0) pari_err_DOMAIN("bernpol", "index", "<", gen_0, stoi(k));
   return gerepileupto(av, bernpol_i(k, v));
 }
+GEN
+bernpol_eval(long k, GEN x)
+{
+  pari_sp av = avma;
+  GEN B;
+  if (!x) return bernpol(k, 0);
+  if (gequalX(x)) return bernpol(k, varn(x));
+  if (k < 0) pari_err_DOMAIN("bernpol", "index", "<", gen_0, stoi(k));
+  B = bernpol_i(k, fetch_var_higher());
+  return gerepileupto(av, poleval(B, x));
+}
+
 /* x := pol_x(v); return 1^e + ... + x^e = x^e + (B_{e+1}(x) - B_{e+1})/(e+1) */
 static GEN
 faulhaber(long e, long v)
@@ -473,7 +485,7 @@ eulerset(GEN *y, long m, long n)
   t = cgetg(r+1, t_VEC); b = int2n(p); /* fixed point */
   for (j = 3; j <= r; j += 2)
   {
-    GEN z = cgeti(prec);
+    GEN z = cgeti(nbits2lg(p));
     pari_sp av2 = avma;
     affii(divii(b, powuu(j, N+1)), z);
     gel(t,j) = z; set_avma(av2);
@@ -579,7 +591,7 @@ inv_lfun4(long n, long prec)
 static GEN
 eulerreal_using_lfun4(long n, long prec)
 {
-  GEN pisur2 = Pi2n(-1, prec+EXTRAPRECWORD);
+  GEN pisur2 = Pi2n(-1, prec+EXTRAPREC64);
   GEN iz = inv_lfun4(n+1, prec);
   GEN z = divrr(mpfactr(n, prec), mulrr(powru(pisur2, n+1), iz));
   if ((n & 3L) == 2) setsigne(z, -1);
