@@ -162,7 +162,7 @@ direuler_factor(GEN s, long n)
       break;
     }
     case t_SER:
-      if (!signe(s) || valp(s) || !gequal1(gel(s,2))) err_direuler(s);
+      if (!signe(s) || valser(s) || !gequal1(gel(s,2))) err_direuler(s);
       break;
     default: pari_err_TYPE("direuler", s);
   }
@@ -325,7 +325,7 @@ vecpowuu(long N, ulong B)
   return v;
 }
 
-/* does n^s require log(x) ? */
+/* does x^s require log(x) ? */
 static long
 get_needlog(GEN s)
 {
@@ -400,6 +400,23 @@ dirpowers(long n, GEN x, long prec)
   return gerepilecopy(av, v);
 }
 
+/* f is a totally multiplicative function of modulus 0 or 1
+ * (essentially a Dirichlet character). Compute simultaneously
+ * sum_{0 < n <= N} f(n)n^s and sum_{0 < n <= N} f(n)n^{-1-conj(s)}
+ * Warning: s is conjugated, but not f. Main application for Riemann-Siegel,
+ * where we need R(chi,s) and conj(R(chi,1-conj(s))). */
+
+static GEN
+vecmulsqlv(GEN Q, GEN V)
+{
+  long l, i;
+  GEN W;
+  if (typ(V) != t_VEC) return RgV_Rg_mul(Q, V);
+  l = lg(Q); W = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++) gel(W, i) = vecmul(gel(Q, i), V);
+  return W;
+}
+
 /* P = prime divisors of (squarefree) n, V[i] = i^s for i <= sq.
  * Return NULL if n is not sq-smooth, else f(n)n^s */
 static GEN
@@ -412,57 +429,120 @@ smallfact(ulong n, GEN P, ulong sq, GEN V)
   l = lg(P); m = p = uel(P, l-1); if (p > sq) return NULL;
   for (i = l-2; i > 1; i--, m = o) { p = uel(P,i); o = m*p; if (o > sq) break; }
   c = gel(V,m); n /= m; /* m <= sq, o = m * p > sq */
-  if (n > sq) { c = gmul(c, gel(V,p)); n /= p; }
-  return gmul(c, gel(V,n));
+  if (n > sq) { c = vecmul(c, gel(V,p)); n /= p; }
+  return vecmul(c, gel(V,n));
 }
+
+static GEN
+_Qtor(GEN x, long prec)
+{ return typ(x) == t_FRAC? fractor(x, prec): x; }
 static GEN
 Qtor(GEN x, long prec)
-{ return typ(x) == t_FRAC? fractor(x, prec): x; }
-/* sum_{n <= N} f(n)n^s. */
-GEN
-dirpowerssumfun(ulong N, GEN s, void *E, GEN (*f)(void *, ulong, long),
-                long prec)
 {
-  const ulong step = 2048;
-  pari_sp av = avma, av2;
-  GEN P, V, W, Q, c2, Q2, Q3, Q6, S, Z, logp;
-  forprime_t T;
-  long gp[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3),0};
-  ulong a, b, c, e, q, x1, n, sq, p, precp;
-  long prec0, prec1, needlog;
+  long tx = typ(x);
+  if (tx == t_VEC || tx == t_COL) pari_APPLY_same(_Qtor(gel(x, i), prec));
+  return tx == t_FRAC? fractor(x, prec): x;
+}
 
-  if (!N) return gen_0;
+static GEN
+vecf(long N, void *E, GEN (*f)(void *, ulong, long), long prec)
+{
+  GEN v;
+  long n;
+  if (!f) return NULL;
+  v = cgetg(N + 1, t_VEC);
+  for (n = 1; n <= N; n++) gel(v,n) = f(E, n, prec);
+  return v;
+}
+
+/* Here #V = #F > 0 is small. Analogous to dotproduct, with following
+ * semantic differences: uses V[1] = 1; V has scalar values but F may have
+ * vector values */
+static GEN
+naivedirpowerssum(GEN V, GEN F, long prec)
+{
+  GEN S;
+  if (!F) S = RgV_sum(V);
+  else
+  {
+    long n, l = lg(V);
+    S = gel(F,1); /* V[1] = 1 */
+    for (n = 2; n < l; n++) S = gadd(S, gmul(gel(V, n), gel(F, n)));
+  }
+  return Qtor(S, prec);
+}
+
+static GEN
+smalldirpowerssum(long N, GEN s, void *E, GEN (*f)(void *, ulong, long),
+                  long both, long prec)
+{
+  GEN F, V, S, SB, sb;
+  if (!N)
+  {
+    if (!f) return both? mkvec2(gen_0, gen_0): gen_0;
+    return gmul(gen_0, f(E, 1, prec));
+  }
+  V = vecpowug(N, s, prec);
+  F = vecf(N, E, f, prec);
+  S = naivedirpowerssum(V, F, prec);
+  if (!both) return S;
+
+  sb = conj_i(gsubsg(-1, s));
+  if ((both==2 || !f) && gequal(s,sb))
+    SB = S;
+  else
+  {
+    GEN VB = cgetg(N+1, t_VEC), FB = (both == 1 && F)? conj_i(F): F;
+    long n;
+    gel(VB, 1) = gel(V, 1); /* = 1 */
+    for (n = 2; n <= N; n++) gel(VB, n) = ginv(gmulsg(n, gel(V, n)));
+    SB = naivedirpowerssum(VB, FB, prec);
+  }
+  return mkvec2(S, SB);
+}
+
+static void
+v2unpack(GEN v, GEN *pV, GEN *pVB)
+{
+  if (typ(v) == t_COL) { *pV = gel(v,1); *pVB = gel(v,2); }
+  else { *pV = v; *pVB = NULL; }
+}
+static GEN
+v2pack(GEN V, GEN VB) { return VB? mkcol2(V,VB): V; }
+
+static GEN
+dirpowsuminit(GEN s, GEN onef, GEN zerf, void *E, GEN (*f)(void *, ulong, long),              GEN data, long both)
+{
+  long needlog = data[1], prec0 = data[2], prec1 = data[3];
+  ulong a, b, c, e, q, n, sq = usqrt(data[4]);
+  GEN V = cgetg(sq+1, t_VEC), W = cgetg(sq+1, t_VEC), VB = NULL, WB = NULL;
+  GEN Q = cgetg(sq+1, t_VEC), Z = cgetg(sq+1, t_VEC), QB = NULL, ZB = NULL;
+  GEN logp, c2, Q2, Q3, Q6, c2B = NULL, Q2B = NULL, Q3B = NULL, Q6B = NULL;
+  long gp[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3),0};
+
+  if (both == 1 || (both == 2 && !gequal(real_i(s), gneg(ghalf))))
+  { VB = cgetg(sq+1, t_VEC); WB = cgetg(sq+1, t_VEC); QB = cgetg(sq+1, t_VEC);}
+  gel(V, 1) = gel(W, 1) = gel(Q, 1) = onef;
+  if (VB) { gel(VB, 1) = gel(WB, 1) = gel(QB, 1) = onef; }
+  c2 = gpow(gen_2, s, prec0); if (VB) c2B = ginv(gmul2n(conj_i(c2), 1));
   if (f)
   {
-    if (N < 49)
-    {
-      V = vecpowug(N, s, prec);
-      S = gen_1;
-      for (n = 2; n <= N; n++)
-        S = gadd(S, gmul(gel(V,n), f(E, n, prec)));
-      return gerepileupto(av, Qtor(S, prec));
-    }
+    GEN tmp2 = f(E, 2, prec0);
+    c2 = gmul(c2, tmp2); if (VB) c2B = gmul(c2B, tmp2);
   }
-  else if (N < 1000UL)
-    return gerepileupto(av, Qtor(RgV_sum(vecpowug(N, s, prec)), prec));
-  sq = usqrt(N);
-  V = cgetg(sq+1, t_VEC);
-  W = cgetg(sq+1, t_VEC);
-  Q = cgetg(sq+1, t_VEC);
-  prec1 = prec0 = prec + EXTRAPRECWORD;
-  s = gprec_w(s, prec0);
-  needlog = get_needlog(s);
-  if (needlog == 1) prec1 = powcx_prec(log2((double)N), s, prec);
-  gel(V,1) = gel(W,1) = gel(Q,1) = gen_1;
-  c2 = gpow(gen_2, s, prec0);
-  if (f) c2 = gmul(c2, f(E, 2, prec));
   gel(V,2) = c2; /* f(2) 2^s */
-  gel(W,2) = Qtor(gaddgs(c2, 1), prec0);
-  gel(Q,2) = Qtor(gaddgs(gsqr(c2), 1), prec0);
+  gel(W,2) = Qtor(gadd(c2, onef), prec0);
+  gel(Q,2) = Qtor(gadd(vecsqr(c2), onef), prec0);
+  if (VB)
+  {
+    gel(VB, 2) = c2B; gel(WB, 2) = Qtor(gadd(c2B, onef), prec0);
+    gel(QB, 2) = Qtor(gadd(vecsqr(c2B), onef), prec0);
+  }
   logp = NULL;
   for (n = 3; n <= sq; n++)
   {
-    GEN u;
+    GEN u = NULL, uB = NULL, ks = f ? f(E, n, prec0) : gen_1;
+    if (gequal0(ks)) ks = NULL;
     if (odd(n))
     {
       gp[2] = n;
@@ -475,120 +555,353 @@ dirpowerssumfun(ulong N, GEN s, void *E, GEN (*f)(void *, ulong, long),
           GEN z = atanhuu(1, n - 1, prec1);
           shiftr_inplace(z, 1); logp = addrr(logp, z);
         }
-        u = needlog == 1? powcx(gp, logp, s, prec0)
-                        : mpexp(gmul(s, logp));
-
+        if (ks)
+          u = needlog == 1? powcx(gp, logp, s, prec0) : mpexp(gmul(s, logp));
+      }
+      else if (ks) u = gpow(gp, s, prec0);
+      if (ks)
+      {
+        if (VB) uB = gmul(ginv(gmulsg(n, conj_i(u))), ks);
+        u = gmul(u, ks); /* f(n) n^s */
       }
-      else
-        u = gpow(gp, s, prec0);
-      if (f) u = gmul(u, f(E, n, prec0)); /* f(n) n^s */
     }
     else
-      u = gmul(c2, gel(V, n>> 1));
-    gel(V,n) = u; /* f(n) n^s */
-    gel(W,n) = gadd(gel(W,n-1), gel(V,n));       /* = sum_{i<=n} f(i)i^s */
-    gel(Q,n) = gadd(gel(Q,n-1), gsqr(gel(V,n))); /* = sum_{i<=n} f(i^2)i^2s */
+    {
+      u = vecmul(c2, gel(V, n >> 1));
+      if (VB) uB = vecmul(c2B, gel(VB, n >> 1));
+    }
+    if (ks)
+    {
+      gel(V,n) = u; /* f(n) n^s */
+      gel(W,n) = gadd(gel(W, n-1), gel(V,n));        /*= sum_{i<=n} f(i)i^s */
+      gel(Q,n) = gadd(gel(Q, n-1), vecsqr(gel(V,n)));/*= sum_{i<=n} f(i^2)i^2s*/
+      if (VB)
+      {
+        gel(VB,n) = uB;
+        gel(WB,n) = gadd(gel(WB,n-1), gel(VB,n));
+        gel(QB,n) = gadd(gel(QB,n-1), vecsqr(gel(VB,n)));
+      }
+    }
+    else
+    {
+      gel(V,n) = zerf; gel(W,n) = gel(W, n-1); gel(Q,n) = gel(Q, n-1);
+      if (VB)
+      { gel(VB,n) = zerf; gel(WB,n) = gel(WB, n-1); gel(QB,n) = gel(QB, n-1); }
+    }
   }
-  Q2 = RgV_Rg_mul(Q, gel(V,2));
-  Q3 = RgV_Rg_mul(Q, gel(V,3));
-  Q6 = RgV_Rg_mul(Q, gel(V,6));
-  precp = 0; logp = NULL; S = gen_0;
-  u_forprime_init(&T, sq + 1, N);
-  av2 = avma;
-  while ((p = u_forprime_next(&T)))
+  Q2 = vecmulsqlv(Q, gel(V,2));
+  Q3 = vecmulsqlv(Q, gel(V,3));
+  Q6 = vecmulsqlv(Q, gel(V,6));
+  if (VB)
   {
-    GEN u;
+    Q2B = vecmulsqlv(QB, gel(VB,2));
+    Q3B = vecmulsqlv(QB, gel(VB,3));
+    Q6B = vecmulsqlv(QB, gel(VB,6));
+  }
+  /* a,b,c,e = sqrt(q), sqrt(q/2), sqrt(q/3), sqrt(q/6)
+   * Z[q] = Q[a] + 2^s Q[b] + 3^s Q[c] + 6^s Q[e], with Q[0] = 0 */
+  gel(Z, 1) = onef;
+  gel(Z, 2) = gel(W, 2);
+  gel(Z, 3) = gel(W, 3);
+  gel(Z, 4) = gel(Z, 5) = gel(W, 4);
+  gel(Z, 6) = gel(Z, 7) = gadd(gel(W, 4), gel(V, 6));
+  if (VB)
+  {
+    ZB = cgetg(sq+1, t_VEC);
+    gel(ZB, 1) = onef;
+    gel(ZB, 2) = gel(WB, 2);
+    gel(ZB, 3) = gel(WB, 3);
+    gel(ZB, 4) = gel(ZB, 5) = gel(WB, 4);
+    gel(ZB, 6) = gel(ZB, 7) = gadd(gel(WB, 4), gel(VB, 6));
+  }
+  a = 2; b = c = e = 1;
+  for (q = 8; q <= sq; q++)
+  { /* Gray code: at most one of a,b,c,d differs (by 1) from previous value */
+    GEN z = gel(Z, q - 1), zB = NULL;
+    ulong na, nb, nc, ne, na2, nb2, nc2, ne2;
+    if (VB) zB = gel(ZB, q - 1);
+    if ((na = usqrt(q)) != a)
+    { a = na; na2 = na * na; z = gadd(z, gel(V, na2));
+      if (VB) zB = gadd(zB, gel(VB, na2)); }
+    else if ((nb = usqrt(q / 2)) != b)
+    { b = nb; nb2 = 2 * nb * nb; z = gadd(z, gel(V, nb2));
+      if (VB) zB = gadd(zB, gel(VB, nb2)); }
+    else if ((nc = usqrt(q / 3)) != c)
+    { c = nc; nc2 = 3 * nc * nc; z = gadd(z, gel(V, nc2));
+      if (VB) zB = gadd(zB, gel(VB, nc2)); }
+    else if ((ne = usqrt(q / 6)) != e)
+    { e = ne; ne2 = 6 * ne * ne; z = gadd(z, gel(V, ne2));
+      if (VB) zB = gadd(zB, gel(VB, ne2)); }
+    gel(Z, q) = z; if (VB) gel(ZB, q) = zB;
+  }
+  return v2pack(mkvecn(7, V, W, Q, Q2, Q3, Q6, Z),
+                VB? mkvecn(7, VB, WB, QB, Q2B, Q3B, Q6B, ZB): NULL);
+}
+
+static GEN
+sumprimeloop(forprime_t *pT, GEN s, long N, GEN data, GEN S, GEN W, GEN WB,
+             void *E, GEN (*f)(void *, ulong, long))
+{
+  pari_sp av = avma;
+  long needlog = data[1], prec0 = data[2], prec1 = data[3];
+  long gp[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3),0};
+  ulong p, precp = 0;
+  GEN logp = NULL, SB = WB? S: NULL;
+
+  while ((p = u_forprime_next(pT)))
+  {
+    GEN u = NULL, ks;
+    if (!f) ks = gen_1;
+    else
+    {
+      ks = f(E, p, prec1);
+      if (gequal0(ks)) ks = NULL;
+    }
     gp[2] = p;
     if (needlog)
     {
       if (!logp)
         logp = logr_abs(utor(p, prec1));
       else
-      { /* log p = log(precp) + 2 atanh((p - precp) / (p + precp)) */
+      { /* Assuming p and precp are odd,
+         * log p = log(precp) + 2 atanh((p - precp) / (p + precp)) */
         ulong a = p >> 1, b = precp >> 1; /* p = 2a + 1, precp = 2b + 1 */
         GEN z = atanhuu(a - b, a + b + 1, prec1); /* avoid overflow */
         shiftr_inplace(z, 1); logp = addrr(logp, z);
       }
-      u = needlog == 1? powcx(gp, logp, s, prec0)
-                      : mpexp(gmul(s, logp));
+      if (ks) u = needlog == 1? powcx(gp, logp, s, prec0)
+                              : mpexp(gmul(s, logp));
+    }
+    else if (ks) u = gpow(gp, s, prec0);
+    if (ks)
+    {
+      S = gadd(S, vecmul(gel(W, N / p), ks == gen_1? u: gmul(ks, u)));
+      if (WB)
+      {
+        GEN w = gel(WB, N / p);
+        if (ks != gen_1) w = vecmul(ks, w);
+        SB = gadd(SB, gdiv(w, gmulsg(p, conj_i(u))));
+      }
     }
-    else
-      u = gpow(gp, s, prec0);
-    if (f) u = gmul(u, f(E, p, prec0)); /* f(p) p^s */
-    S = gadd(S, gmul(gel(W, N / p), u));
     precp = p;
     if ((p & 0x1ff) == 1)
     {
-      if (!logp) S = gerepileupto(av2,S);
-      else gerepileall(av2, 2, &S, &logp);
+      if (!logp) gerepileall(av, SB? 2: 1, &S, &SB);
+      else gerepileall(av, SB? 3: 2, &S, &logp, &SB);
     }
   }
-  P = mkvecsmall2(2, 3);
-  Z = cgetg(sq+1, t_VEC);
-  /* a,b,c,e = sqrt(q), sqrt(q/2), sqrt(q/3), sqrt(q/6)
-   * Z[q] = Q[a] + 2^s Q[b] + 3^s Q[c] + 6^s Q[e], with Q[0] = 0 */
-  gel(Z, 1) = gen_1;
-  gel(Z, 2) = gel(W, 2);
-  gel(Z, 3) = gel(W, 3);
-  gel(Z, 4) = gel(Z, 5) = gel(W,4);
-  gel(Z, 6) = gel(Z, 7) = gadd(gel(W,4), gel(V,6));
-  a = 2; b = c = e = 1;
-  for (q = 8; q <= sq; q++)
-  { /* Gray code: at most one of a,b,c,d differs (by 1) from previous value */
-    GEN z = gel(Z, q - 1);
-    ulong na, nb, nc, ne;
-    if ((na = usqrt(q)) != a)
-    { a = na; z = gadd(z, gel(V, na * na)); }
-    else if ((nb = usqrt(q / 2)) != b)
-    { b = nb; z = gadd(z, gel(V, 2 * nb * nb)); }
-    else if ((nc = usqrt(q / 3)) != c)
-    { c = nc; z = gadd(z, gel(V, 3 * nc * nc)); }
-    else if ((ne = usqrt(q / 6)) != e)
-    { e = ne; z = gadd(z, gel(V, 6 * ne * ne)); }
-    gel(Z,q) = z;
-  }
-  av2 = avma;
-  for(x1 = 1;; x1 += step)
-  { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
-    ulong j, lv, x2 = (N >= 2*step && N - 2*step >= x1)? x1-1 + step: N;
-    GEN v = vecfactorsquarefreeu_coprime(x1, x2, P);
-    lv = lg(v);
-    for (j = 1; j < lv; j++) if (gel(v,j))
+  return v2pack(S, SB);
+}
+
+static GEN
+add4(GEN a, GEN b, GEN c, GEN d) { return gadd(gadd(a,b), gadd(c,d)); }
+
+static const long step = 2048;
+static int
+mksqfloop(long N, long x1, GEN R, GEN RB, GEN *pS, GEN *pSB)
+{
+  GEN V = gel(R,1), Q = gel(R,3), Q2 = gel(R,4);
+  GEN Q3 = gel(R,5), Q6 = gel(R,6), Z = gel(R,7);
+  GEN v, VB = NULL, QB = NULL, Q2B = NULL, Q3B = NULL, Q6B = NULL, ZB = NULL;
+  long x2, j, lv, sq = lg(V)-1;
+
+  if (RB)
+  { VB = gel(RB,1); QB = gel(RB,3); Q2B = gel(RB,4);
+    Q3B = gel(RB,5), Q6B = gel(RB,6); ZB = gel(RB,7); }
+  /* beware overflow, fuse last two bins (avoid a tiny remainder) */
+  x2 = (N >= 2*step && N - 2*step >= x1)? x1-1 + step: N;
+  v = vecfactorsquarefreeu_coprime(x1, x2, mkvecsmall2(2, 3));
+  lv = lg(v);
+  for (j = 1; j < lv; j++)
+    if (gel(v,j))
     {
-      ulong d = x1-1 + j; /* squarefree, coprime to 6 */
-      GEN t = smallfact(d, gel(v,j), sq, V), u; /* = d^s */
-      if (!t) continue;
-      /* S += d^s * Z[q] */
+      ulong d = x1 - 1 + j; /* squarefree, coprime to 6 */
+      GEN t = smallfact(d, gel(v,j), sq, V), u;
+      GEN tB = NULL, uB = NULL; /* = f(d) d^s */
+      long a, b, c, e, q;
+      if (!t || gequal0(t)) continue;
+      if (VB) tB = vecinv(gmulsg(d, conj_i(t)));
+      /* warning: gives 1/conj(f(d)) d^(-1-conj(s)), equal to
+         f(d) d^(-1-conj(s)) only if |f(d)|=1. */
+      /* S += f(d) * d^s * Z[q] */
       q = N / d;
-      if (q == 1) { S = gadd(S, t); continue; }
-      if (q <= sq) u = gel(Z, q);
+      if (q == 1)
+      {
+        *pS = gadd(*pS, t); if (VB) *pSB = gadd(*pSB, tB);
+        continue;
+      }
+      if (q <= sq) { u = gel(Z, q); if (VB) uB = gel(ZB, q); }
       else
       {
         a = usqrt(q); b = usqrt(q / 2); c = usqrt(q / 3); e = usqrt(q / 6);
-        u = gadd(gadd(gel(Q,a), gel(Q2,b)), gadd(gel(Q3,c), gel(Q6,e)));
+        u = add4(gel(Q,a), gel(Q2,b), gel(Q3,c), gel(Q6,e));
+        if (VB) uB = add4(gel(QB,a), gel(Q2B,b), gel(Q3B,c), gel(Q6B,e));
       }
-      S = gadd(S, gmul(t, u));
+      *pS = gadd(*pS, vecmul(t, u)); if (VB) *pSB = gadd(*pSB, vecmul(tB, uB));
     }
-    if (x2 == N) break;
-    S = gerepileupto(av2, S);
-  }
-  return gerepileupto(av, S);
+  return x2 == N;
 }
-GEN
-dirpowerssum(ulong N, GEN s, long prec)
-{ return dirpowerssumfun(N, s, NULL, NULL, prec); }
+
+static GEN
+mkdata(long N, GEN s, long prec)
+{
+  long needlog, prec0, prec1, m = mt_nbthreads(), STEP = maxss(N / (m * m), 1);
+  prec1 = prec0 = prec + EXTRAPRECWORD;
+  needlog = get_needlog(s);
+  if (needlog == 1) prec1 = powcx_prec(log2((double)N), s, prec);
+  return mkvecsmalln(5, needlog, prec0, prec1, N, STEP);
+}
+
 static GEN
 gp_callUp(void *E, ulong x, long prec)
 {
-  long court[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3),0};
-  court[2] = x; return gp_callprec(E, court, prec);
+  long n[] = {evaltyp(t_INT)|_evallg(3), evalsigne(1)|evallgefint(3),0};
+  n[2] = x; return gp_callprec(E, n, prec);
+}
+
+/* set *p0 and *p1 to 0 and 1 in the algebra where f takes its values */
+static int
+mk01(void *E, GEN (*f)(void*,ulong,long), long prec, GEN *p0, GEN *p1)
+{
+  *p0 = gen_0; *p1 = gen_1;
+  if (!f) return 1;
+  *p1 = f(E, 1, prec);
+  if (is_vec_t(typ(*p1)))
+  {
+    long l = lg(*p1);
+    if (l == 1) { *p0 = *p1 = NULL; return 0; }
+    *p0 = const_vec(l-1, gen_0);
+  }
+  return 1;
+}
+static GEN
+mktrivial(long both)
+{
+  if (!both) return cgetg(1, t_VEC);
+  return mkvec2(cgetg(1,t_VEC), cgetg(1,t_VEC));
+}
+
+/* both is
+ * 0: sum_{n<=N}f(n)n^s
+ * 1: sum for (f,s) and (conj(f),-1-s)
+ * 2: sum for (f,s) and (f,-1-s), assuming |f(n)| in {0,1} */
+static GEN
+dirpowerssumfun_i(ulong N, GEN s, void *E, GEN (*f)(void *, ulong, long),
+                  long both, long prec)
+{
+  forprime_t T;
+  pari_sp av;
+  GEN onef, zerf, R, RB, W, WB, S, SB, data;
+  ulong x1;
+
+  if ((f && N < 49) || (!f && N < 1000))
+    return smalldirpowerssum(N, s, E, f, both, prec);
+  if (!mk01(E, f, prec, &zerf, &onef)) return mktrivial(both);
+  data = mkdata(N, s, prec); s = gprec_w(s, prec + EXTRAPRECWORD);
+  v2unpack(dirpowsuminit(s, onef, zerf, E, f, data, both), &R, &RB);
+  W = gel(R,2); WB = RB? gel(RB,2): NULL;
+  av = avma; u_forprime_init(&T, lg(W), N);
+  v2unpack(sumprimeloop(&T, s, N, data, zerf, W, WB, E, f), &S, &SB);
+  for(x1 = 1;; x1 += step)
+  {
+    if (mksqfloop(N, x1, R, RB, &S, &SB))
+      return both? mkvec2(S, conj_i(SB? SB: S)): S;
+    gerepileall(av, SB? 2: 1, &S, &SB);
+  }
 }
 GEN
-dirpowerssum0(GEN N, GEN s, GEN f, long prec)
+dirpowerssumfun(ulong N, GEN s, void *E, GEN (*f)(void *, ulong, long),
+                long both, long prec)
+{
+  pari_sp av = avma;
+  return gerepilecopy(av, dirpowerssumfun_i(N, s, E, f, both, prec));
+}
+
+GEN
+dirpowerssum(ulong N, GEN s, long both, long prec)
+{ return dirpowerssumfun(N, s, NULL, NULL, both, prec); }
+
+GEN
+dirpowerssum0(GEN N, GEN s, GEN f, long both, long prec)
 {
   if (typ(N) != t_INT) pari_err_TYPE("dirpowerssum", N);
-  if (signe(N) <= 0) return gen_0;
-  if (!f) return dirpowerssum(itou(N), s, prec);
+  if (signe(N) <= 0) N = gen_0;
+  if (!f) return dirpowerssum(itou(N), s, both, prec);
   if (typ(f) != t_CLOSURE) pari_err_TYPE("dirpowerssum", f);
-  return dirpowerssumfun(itou(N), s, (void*)f, gp_callUp, prec);
+  return dirpowerssumfun(itou(N), s, (void*)f, gp_callUp, both, prec);
 }
+
+GEN
+parsqf_worker(GEN gk, GEN vR, GEN gN)
+{
+  pari_sp av = avma;
+  GEN R, RB, onef, S, SB;
+  long k = itou(gk), N = itou(gN), x1 = 1 + step * k;
+  v2unpack(vR, &R, &RB); onef = gmael(R,1,1);
+  S = SB = is_vec_t(typ(onef)) ? zerovec(lg(onef) - 1): gen_0;
+  (void)mksqfloop(N, x1, R, RB, &S, &SB);
+  return gerepilecopy(av, v2pack(S, RB? SB: NULL));
+}
+
+static GEN
+mycallvec(void *f, ulong n, long prec)
+{
+  GEN F = (GEN)f;
+  if (!f) return gen_1;
+  if (typ(F) == t_CLOSURE) return gp_callUp(f, n, prec);
+  return gel(F, (n-1) % (lg(F)-1) + 1);
+}
+
+GEN
+parsumprimefun_worker(GEN gk, GEN s, GEN zerf, GEN data, GEN vW, GEN f)
+{
+  forprime_t T;
+  GEN W, WB;
+  long k = itou(gk), sq, N = data[4], STEP = data[5];
+
+  v2unpack(vW, &W, &WB); sq = lg(W)-1;
+  if (isintzero(f)) f = NULL;
+  u_forprime_init(&T, k * STEP + sq + 1, minss(N, (k + 1) * STEP + sq));
+  return sumprimeloop(&T, s, N, data, zerf, W, WB, (void*)f, mycallvec);
+}
+
+static GEN
+vR_get_vW(GEN vR)
+{
+  GEN R, RB, W, WB;
+  v2unpack(vR, &R, &RB); W = gel(R,2); WB = RB? gel(RB,2): NULL;
+  return v2pack(W, WB);
+}
+
+static GEN
+pardirpowerssumfun_i(GEN f, ulong N, GEN s, long both, long prec)
+{
+  GEN worker, worker2, data, vR, onef, zerf;
+
+  if ((f && N < 49) || (!f && N < 10000UL))
+    return smalldirpowerssum(N, s, (void*)f, mycallvec, both, prec);
+  if (!mk01((void*)f, mycallvec, prec, &zerf, &onef)) return mktrivial(both);
+  data = mkdata(N, s, prec); s = gprec_w(s, prec + EXTRAPRECWORD);
+  vR = dirpowsuminit(s, onef, zerf, (void*)f, mycallvec, data, both);
+  worker = snm_closure(is_entry("_parsumprimefun_worker"),
+                       mkvecn(5, s, zerf, data, vR_get_vW(vR), f? f: gen_0));
+  worker2 = snm_closure(is_entry("_parsqf_worker"), mkvec2(vR, utoi(N)));
+  return gadd(parsum(gen_0, utoipos((N-1) / data[5]), worker),
+              parsum(gen_0, utoipos(maxss((N-1) / step - 1, 0)), worker2));
+}
+
+GEN
+pardirpowerssumfun(GEN f, ulong N, GEN s, long both, long prec)
+{
+  pari_sp av = avma;
+  return gerepilecopy(av, pardirpowerssumfun_i(f, N, s, both, prec));
+}
+GEN
+pardirpowerssum0(GEN N, GEN s, GEN f, long both, long prec)
+{
+  if (typ(N) != t_INT) pari_err_TYPE("pardirpowerssum", N);
+  return pardirpowerssumfun(f, itou(N), s, both, prec);
+}
+GEN
+pardirpowerssum(ulong N, GEN s, long prec)
+{ return pardirpowerssumfun(NULL, N, s, 0, prec); }
