@@ -888,6 +888,7 @@ matcassels(GEN F, GEN M)
       { F2m_set(C,i,j); F2m_set(C,j,i); }
     }
   }
+  if (DEBUGLEVEL>=2) err_printf("Cassels Pairing: %Ps\n", F2m_to_ZM(C));
   return gc_const(av, C);
 }
 
@@ -1270,9 +1271,10 @@ nf2selmer_quad(GEN nf, GEN S)
 {
   pari_sp ltop = avma;
   GEN D = nf_get_disc(nf), factD = nf_get_ramified_primes(nf);
-  GEN SlistQ = prV_primes(S), QS2gen, gen, Hlist, H, KerH, norms, LS2gen;
+  GEN SlistQ = prV_primes(S), QS2gen, gen, Hlist, H, KerH, LS2gen;
   GEN chpol, Q, kerval, S2, G, e, f, b, c, bad;
   long lS = lg(S), l, lHlist, i, j, k;
+  GEN nfa = nf_get_ramified_primes(nf);
 
   QS2gen = vec_prepend(SlistQ, gen_m1);
   bad = ZV_sort_uniq_shallow(shallowconcat(factD, SlistQ));
@@ -1288,28 +1290,26 @@ nf2selmer_quad(GEN nf, GEN S)
     gel(H, j) = Flv_to_F2v(v);
   }
   KerH = F2m_ker(H); l = lg(KerH);
-  norms = cgetg(l, t_VEC);
-  for (i = 1; i < l; i++)
-    gel(norms, i) = factorback2(QS2gen, F2c_to_ZC(gel(KerH, i)));
   LS2gen = cgetg(l, t_VEC);
   chpol = QXQ_charpoly(gel(nf_get_zk(nf), 2), nf_get_pol(nf), 0);
-  b = gdivgu(negi(gel(chpol, 3)), 2);
-  c = gel(chpol, 2);
-  Q = mkmat3(mkcol3(gen_1, b, gen_0),
+  b = negi(gel(chpol, 3));
+  c = shifti(gel(chpol, 2),1);
+  Q = mkmat3(mkcol3(gen_2, b, gen_0),
              mkcol3(b, c, gen_0),
              mkcol3(gen_0, gen_0, gen_0));
   for (k = 1; k < l; k++)
   {
-    GEN sol;
-    gcoeff(Q, 3, 3) = negi(gel(norms, k));
-    sol = qfsolve(Q); /* must be solvable */
+    GEN P = RgV_F2v_extract_shallow(QS2gen, gel(KerH, k));
+    GEN F = ZV_sort_uniq_shallow(shallowconcat(nfa, P)), sol;
+    gcoeff(Q, 3, 3) = mulis(ZV_prod(P), -2);
+    sol = qfsolve(mkvec2(Q, F)); /* must be solvable */
     sol = Q_primpart(mkcol2(gel(sol,1), gel(sol,2)));
     gel(LS2gen, k) = basistoalg(nf, sol);
   }
   if (equalis(D, -4)) G = bad;
   else
   {
-    G = vecsplice(bad, ZV_search(bad, gel(factD, lg(factD)-1)));
+    G = vecsplice(bad, ZV_search(bad, veclast(factD)));
     G = vec_prepend(G, gen_m1);
   }
   LS2gen = shallowconcat(G, LS2gen);
@@ -1494,19 +1494,25 @@ RgXV_cxeval(GEN x, GEN r, GEN ri)
 { pari_APPLY_same(RgX_cxeval(gel(x,i), r, ri)) }
 
 static GEN
-redquadric(GEN base, GEN q2, GEN pol, GEN zc)
+redquadric(GEN base, GEN pol, GEN zc)
 {
-  long i, l, prec = nbits2prec(2*gexpo(q2)) + 1;
-  GEN s = NULL, R = roots(pol, prec);
-  l = lg(R);
-  for (i = 1; i < l; ++i)
+  pari_sp av = avma;
+  long prec = nbits2prec(gexpo(pol)+gexpo(zc)) + EXTRAPRECWORD;
+  for (;;)
   {
-    GEN r = gel(R,i), ri = gexpo(r) > 1? ginv(r): NULL;
-    GEN b = RgXV_cxeval(base, r, ri), z = RgX_cxeval(zc, r, ri);
-    GEN M = RgC_RgV_mulrealsym(RgV_Rg_mul(b, gabs(z, prec)), gconj(b));
-    s = s? RgM_add(s, M): M;
+    GEN R = roots(pol, prec), s = NULL;
+    long i, l = lg(R);
+    for (i = 1; i < l; ++i)
+    {
+      GEN r = gel(R,i), ri = gexpo(r) > 1? ginv(r): NULL;
+      GEN b = RgXV_cxeval(base, r, ri), z = RgX_cxeval(zc, r, ri);
+      GEN M = RgC_RgV_mulrealsym(RgV_Rg_mul(b, gabs(z, prec)), gconj(b));
+      s = s? RgM_add(s, M): M;
+    }
+    s = RgM_Cholesky(s, prec);
+    if (s) return gerepileupto(av, lll(s));
+    prec = precdbl(prec); set_avma(av);
   }
-  return lllgram(s);
 }
 
 static GEN
@@ -1594,6 +1600,16 @@ liftselmerinit(GEN expo, GEN vnf, GEN sqrtLS2, GEN factLS2,
   return shallowconcat1(b);
 }
 
+static GEN
+qf_disc_fact(GEN q, GEN L)
+{
+  GEN D = ZM_det(q), P, E;
+  GEN H = Z_smoothen(D, L, &P, &E);
+  if (H)
+    P = shallowconcat(P, gel(Z_factor(H),1));
+  return mkvec2(q, P);
+}
+
 static GEN
 liftselmer_cover(GEN b, GEN expo, GEN LS2, GEN pol, GEN discF, GEN K)
 {
@@ -1606,10 +1622,10 @@ liftselmer_cover(GEN b, GEN expo, GEN LS2, GEN pol, GEN discF, GEN K)
   tttheta = RgX_shift_shallow(pol, -1);
   polprime = ZX_deriv(pol);
   q2 = Q_primpart(tracematrix(zc, b, pol));
-  U = redquadric(b, q2, pol, QXQ_div(zc, polprime, pol));
-  q2 = qf_apply_RgM(q2, U);
+  U = redquadric(b, pol, QXQ_div(zc, polprime, pol));
+  q2 = qf_RgM_apply(q2, U);
   newb = RgV_RgM_mul(b, U);
-  param = Q_primpart(qfparam(q2, qfsolve(q2), 1));
+  param = Q_primpart(qfparam(q2, qfsolve(qf_disc_fact(q2,gel(discF,2))), 1));
   param = RgM_to_RgXV_reverse(shallowtrans(param), 0);
   q1 = RgM_neg(tracematrix(QXQ_mul(zc, ttheta, pol), newb, pol));
   q1 = Q_remove_denom(qfeval(q1, param), &den);
@@ -1660,12 +1676,11 @@ liftselmer(GEN b, GEN expo, GEN sbase, GEN LS2, GEN pol, GEN discF, GEN K, long
       zc = QXQ_mul(z, QXQ_sqr(r, pol), pol);
     }
     q2 = Q_primpart(tracematrix(zc, b, pol));
-    U = redquadric(b, q2, pol, QXQ_div(zc, polprime, pol));
+    U = redquadric(b, pol, QXQ_div(zc, polprime, pol));
     if (lg(U) < 4) continue;
-    q2 = qf_apply_RgM(q2, U);
+    q2 = qf_RgM_apply(q2, U);
     newb = RgV_RgM_mul(b, U);
-
-    param = Q_primpart(qfparam(q2, qfsolve(q2), 1));
+    param = Q_primpart(qfparam(q2, qfsolve(qf_disc_fact(q2,gel(discF,2))), 1));
     param = RgM_to_RgXV_reverse(shallowtrans(param), 0);
     q1 = RgM_neg(tracematrix(QXQ_mul(zc, ttheta, pol), newb, pol));
     q1 = Q_remove_denom(qfeval(q1, param), &den);
@@ -1779,6 +1794,48 @@ get_row(GEN vnf, GEN K)
   return sK > 0? vecindexmin(R): vecindexmax(R);
 }
 
+static GEN
+Z_factor_addprimes(GEN N, GEN P)
+{
+  GEN Pnew, Enew;
+  GEN H = Z_smoothen(N, P, &Pnew, &Enew);
+  return H ? shallowconcat(P, gel(Z_factor(H),1)): P;
+}
+
+static GEN
+vbnf_discfactors(GEN vbnf)
+{
+  long n = lg(vbnf)-1;
+  switch (n)
+  {
+    case 1:
+    {
+      GEN nf = bnf_get_nf(gel(vbnf,1));
+      GEN L = shallowtrans(nf_get_ramified_primes(nf));
+      return Z_factor_addprimes(nf_get_index(nf), L);
+    }
+    case 2:
+    {
+      GEN nf = gel(vbnf,2), R;
+      GEN L = shallowtrans(nf_get_ramified_primes(nf));
+      L = Z_factor_addprimes(nf_get_index(nf), L);
+      R = absi(ZX_resultant(nf_get_pol(gel(vbnf,1)), nf_get_pol(nf)));
+      return Z_factor_addprimes(R, L);
+    }
+    case 3:
+    {
+      GEN P1 = nf_get_pol(gel(vbnf,1));
+      GEN P2 = nf_get_pol(gel(vbnf,2));
+      GEN P3 = nf_get_pol(gel(vbnf,3));
+      GEN L = gel(Z_factor(absi(ZX_resultant(P1,P2))),1);
+      L = Z_factor_addprimes(absi(ZX_resultant(P2,P3)),L);
+      return Z_factor_addprimes(absi(ZX_resultant(P3,P1)),L);
+    }
+    default: pari_err_BUG("vbnf_discfactors");
+      return NULL;/*LCOV_EXCL_LINE*/
+  }
+}
+
 static GEN
 ell2selmer(GEN ell, GEN ell_K, GEN help, GEN K, GEN vbnf,
            long effort, long flag, long prec)
@@ -1795,7 +1852,7 @@ ell2selmer(GEN ell, GEN ell_K, GEN help, GEN K, GEN vbnf,
   n = lg(vbnf) - 1; tors2 = n - 1;
   KP = gel(absZ_factor(K), 1);
   disc = ZX_disc(pol);
-  factdisc = mkvec3(KP, mkcol(gen_2), gel(absZ_factor(disc), 1));
+  factdisc = mkvec3(KP, mkcol(gen_2), vbnf_discfactors(vbnf));
   factdisc = ZV_sort_uniq_shallow(shallowconcat1(factdisc));
   discF = mkvec2(gmul(K,disc), factdisc);
   badprimes = cgetg(n+1, t_VEC);
@@ -1817,7 +1874,7 @@ ell2selmer(GEN ell, GEN ell_K, GEN help, GEN K, GEN vbnf,
 
     id = idealadd(nf, nf_get_index(nf), ZX_deriv(T));
     f = nf_pV_to_prV(nf, KP); settyp(f, t_COL);
-    f = mkvec3(gel(idealfactor(nf, Tinv), 1),
+    f = mkvec3(gel(idealfactor_partial(nf, Tinv, factdisc), 1),
                gel(idealfactor(nf, id), 1), f);
     gel(badprimes, k) = S = gtoset(shallowconcat1(f));
     if (n == 1)
@@ -1864,6 +1921,7 @@ ell2selmer(GEN ell, GEN ell_K, GEN help, GEN K, GEN vbnf,
   dim = lg(selmer)-1;
   if (DEBUGLEVEL) err_printf("Selmer rank: %ld\n", dim);
   newselmer = Flm_invimage(Flm_mul(LS2chars, selmer, 2), helpchars, 2);
+  if (!newselmer) pari_err_BUG("ellrank, wrong nfeltsign");
   nbpoints = lg(help) - 1;
   if (flag==1)
   {
