@@ -112,6 +112,21 @@ glength(GEN x)
   return lg(x) - lontyp[tx];
 }
 
+long
+gtranslength(GEN x)
+{
+  switch(typ(x))
+  {
+    case t_VEC: case t_COL:
+      return lg(x)-1;
+    case t_MAT:
+      return lg(x)==1 ? 0: nbrows(x);
+    default:
+      pari_err_TYPE("trans",x);
+      return 0; /* LCOV_EXCL_LINE */
+  }
+}
+
 GEN
 matsize(GEN x)
 {
@@ -420,7 +435,7 @@ gequal1(GEN x)
       return gequal1(gel(x,2)) && gequal0(gel(x,3));
 
     case t_POL: return is_monomial_test(x, 2, &gequal1);
-    case t_SER: return is_monomial_test(x, 2 - valp(x), &gequal1);
+    case t_SER: return is_monomial_test(x, 2 - valser(x), &gequal1);
 
     case t_RFRAC: return gequal(gel(x,1), gel(x,2));
     case t_COL: return col_test(x, &gequal1);
@@ -470,7 +485,7 @@ gequalm1(GEN x)
       return !degpol(gel(x,1)) || gequalm1(gel(x,2));
 
     case t_POL: return is_monomial_test(x, 2, &gequalm1);
-    case t_SER: return is_monomial_test(x, 2 - valp(x), &gequalm1);
+    case t_SER: return is_monomial_test(x, 2 - valser(x), &gequalm1);
 
     case t_RFRAC:
       av = avma; return gc_bool(av, gmequal_try(gel(x,1), gel(x,2)));
@@ -997,8 +1012,8 @@ serequal(GEN x, GEN y)
 {
   long LX, LY, lx, ly, vx, vy;
   if (!signe(x) && !signe(y)) return 1;
-  lx = lg(x); vx = valp(x); LX = lx + vx;
-  ly = lg(y); vy = valp(y); LY = ly + vy;
+  lx = lg(x); vx = valser(x); LX = lx + vx;
+  ly = lg(y); vy = valser(y); LY = ly + vy;
   if (LX > LY) lx = LY - vx; else ly = LX - vy;
   while (lx >= 3 && ly >= 3)
     if (!gequal(gel(x,--lx), gel(y,--ly))) return 0;
@@ -1239,7 +1254,7 @@ gvaluation(GEN x, GEN p)
     {
       case t_PADIC: return valp(x);
       case t_POL: return RgX_val(x);
-      case t_SER: return valp(x);
+      case t_SER: return valser(x);
       default: pari_err_TYPE("gvaluation", x);
     }
   tp  = typ(p);
@@ -1341,7 +1356,7 @@ gvaluation(GEN x, GEN p)
         {
           long val = RgX_val(p);
           if (!val) pari_err_DOMAIN("gvaluation", "p", "=", p, p);
-          return (long)(valp(x) / val);
+          return (long)(valser(x) / val);
         }
         if (varncmp(vx, vp) > 0) return 0;
       }
@@ -1850,6 +1865,56 @@ ZX_pvalrem(GEN x, GEN p, GEN *px) { return gen_pvalrem(x,p,px, 2); }
 long
 ZV_pvalrem(GEN x, GEN p, GEN *px) { return gen_pvalrem(x,p,px, 1); }
 
+static long
+ZX_gen_pvalrem(GEN x, GEN p, GEN *px, long imin)
+{
+  long i, lx, v;
+  GEN y;
+  y = cgetg_copy(x, &lx);
+  y[1] = x[1];
+  x = leafcopy(x);
+  for (i = imin; i < lx; i++)
+    if (typ(gel(x, i)) != t_INT)
+    {
+      gel(x, i) = leafcopy(gel(x,i));
+      gel(y, i) = leafcopy(gel(x,i));
+    }
+  for(v = 0;; v++)
+  {
+#if 0
+    if (v == VAL_DC_THRESHOLD) /* TODO */
+    {
+      if (is_pm1(p)) pari_err_DOMAIN("ZX_gen_pvalrem", "p", "=", p, p);
+      return v + ZX_gen_pvalrem_DC(x, p, px, imin);
+    }
+#endif
+
+    for (i = imin; i < lx; i++)
+    {
+      GEN r, xi = gel(x,i);
+      if (typ(xi) == t_INT)
+      {
+        gel(y,i) = dvmdii(xi, p, &r);
+        if (r != gen_0) { *px = x; return v; }
+      } else
+      {
+        long j, lxi = lg(xi);
+        for(j = 2; j < lxi; j++)
+        {
+          gmael(y,i,j) = dvmdii(gel(xi,j), p, &r);
+          if (r != gen_0) { *px = x; return v; }
+        }
+      }
+    }
+    swap(x, y);
+  }
+}
+
+long
+ZXX_pvalrem(GEN x, GEN p, GEN *px) { return ZX_gen_pvalrem(x,p,px, 2); }
+long
+ZXV_pvalrem(GEN x, GEN p, GEN *px) { return ZX_gen_pvalrem(x,p,px, 1); }
+
 /*******************************************************************/
 /*                                                                 */
 /*                       NEGATION: Create -x                       */
@@ -2051,11 +2116,14 @@ gabs(GEN x, long prec)
 
     case t_SER:
      if (!signe(x)) pari_err_DOMAIN("abs", "argument", "=", gen_0, x);
-     if (valp(x)) pari_err_DOMAIN("abs", "series valuation", "!=", gen_0, x);
+     if (valser(x)) pari_err_DOMAIN("abs", "series valuation", "!=", gen_0, x);
      return is_negative(gel(x,2))? gneg(x): gcopy(x);
 
     case t_VEC: case t_COL: case t_MAT:
       pari_APPLY_same(gabs(gel(x,i),prec));
+
+    case t_INFINITY:
+      return mkoo();
   }
   pari_err_TYPE("gabs",x);
   return NULL; /* LCOV_EXCL_LINE */
@@ -2118,7 +2186,10 @@ GEN
 vecmax0(GEN x, GEN *pi)
 {
   long i, lx = lg(x), tx = typ(x);
-  if (!is_matvec_t(tx) && tx != t_VECSMALL) return gcopy(x);
+  if (!is_matvec_t(tx) && tx != t_VECSMALL
+      && (tx != t_LIST || list_typ(x) != t_LIST_RAW)) return gcopy(x);
+  if (tx == t_LIST)
+  { if (list_data(x)) { x = list_data(x); lx = lg(x); } else lx = 1; }
   if (lx==1) pari_err_DOMAIN("vecmax", "empty argument", "=", x,x);
   switch(typ(x))
   {
@@ -2149,7 +2220,10 @@ GEN
 vecmin0(GEN x, GEN *pi)
 {
   long i, lx = lg(x), tx = typ(x);
-  if (!is_matvec_t(tx) && tx != t_VECSMALL) return gcopy(x);
+  if (!is_matvec_t(tx) && tx != t_VECSMALL
+      && (tx != t_LIST || list_typ(x) != t_LIST_RAW)) return gcopy(x);
+  if (tx == t_LIST)
+  { if (list_data(x)) { x = list_data(x); lx = lg(x); } else lx = 1; }
   if (lx==1) pari_err_DOMAIN("vecmin", "empty argument", "=", x,x);
   switch(typ(x))
   {
@@ -2672,7 +2746,7 @@ sizedigit(GEN x)
 GEN
 normalizeser(GEN x)
 {
-  long i, lx = lg(x), vx=varn(x), vp=valp(x);
+  long i, lx = lg(x), vx=varn(x), vp=valser(x);
   GEN y, z;
 
   if (lx == 2) { setsigne(x,0); return x; }
@@ -2683,7 +2757,7 @@ normalizeser(GEN x)
     if (isexactzero(z)) {
       /* dangerous case: already normalized ? */
       if (!signe(x)) return x;
-      setvalp(x,vp+1); /* no: normalize */
+      setvalser(x,vp+1); /* no: normalize */
     }
     setsigne(x,0); return x;
   }
@@ -2697,14 +2771,14 @@ normalizeser(GEN x)
     i -= 3; y = x + i;
     stackdummy((pari_sp)y, (pari_sp)x);
     gel(y,2) = z;
-    y[1] = evalsigne(0) | evalvalp(lx-2+vp) | evalvarn(vx);
+    y[1] = evalsigne(0) | evalvalser(lx-2+vp) | evalvarn(vx);
     y[0] = evaltyp(t_SER) | _evallg(3);
     return y;
   }
 
   i -= 2; y = x + i; lx -= i;
-  y[1] = evalsigne(1) | evalvalp(vp+i) | evalvarn(vx);
-  y[0] = evaltyp(t_SER) | evallg(lx);
+  y[1] = evalsigne(1) | evalvalser(vp+i) | evalvarn(vx);
+  y[0] = evaltyp(t_SER) | _evallg(lx);
 
   stackdummy((pari_sp)y, (pari_sp)x);
   for (i = 2; i < lx; i++)
@@ -2734,7 +2808,7 @@ normalizepol_lg(GEN x, long lx)
     if (! gequal0(z) ) {
       if (!LX) LX = i+1;
       stackdummy((pari_sp)(x + lg(x)), (pari_sp)(x + LX));
-      x[0] = evaltyp(t_POL) | evallg(LX);
+      x[0] = evaltyp(t_POL) | _evallg(LX);
       setsigne(x,1); return x;
     } else if (!isexactzero(z)) {
       if (!LX) LX = i+1; /* to be kept as leading coeff */
@@ -2749,7 +2823,7 @@ normalizepol_lg(GEN x, long lx)
       LX = 2; /* Pol(0) */
   }
   stackdummy((pari_sp)(x + lg(x)), (pari_sp)(x + LX));
-  x[0] = evaltyp(t_POL) | evallg(LX);
+  x[0] = evaltyp(t_POL) | _evallg(LX);
   setsigne(x,0); return x;
 }
 
@@ -2904,7 +2978,8 @@ listput0(GEN L, GEN x, long index)
 {
   if (typ(L) != t_LIST || list_typ(L) != t_LIST_RAW)
     pari_err_TYPE("listput",L);
-  return listput(L, x, index);
+  (void) listput(L, x, index);
+  return x;
 }
 
 GEN
@@ -2913,8 +2988,6 @@ listinsert(GEN L, GEN x, long index)
   long l, i;
   GEN z;
 
-  if (typ(L) != t_LIST || list_typ(L) != t_LIST_RAW)
-    pari_err_TYPE("listinsert",L);
   z = list_data(L); l = z? lg(z): 1;
   if (index <= 0) pari_err_COMPONENT("listinsert", "<=", gen_0, stoi(index));
   if (index > l) index = l;
@@ -2928,6 +3001,15 @@ listinsert(GEN L, GEN x, long index)
   return gel(z,index);
 }
 
+GEN
+listinsert0(GEN L, GEN x, long index)
+{
+  if (typ(L) != t_LIST || list_typ(L) != t_LIST_RAW)
+    pari_err_TYPE("listinsert",L);
+  (void) listinsert(L, x, index);
+  return x;
+}
+
 void
 listpop(GEN L, long index)
 {
@@ -2942,7 +3024,7 @@ listpop(GEN L, long index)
   if (!index || index > l) index = l;
   BLOCK_SIGINT_START
   gunclone_deep( gel(z, index) );
-  z[0] = evaltyp(t_VEC) | evallg(l);
+  z[0] = evaltyp(t_VEC) | _evallg(l);
   for (i=index; i < l; i++) z[i] = z[i+1];
   BLOCK_SIGINT_END
 }
