@@ -326,7 +326,7 @@ hyperellpadicfrobenius(GEN H, ulong p, long n)
 {
   pari_sp av = avma;
   GEN M = ZlX_hyperellpadicfrobenius(H, p, n);
-  GEN q = zeropadic(utoipos(p),n);
+  GEN q = zeropadic_shallow(utoipos(p),n);
   return gerepileupto(av, ZM_to_padic(M, q));
 }
 
@@ -599,7 +599,7 @@ GEN
 nfhyperellpadicfrobenius(GEN H, GEN T, ulong p, long n)
 {
   pari_sp av = avma;
-  GEN pp = utoipos(p), q = zeropadic(pp, n);
+  GEN pp = utoipos(p), q = zeropadic_shallow(pp, n);
   GEN M = ZlXQX_hyperellpadicfrobenius(lift_shallow(H),T,p,n);
   GEN MM = ZpXQM_prodFrobenius(M, T, pp, n);
   GEN m = gmul(ZXM_to_padic(MM, q), gmodulo(gen_1, T));
@@ -828,6 +828,30 @@ hyperellisoncurve(GEN W, GEN P)
   return gc_int(av, res);
 }
 
+GEN
+hyperellordinate(GEN W, GEN x)
+{
+  pari_sp av = avma;
+  if (typ(W)==t_POL)
+  {
+    GEN d = poleval(W,x), y;
+    if (gequal0(d)) { return gerepilecopy(av, mkvec(d)); }
+    if (!issquareall(d, &y)) { set_avma(av); return cgetg(1,t_VEC); }
+    return gerepilecopy(av, mkvec2(y, gneg(y)));
+  }
+  else
+  {
+    GEN b, c, d, rd, y;
+    if (typ(W)!=t_VEC || lg(W)!=3) pari_err_TYPE("hyperellisoncurve",W);
+    b = poleval(gel(W,2), x); c = poleval(gel(W,1), x);
+    d = gadd(gsqr(b), gmul2n(c, 2));
+    if (gequal0(d)) { return gerepilecopy(av, mkvec(gmul2n(gneg(b),-1))); }
+    if (!issquareall(d, &rd)) { set_avma(av); return cgetg(1,t_VEC); }
+    y = gmul2n(gsub(rd, b), -1);
+    return gerepilecopy(av, mkvec2(y, gsub(y,rd)));
+  }
+}
+
 GEN
 hyperelldisc(GEN PQ)
 {
@@ -877,7 +901,7 @@ algo51(GEN W, GEN M)
     if (vP==1) break;
     /* 2d */
     r = minss(2*vQ, vP)>>1;
-    gel(M,1) = shifti(gel(M,1), r);
+    if (M) gel(M,1) = shifti(gel(M,1), r);
     P = ZX_shifti(P, -2*r);
     Q = ZX_shifti(Q, -r);
   }
@@ -898,7 +922,7 @@ algo52(GEN W, GEN c, long *pt_lambda)
     /* 2 */
     if (2*mQ <= mP) { lambda = 2*mQ; break; }
     /* 3 */
-    if (mP%2 == 1) { lambda = mP; break; }
+    if (odd(mP)) { lambda = mP; break; }
     /* 4 */
     RgX_even_odd(FpX_red(ZX_shifti(Pc, -mP),gen_2),&H, &H1);
     if (signe(H1)) { lambda = mP; break; }
@@ -910,17 +934,10 @@ algo52(GEN W, GEN c, long *pt_lambda)
   return mkvec2(P,Q);
 }
 
-static GEN
-algo541(GEN F, GEN p, long ep, long g)
-{
-  GEN Fe = FpX_red(ep ? ZX_Z_divexact(F,p): F, p);
-  return FpX_roots_mult(Fe, g+2-ep, p);
-}
-
 static long
 test53(long lambda, long ep, long g)
 {
-  return (lambda <= g+1) || (g%2 && lambda<g+3 && ep==1);
+  return (lambda <= g+1) || (odd(g) && lambda<g+3 && ep==1);
 }
 
 static long
@@ -994,6 +1011,55 @@ algo56(GEN W, long g)
   return mkvec2(W, M);
 }
 
+static GEN
+algo56bis(GEN W, long g, long inf, long thr)
+{
+  pari_sp av = avma;
+  GEN vl = cgetg(3,t_VEC);
+  long nl = 1;
+  W = algo51(W, NULL);
+  if (inf)
+  {
+    GEN Woo = hyperell_reverse(W, g);
+    GEN Pc = ZX_unscale(gel(W,1), gen_2);
+    GEN Qc = ZX_unscale(gel(W,2), gen_2);
+    long ep = get_ep(Woo);
+    if (test55(mkvec2(Pc, Qc), ep, g))
+    {
+      long lambda;
+      Woo = algo52(Woo, gen_0, &lambda);
+      if (lambda == thr)
+      {
+        long r = lambda>>1;
+        gel(vl,nl++) = mkvec2(ZX_shifti(ZX_unscale(gel(Woo,1), gen_2), -2*r),
+                              ZX_shifti(ZX_unscale(gel(Woo,2), gen_2), -r));
+      }
+    }
+  }
+  {
+    long j, ep = get_ep(W);
+    for (j = 0; j < 2; j++)
+    {
+      long lambda;
+      GEN c = utoi(j);
+      GEN Pc = ZX_affine(gel(W,1), gen_2, c);
+      GEN Qc = ZX_affine(gel(W,2), gen_2, c);
+      if (test55(mkvec2(Pc, Qc), ep, g))
+      {
+        GEN Wc = algo52(W, c, &lambda);
+        if (lambda == thr)
+        {
+          long r = lambda>>1;
+          gel(vl,nl++) = mkvec2(ZX_shifti(ZX_affine(gel(Wc,1), gen_2,c), -2*r),
+                                ZX_shifti(ZX_affine(gel(Wc,2), gen_2,c), -r));
+        }
+      }
+    }
+  }
+  setlg(vl, nl);
+  return gerepilecopy(av,vl);
+}
+
 /* return the (degree 2) apolar invariant (the nth transvectant of P and P) */
 static GEN
 ZX_apolar(GEN P, long n)
@@ -1033,8 +1099,7 @@ algo57(GEN F, long g, GEN pr)
     long ep;
     GEN p = gel(pr, i), ps2 = shifti(p,-1), Fe;
     if (equaliu(p,2) || Z_pval(D,p) < minvd) continue;
-    ep = ZX_pval(F,p);
-    Fe = FpX_red(ep ? ZX_Z_divexact(F,p): F, p);
+    ep = ZX_pvalrem(F,p, &Fe); Fe = FpX_red(Fe, p);
     if (degpol(Fe) < g+1+ep)
     {
       GEN Fi = ZX_unscale(RgXn_recip_shallow(F,2*g+3), p);
@@ -1049,9 +1114,9 @@ algo57(GEN F, long g, GEN pr)
     }
     for(;;)
     {
-      long ep = ZX_pval(F,p);
-      GEN R = algo541(F, p, ep, g);
-      long j, lR = lg(R);
+      GEN Fe, R;
+      long j, lR, ep = ZX_pvalrem(F,p, &Fe);
+      R = FpX_roots_mult(FpX_red(Fe, p), g+2-ep, p); lR = lg(R);
       for (j = 1; j<lR; j++)
       {
         GEN c = Fp_center(gel(R,j), p, ps2);
@@ -1072,6 +1137,76 @@ algo57(GEN F, long g, GEN pr)
   return mkvec2(F, M);
 }
 
+/* if inf=0, ignore point at infinity */
+static GEN
+algo57bis(GEN F, long g, GEN p, long inf, long thr)
+{
+  pari_sp av = avma;
+  GEN vl = cgetg(3,t_VEC), Fe;
+  long nl = 1, ep = ZX_pvalrem(F,p, &Fe);
+  Fe = FpX_red(Fe, p);
+  {
+    GEN R = FpX_roots_mult(Fe, thr-ep, p);
+    long j, lR = lg(R);
+    for (j = 1; j<lR; j++)
+    {
+      GEN Fj = ZX_affine(F, p, gel(R,j));
+      long lambda = ZX_pvalrem(Fj, p, &Fj);
+      if (lambda == thr) gel(vl,nl++) = odd(lambda)? ZX_Z_mul(Fj, p): Fj;
+    }
+  }
+  if (inf==1 && 2*g+2-degpol(Fe) >= thr-ep)
+  {
+    GEN Fj = ZX_unscale(RgXn_recip_shallow(F,2*g+3), p);
+    long lambda = ZX_pvalrem(Fj, p, &Fj);
+    if (lambda == thr) gel(vl,nl++) = odd(lambda)? ZX_Z_mul(Fj, p): Fj;
+  }
+  setlg(vl, nl);
+  return gerepilecopy(av,vl);
+}
+
+static GEN
+next_model(GEN G, long g, GEN p, long inf, long thr)
+{
+  return equaliu(p,2) ? algo56bis(G, g,    inf, thr)
+                      : algo57bis(G, g, p, inf, thr);
+}
+
+static GEN
+get_extremal(GEN F, GEN G, long g, GEN p)
+{
+  while (1)
+  {
+    GEN Wi;
+    Wi = next_model(G, g, p, 0, g+2);
+    if (lg(Wi)==1) return F;
+    F = gel(Wi,1);
+    Wi = next_model(F, g, p, 0, g+1);
+    if (lg(Wi)==1) return F;
+    G = gel(Wi,1);
+  }
+}
+
+GEN
+hyperellextremalmodels(GEN F, long g, GEN p)
+{
+  pari_sp av = avma;
+  GEN R, W;
+  long l;
+  if (equaliu(p,2))
+  {
+    if (get_ep(F)>0) return mkvec(F);
+  } else
+    if( ZX_pval(F,p) > 0) return mkvec(F);
+  W = next_model(F, g, p, 1, g+1); l = lg(W);
+  if (l==1) { set_avma(av); return mkvec(F); }
+  R = cgetg(3, t_VEC);
+  gel(R,1) = get_extremal(F, gel(W,1), g, p);
+  gel(R,2) = l==3 ? get_extremal(F, gel(W,2), g, p) : F;
+  if (gel(R,2) == gel(R,1)) setlg(R,2);
+  return gerepilecopy(av, R);
+}
+
 static GEN
 RgX_RgM2_eval(GEN P, GEN A, GEN Bp, long d)
 {
@@ -1194,7 +1329,7 @@ redqfbsplit(GEN a, GEN b, GEN c, GEN d)
 
   if (!equali1(w)) { p = diviiexact(p, w); q = diviiexact(q, w); }
   U = mkmat22(p, negi(v), q, u);
-  Q = qfb_apply_ZM(mkvec3(a,b,c), U);
+  Q = qfb_ZM_apply(mkvec3(a,b,c), U);
   b = gel(Q, 2); c = gel(Q,3);
   if (signe(b) < 0) gel(U,2) = mkcol2(v, negi(u));
   gel(U,2) = ZC_lincomb(gen_1, truedivii(negi(c), d), gel(U,2), gel(U,1));
@@ -1215,20 +1350,25 @@ static GEN
 red_Cremona_Stoll(GEN P, GEN *pM)
 {
   GEN q1, q2, q3, M, R;
-  long i, prec = nbits2prec(2*gexpo(P)) + 1, d = degpol(P);
-  GEN dP = ZX_deriv(P), r = QX_complex_roots(P, prec);
-  q1 = gen_0; q2 = gen_0; q3 = gen_0;
-  for (i = 1; i <= d; i++)
+  long i, prec = nbits2prec(2*gexpo(P)) + EXTRAPRECWORD, d = degpol(P);
+  GEN dP = ZX_deriv(P);
+  for (;;)
   {
-    GEN ri = gel(r,i);
-    GEN s = ginv(gabs(RgX_cxeval(dP,ri,NULL), prec));
-    if (d!=4) s = gpow(s, gdivgs(gen_2,d-2), prec);
-    q1 = gadd(q1, s);
-    q2 = gsub(q2, gmul(real_i(ri), s));
-    q3 = gadd(q3, gmul(gnorm(ri), s));
+    GEN r = QX_complex_roots(P, prec);
+    q1 = gen_0; q2 = gen_0; q3 = gen_0;
+    for (i = 1; i <= d; i++)
+    {
+      GEN ri = gel(r,i);
+      GEN s = ginv(gabs(RgX_cxeval(dP,ri,NULL), prec));
+      if (d!=4) s = gpow(s, gdivgs(gen_2,d-2), prec);
+      q1 = gadd(q1, s);
+      q2 = gsub(q2, gmul(real_i(ri), s));
+      q3 = gadd(q3, gmul(gnorm(ri), s));
+    }
+    M = lllgram(mkmat22(q1,q2,q2,q3));
+    if (M && lg(M) == 3) break;
+    prec = precdbl(prec);
   }
-  M = lllgram(mkmat22(q1,q2,q2,q3));
-  if (lg(M) != 3) M = matid(2);
   R = polreduce(P, M);
   *pM = M;
   return R;
