@@ -20,9 +20,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /********************************************************/
 /*                   Hurwitz zeta function              */
 /********************************************************/
-struct hurwitzp_t { GEN B, _1, s1; };
+struct Qp_zetahurwitz_t { GEN B, _1, s1; };
 static void
-hurwitzp_init(struct hurwitzp_t *S, long prec, GEN s)
+Qp_zetahurwitz_init(struct Qp_zetahurwitz_t *S, long prec, GEN s)
 {
   GEN B, C = gen_1, s1 = gsubgs(s, 1), p = gel(s, 2);
   long j, J = ((equaliu(p,2)? (prec >> 1): prec) + 2) >> 1;
@@ -39,9 +39,9 @@ hurwitzp_init(struct hurwitzp_t *S, long prec, GEN s)
   S->B = B;
 }
 
-/* s1 = s-1 or NULL (if s=1) */
+/* v_p(x) < (p==2)?-1: 0; s1 = s-1 or NULL (if s=1) */
 static GEN
-hurwitzp_i(struct hurwitzp_t *S, GEN x)
+Qp_zetahurwitz_0(struct Qp_zetahurwitz_t *S, GEN x)
 {
   GEN z, x2, x2j, s1 = S->s1;
   long j, J = lg(S->B) - 2;
@@ -64,23 +64,23 @@ pprec(GEN x) { return maxss(valp(x) + precp(x), 1); }
 
 /* L_p(s, (D, .)); assume s != 1 if D = 1 */
 static GEN
-zetap_i(GEN s, long D)
+Qp_zeta_i(GEN s, long D)
 {
   pari_sp av = avma;
   GEN z, va, gp = gel(s,2);
   ulong a, p = itou(gp), m;
   long prec = pprec(s);
-  struct hurwitzp_t S;
+  struct Qp_zetahurwitz_t S;
 
   if (D <= 0) pari_err_DOMAIN("p-adic L-function", "D", "<=", gen_0, stoi(D));
   if (!uposisfundamental(D))
     pari_err_TYPE("p-adic L-function [D not fundamental]", stoi(D));
-  hurwitzp_init(&S, prec, s);
+  Qp_zetahurwitz_init(&S, prec, s);
   m = ulcm(D, p == 2? 4: p); va = coprimes_zv(m);
   for (a = 1, z = gen_0; a <= (m >> 1); a++)
     if (va[a])
     {
-      GEN h = hurwitzp_i(&S, uutoQ(a, m));
+      GEN h = Qp_zetahurwitz_0(&S, uutoQ(a, m));
       if (D != 1 && kross(D, a) < 0) h = gneg(h);
       z = gadd(z, h);
     }
@@ -89,16 +89,16 @@ zetap_i(GEN s, long D)
   return gerepileupto(av, z);
 }
 GEN
-Qp_zeta(GEN s) { return zetap_i(s, 1); }
+Qp_zeta(GEN s) { return Qp_zeta_i(s, 1); }
 
-/* s a t_PADIC; gerepileupto-safe */
+/* s a t_PADIC; gerepileupto-safe. Could be exported */
 static GEN
-hurwitzp(GEN s, GEN x)
+Qp_zetahurwitz_ii(GEN s, GEN x, long k)
 {
   GEN gp = gel(s,2);
   long p = itou(gp), prec = pprec(s);
-  struct hurwitzp_t S;
-  hurwitzp_init(&S, prec, s);
+  struct Qp_zetahurwitz_t S;
+  Qp_zetahurwitz_init(&S, prec, s);
   if (typ(x) != t_PADIC) x = gadd(x, zeropadic_shallow(gp, prec));
   if (valp(x) >= ((p==2)? -1: 0))
   {
@@ -107,12 +107,40 @@ hurwitzp(GEN s, GEN x)
     for (j = 0; j < M; j++)
     {
       GEN y = gaddsg(j, x);
-      if (valp(y) <= 0) z = gadd(z, hurwitzp_i(&S, gdivgu(y, M)));
+      if (valp(y) <= 0)
+      {
+        GEN tmp = Qp_zetahurwitz_0(&S, gdivgu(y, M));
+        if (k) tmp = gmul(tmp, gpowgs(teich(y), k));
+        z = gadd(z, tmp);
+      }
     }
     return gdivgu(z, M);
   }
-  if (valp(s) < 0) pari_err_DOMAIN("hurwitzp", "v(s)", "<", gen_0, s);
-  return hurwitzp_i(&S, x);
+  if (valp(s) < 0) pari_err_DOMAIN("Qp_zetahurwitz", "v(s)", "<", gen_0, s);
+  return Qp_zetahurwitz_0(&S, x);
+}
+
+/* x or s must be p-adic */
+static GEN
+Qp_zetahurwitz_i(GEN s, GEN x, long k)
+{
+  if (typ(x) == t_PADIC)
+  {
+    pari_sp av = avma;
+    GEN p = gel(x,2);
+    long e = pprec(x);
+    e += sdivsi(e, subis(p, 1));
+    s = gadd(s, zeropadic_shallow(p, e));
+    return gerepileupto(av, Qp_zetahurwitz_ii(s, x, k));
+  }
+  return Qp_zetahurwitz_ii(s, x, k);
+}
+
+GEN
+Qp_zetahurwitz(GEN s, GEN x, long k)
+{
+  pari_sp av = avma;
+  return gerepileupto(av, Qp_zetahurwitz_i(s, x, k));
 }
 
 static void
@@ -146,6 +174,13 @@ static GEN
 deg1ser_shallow(GEN a1, GEN a0, long v, long e)
 { return RgX_to_ser(deg1pol_shallow(a1, a0, v), e+2); }
 
+static long
+hurwitz_cutoff(GEN s, long bit)
+{
+  return typ(s) == t_COMPLEX &&
+         fabs(gtodouble(gel(s,2))) > 5.37 * pow(bit, 1.4) / mt_nbthreads();
+}
+
 /* New zetahurwitz, from Fredrik Johansson. */
 GEN
 zetahurwitz(GEN s, GEN x, long der, long bitprec)
@@ -173,12 +208,8 @@ zetahurwitz(GEN s, GEN x, long der, long bitprec)
     }
     return gerepileupto(av,z);
   }
-  if (typ(s) == t_PADIC) return gerepileupto(av, hurwitzp(s, x));
-  if (typ(x) == t_PADIC)
-  {
-    s = gadd(s, zeropadic_shallow(gel(x,2), pprec(x)));
-    return gerepileupto(av, hurwitzp(s, x));
-  }
+  if (typ(s) == t_PADIC || typ(x) == t_PADIC)
+    return gerepileupto(av, Qp_zetahurwitz_i(s, x, 0));
   switch(typ(x))
   {
     case t_INT: case t_REAL: case t_FRAC: case t_COMPLEX: break;
@@ -193,10 +224,13 @@ zetahurwitz(GEN s, GEN x, long der, long bitprec)
   switch (typ(s))
   {
     long v, pr;
-    case t_INT: case t_REAL: case t_FRAC: case t_COMPLEX: break;
+    case t_INT: case t_REAL: case t_FRAC: case t_COMPLEX:
+      if (!der && hurwitz_cutoff(s, bitprec))
+        return zetahurwitzlarge(s, x, prec);
+      break;
     default:
       if (!(y = toser_i(s))) pari_err_TYPE("zetahurwitz", s);
-      if (valp(y) < 0) pari_err_DOMAIN("zetahurwitz", "val(s)", "<", gen_0, s);
+      if (valser(y) < 0) pari_err_DOMAIN("zetahurwitz", "val(s)", "<", gen_0, s);
       s0 = polcoef_i(y, 0, -1);
       switch(typ(s0))
       {
@@ -205,7 +239,7 @@ zetahurwitz(GEN s, GEN x, long der, long bitprec)
         default: pari_err_TYPE("zetahurwitz", s0);
       }
       sch = gequal0(s0)? y: serchop0(y);
-      v = valp(sch);
+      v = valser(sch);
       pr = (lg(y) + v + 1) / v;
       if (gequal1(s0)) pr += v;
       s = deg1ser_shallow(gen_1, s0, 0, pr);
@@ -245,7 +279,11 @@ zetahurwitz(GEN s, GEN x, long der, long bitprec)
     C = gmul2n(gabs(C,LOWDEFAULTPREC), bitprec + 1);
     C = gpow(C, ginv(gsubsg(k+1, ra)), LOWDEFAULTPREC);
     /* need |N + x - 1|^2 > C^2 */
-    if (!gequal0(ix)) C = gsqrt(gsub(gsqr(C), gsqr(ix)), LOWDEFAULTPREC);
+    if (!gequal0(ix))
+    {
+      GEN tmp = gsub(gsqr(C), gsqr(ix));
+      if (gsigne(tmp) >= 0) C = gsqrt(tmp, LOWDEFAULTPREC);
+    }
     /* need |N + re(x) - 1| > C */
     C = gceil(gadd(C, gsubsg(1, rx)));
     if (typ(C) != t_INT) pari_err_TYPE("zetahurwitz",s);
@@ -269,10 +307,11 @@ zetahurwitz(GEN s, GEN x, long der, long bitprec)
   a = gneg(s);
   if (DEBUGLEVEL>2) timer_start(&T);
   incrprec(prec);
-  Nx = gmul(real_1(prec), gaddsg(N - 1, x));
+  Nx = gaddsg(N - 1, x);
+  Nx = typ(Nx) == t_SER? RgX_gtofp(Nx, prec): gtofp(Nx, prec);
   S1 = S3 = gpow(Nx, a, prec);
   av2 = avma;
-  if (gequal1(x)) S1 = dirpowerssum(N, a, prec);
+  if (gequal1(x)) S1 = dirpowerssum(N, a, 0, prec);
   else
     for (m = N - 2; m >= 0; m--)
     {
@@ -559,5 +598,6 @@ lerchzeta(GEN s, GEN a, GEN lam, long prec)
   if (!iscplx(z)) pari_err_TYPE("lerchzeta", z);
   if (!iscplx(s)) pari_err_TYPE("lerchzeta", s);
   if (!iscplx(a)) pari_err_TYPE("lerchzeta", a);
+  if (hurwitz_cutoff(s, prec)) return lerchzetalarge(s, a, lam, prec);
   return gerepileupto(av, _lerchphi(z, s, a, prec));
 }
