@@ -17,6 +17,431 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #define DEBUGLEVEL DEBUGLEVEL_qflll
 
+static int
+RgM_is_square_mat(GEN x) { long l = lg(x); return l == 1 || l == lgcols(x); }
+
+static long
+ZM_is_upper(GEN R)
+{
+  long i,j, l = lg(R);
+  if (l != lgcols(R)) return 0;
+  for(i = 1; i < l; i++)
+    for(j = 1; j < i; j++)
+      if (signe(gcoeff(R,i,j))) return 0;
+  return 1;
+}
+
+static long
+ZM_is_knapsack(GEN R)
+{
+  long i,j, l = lg(R);
+  if (l != lgcols(R)) return 0;
+  for(i = 2; i < l; i++)
+    for(j = 1; j < l; j++)
+      if ( i!=j && signe(gcoeff(R,i,j))) return 0;
+  return 1;
+}
+
+static long
+ZM_is_lower(GEN R)
+{
+  long i,j, l = lg(R);
+  if (l != lgcols(R)) return 0;
+  for(i = 1; i < l; i++)
+    for(j = 1; j < i; j++)
+      if (signe(gcoeff(R,j,i))) return 0;
+  return 1;
+}
+
+static GEN
+RgM_flip(GEN R)
+{
+  GEN M;
+  long i,j,l;
+  M = cgetg_copy(R, &l);
+  for(i = 1; i < l; i++)
+  {
+    gel(M,i) = cgetg(l, t_COL);
+    for(j = 1; j < l; j++)
+      gmael(M,i,j) = gmael(R,l-i, l-j);
+  }
+  return M;
+}
+
+static GEN
+RgM_flop(GEN R)
+{
+  GEN M;
+  long i,j,l;
+  M = cgetg_copy(R, &l);
+  for(i = 1; i < l; i++)
+  {
+    gel(M,i) = cgetg(l, t_COL);
+    for(j = 1; j < l; j++)
+      gmael(M,i,j) = gmael(R,i, l-j);
+  }
+  return M;
+}
+
+/* Assume x and y has same type! */
+INLINE int
+mpabscmp(GEN x, GEN y)
+{
+  return (typ(x)==t_INT) ? abscmpii(x,y) : abscmprr(x,y);
+}
+
+/****************************************************************************/
+/***                             FLATTER                                  ***/
+/****************************************************************************/
+/* Implementation of "FLATTER" algorithm based on
+ * <https://eprint.iacr.org/2023/237>
+ * Fast Practical Lattice Reduction through Iterated Compression
+ *
+ * Keegan Ryan, University of California, San Diego
+ * Nadia Heninger, University of California, San Diego. BA20230925 */
+static long
+drop(GEN R)
+{
+  long i, n = lg(R)-1;
+  long s = 0, m = mpexpo(gcoeff(R, 1, 1));
+  for (i = 2; i <= n; ++i)
+  {
+    if (mpabscmp(gcoeff(R, i, i), gcoeff(R, i - 1, i - 1)) >= 0)
+    {
+      s += m - mpexpo(gcoeff(R, i - 1, i - 1));
+      m = mpexpo(gcoeff(R, i, i));
+    }
+  }
+  s += m - mpexpo(gcoeff(R, n, n));
+  return s;
+}
+
+static long
+potential(GEN R)
+{
+  long i, n = lg(R)-1;
+  long s = 0, mul = n-1;;
+  for (i = 1; i <= n; i++, mul-=2) s += mul * mpexpo(gcoeff(R,i,i));
+  return s;
+}
+
+/* U upper-triangular invertible:
+ * Bound on the exponent of the condition number of U.
+ * Algo 8.13 in Higham, Accuracy and stability of numercal algorithms. */
+static long
+condition_bound(GEN U, int lower)
+{
+  long n = lg(U)-1, e, i, j;
+  GEN y;
+  pari_sp av = avma;
+  y = cgetg(n+1, t_VECSMALL);
+  e = y[n] = -gexpo(gcoeff(U,n,n));
+  for (i=n-1; i>0; i--)
+  {
+    long s = 0;
+    for (j=i+1; j<=n; j++)
+      s = maxss(s, (lower? gexpo(gcoeff(U,j,i)): gexpo(gcoeff(U,i,j))) + y[j]);
+    y[i] = s - gexpo(gcoeff(U,i,i));
+    e = maxss(e, y[i]);
+  }
+  return gc_long(av, gexpo(U) + e);
+}
+
+static long
+gsisinv(GEN M)
+{
+  long i, l = lg(M);
+  for (i = 1; i < l; ++i)
+    if (! signe(gmael(M, i, i))) return 0;
+  return 1;
+}
+
+INLINE long
+nbits2prec64(long n)
+{
+  return nbits2prec(((n+63)>>6)<<6);
+}
+
+static long
+spread(GEN R)
+{
+  long i, n = lg(R)-1, m = mpexpo(gcoeff(R, 1, 1)), M = m;
+  for (i = 2; i <= n; ++i)
+  {
+    long e = mpexpo(gcoeff(R, i, i));
+    if (e < m) m = e;
+    if (e > M) M = e;
+  }
+  return M - m;
+}
+
+static long
+GS_extraprec(GEN L, int lower)
+{
+  long C = condition_bound(L, lower), S = spread(L), n = lg(L)-1;
+  return maxss(2*S+2*n, C-S-2*n); /* = 2*S + 2*n + maxss(0, C-3*S-4*n) */
+}
+
+static GEN
+RgM_Cholesky_dynprec(GEN M)
+{
+  pari_sp ltop = avma;
+  GEN L;
+  long minprec = lg(M) + 30, bitprec = minprec, prec;
+  while (1)
+  {
+    long mbitprec;
+    prec = nbits2prec64(bitprec);
+    L = RgM_Cholesky(RgM_gtofp(M, prec), prec); /* upper-triangular */
+    if (!L)
+    {
+      bitprec *= 2;
+      set_avma(ltop);
+      continue;
+    }
+    mbitprec = minprec + GS_extraprec(L, 0);
+    if (bitprec >= mbitprec)
+      break;
+    bitprec = maxss((4*bitprec)/3, mbitprec);
+    set_avma(ltop);
+  }
+  return gerepilecopy(ltop, L);
+}
+
+static GEN
+gramschmidt_upper(GEN M)
+{
+  long bitprec = lg(M)-1 + 31 + GS_extraprec(M, 0);
+  return RgM_gtofp(M, nbits2prec64(bitprec));
+}
+
+static GEN
+gramschmidt_dynprec(GEN M)
+{
+  pari_sp ltop = avma;
+  long minprec = lg(M) + 30, bitprec = minprec;
+  if (ZM_is_upper(M)) return gramschmidt_upper(M);
+  while (1)
+  {
+    GEN B, Q, L;
+    long prec = nbits2prec64(bitprec), mbitprec;
+    if (!QR_init(RgM_gtofp(M, prec), &B, &Q, &L, prec) || !gsisinv(L))
+    {
+      bitprec *= 2;
+      set_avma(ltop);
+      continue;
+    }
+    mbitprec = minprec + GS_extraprec(L, 1);
+    if (bitprec >= mbitprec)
+      return gerepilecopy(ltop, shallowtrans(L));
+    bitprec = maxss((4*bitprec)/3, mbitprec);
+    set_avma(ltop);
+  }
+}
+/* return -T1 * round(T1^-1*(R1^-1*R2)*T3) */
+static GEN
+sizered(GEN T1, GEN T3, GEN R1, GEN R2)
+{
+  pari_sp ltop = avma;
+  long e;
+  return gerepileupto(ltop, ZM_mul(ZM_neg(T1), grndtoi(gmul(ZM_inv(T1,NULL),
+         RgM_mul(RgM_mul(RgM_inv_upper(R1), R2), T3)), &e)));
+}
+
+static GEN
+flat(GEN M, long flag, GEN *pt_T, long *pt_s, long *pt_pot)
+{
+  pari_sp ltop = avma;
+  GEN R, R1, R2, R3, T1, T2, T3, T, S;
+  long k = lg(M)-1, n = k>>1, n2 = k - n, m = n>>1;
+  long keepfirst = flag & LLL_KEEP_FIRST, inplace = flag & LLL_INPLACE;
+  /* for k = 3, we want n = 1; n2  = 2; m = 0 */
+  /* for k = 5,         n = 2; n2 = 3; m = 1 */
+  R = gramschmidt_dynprec(M);
+  R1 = matslice(R, 1, n, 1, n);
+  R2 = matslice(R, 1, n, n + 1, k);
+  R3 = matslice(R, n + 1, k, n + 1, k);
+  T1 = lllfp(R1, 0.99, LLL_IM| LLL_UPPER| LLL_NOCERTIFY| (keepfirst ? LLL_KEEP_FIRST: 0));
+  T3 = lllfp(R3, 0.99, LLL_IM| LLL_UPPER| LLL_NOCERTIFY);
+  T2 = sizered(T1, T3, R1, R2);
+  T = shallowmatconcat(mkmat22(T1,T2,gen_0,T3));
+  M = ZM_mul(M, T);
+  R = gramschmidt_dynprec(M);
+  R3 = matslice(R, m + 1, m + n2, m + 1, m + n2);
+  T3 = lllfp(R3, 0.99, LLL_IM| LLL_UPPER| LLL_NOCERTIFY);
+  S = shallowmatconcat(diagonal(
+       m == 0     ? mkvec2(T3, matid(k - m - n2))
+     : m+n2 == k  ? mkvec2(matid(m), T3)
+                  : mkvec3(matid(m), T3, matid(k - m - n2))));
+  M = ZM_mul(M, S);
+  if (!inplace) *pt_T = ZM_mul(T, S);
+  *pt_s = drop(R);
+  *pt_pot = potential(R);
+  return gc_all(ltop, inplace ? 1: 2, &M, pt_T);
+}
+
+static GEN
+ZM_flatter(GEN M, long flag)
+{
+  pari_sp ltop = avma, btop;
+  long i, n = lg(M)-1;
+  long s = -1, pot = LONG_MAX;
+  GEN T;
+  pari_timer ti;
+  long lti = 1;
+  long inplace = flag & LLL_INPLACE, cert = !(flag & LLL_NOCERTIFY);
+  T = matid(n);
+  if (DEBUGLEVEL>=3)
+  {
+    timer_start(&ti);
+    if (cert)
+      err_printf("flatter dim = %ld size = %ld\n", n, ZM_max_expi(M));
+  }
+  btop = avma;
+  for (i = 1;;i++)
+  {
+    long t, pot2;
+    GEN U, M2 = flat(M, flag, &U, &t, &pot2);
+    if (t==0) { s = t; break; }
+    if (s >= 0)
+    {
+      if (s==t && pot>=pot2)
+        break;
+      if (s<t && i > 20)
+      {
+        if (DEBUGLEVEL >= 3) err_printf("BACK:%ld:%ld:%g\n", n, i, s);
+        break;
+      }
+    }
+    if (DEBUGLEVEL>=3 && (cert || timer_get(&ti) > 1000))
+    {
+      if (i==lti)
+        timer_printf(&ti, "FLATTER, dim %ld, step %ld: \t slope=%0.10g \t pot=%0.10g", n, i, ((double)t)/n, ((double)pot2)/(n*(n+1)));
+      else
+        timer_printf(&ti, "FLATTER, dim %ld, steps %ld-%ld: \t slope=%0.10g \t pot=%0.10g", n, lti,i, ((double)t)/n, ((double)pot2)/(n*(n+1)));
+      lti = i+1;
+    }
+    s = t;
+    pot = pot2;
+    M = M2;
+    if (!inplace) T = ZM_mul(T, U);
+    if (gc_needed(btop, 1))
+      gerepileall(btop, 2, &M, &T);
+  }
+  if (DEBUGLEVEL>=3 && (cert || timer_get(&ti) > 1000))
+  {
+    if (i==lti)
+      timer_printf(&ti, "FLATTER, dim %ld, final \t slope=%0.10g \t pot=%0.10g", n, ((double)s)/n, ((double)pot)/(n*(n+1)));
+    else
+      timer_printf(&ti, "FLATTER, dim %ld, steps %ld-final:\t slope=%0.10g \t pot=%0.10g", n, lti, ((double)s)/n, ((double)pot)/(n*(n+1)));
+  }
+  return  gerepilecopy(ltop, inplace ? M: T);
+}
+
+static GEN
+ZM_flatter_rank(GEN M, long rank, long flag)
+{
+  pari_timer ti;
+  pari_sp ltop = avma;
+  GEN T;
+  long i, n = lg(M)-1;
+  if (rank == n)  return ZM_flatter(M, flag);
+  T = matid(n);
+  if (DEBUGLEVEL>=3) timer_start(&ti);
+  for (i = 1;; i++)
+  {
+    GEN S = ZM_flatter(vconcat(gshift(M,i),matid(n)), flag);
+    if (DEBUGLEVEL>=3)
+      timer_printf(&ti,"FLATTERRANK step %ld: %ld",i,expi(gnorml2(S)));
+    if (ZM_isidentity(S)) break;
+    T = ZM_mul(T, S);
+    M = ZM_mul(M, S);
+    if (gc_needed(ltop, 1))
+      gerepileall(ltop, 2, &M, &T);
+  }
+  return gerepilecopy(ltop, T);
+}
+
+static GEN
+flattergram_i(GEN M, long flag, long *pt_s)
+{
+  pari_sp ltop = avma;
+  GEN R, T;
+  R = RgM_Cholesky_dynprec(M);
+  *pt_s = drop(R);
+  T =  lllfp(R, 0.99, LLL_IM| LLL_UPPER| LLL_NOCERTIFY | (flag&LLL_KEEP_FIRST));
+  return gerepilecopy(ltop, T);
+}
+
+static GEN
+ZM_flattergram(GEN M, long flag)
+{
+  pari_sp ltop = avma, btop;
+  GEN T;
+  long i, n = lg(M)-1;
+  pari_timer ti;
+  long s = -1;
+  if (DEBUGLEVEL>=3)
+  {
+    timer_start(&ti);
+    err_printf("FLATTERGRAM dim = %ld size = %ld\n", n, ZM_max_expi(M));
+  }
+  btop = avma;
+  T = matid(n);
+  for (i = 1;; i++)
+  {
+    long t;
+    GEN S = flattergram_i(M, flag, &t);
+    t = expi(gnorml2(S));
+    if (t==0) { s = t;  break; }
+    if (s)
+    {
+      double st = s-t;
+      if (st == 0)
+        break;
+      if (st < 0 && i > 20)
+      {
+        if (DEBUGLEVEL >= 3) err_printf("BACK:%ld:%ld:%0.10g\n", n, i, ((double)s)/n);
+        break;
+      }
+    }
+    T = ZM_mul(T, S);
+    M = qf_ZM_apply(M, S);
+    s = t;
+    if (DEBUGLEVEL >= 3)
+      timer_printf(&ti, "FLATTERGRAM, dim %ld step %ld, slope=%0.10g", n, i, ((double)s)/n);
+    if (gc_needed(btop, 1))
+      gerepileall(btop, 2, &M, &T);
+  }
+  if (DEBUGLEVEL >= 3)
+    timer_printf(&ti, "FLATTERGRAM, dim %ld, slope=%0.10g", n, ((double)s)/n);
+  return gerepilecopy(ltop, T);
+}
+
+static GEN
+ZM_flattergram_rank(GEN M, long rank, long flag)
+{
+  pari_timer ti;
+  pari_sp ltop = avma;
+  GEN T;
+  long i, n = lg(M)-1;
+  if (rank == n)  return ZM_flattergram(M, flag);
+  T = matid(n);
+  if (DEBUGLEVEL>=3) timer_start(&ti);
+  for (i = 1;; i++)
+  {
+    GEN S = ZM_flattergram(RgM_Rg_add(gshift(M, i), gen_1), flag);
+    if (DEBUGLEVEL>=3)
+      timer_printf(&ti,"FLATTERGRAMRANK step %ld: %ld",i,expi(gnorml2(S)));
+    if (ZM_isidentity(S)) break;
+    T = ZM_mul(T, S);
+    M = qf_ZM_apply(M, S);
+    if (gc_needed(ltop, 1))
+      gerepileall(ltop, 2, &M, &T);
+  }
+  return gerepilecopy(ltop, T);
+}
+
 static double
 pari_rint(double a)
 {
@@ -66,7 +491,7 @@ static GEN
 vectail_inplace(GEN x, long k)
 {
   if (!k) return x;
-  x[k] = ((ulong)x[0] & ~LGBITS) | evallg(lg(x) - k);
+  x[k] = ((ulong)x[0] & ~LGBITS) | _evallg(lg(x) - k);
   return x + k;
 }
 
@@ -112,6 +537,23 @@ submulshift(GEN x, GEN y, GEN z, long e)
   t = shifti(mulii(z, y), e);
   set_avma(av); return subii(x, t);
 }
+static void
+subzi(GEN *a, GEN b)
+{
+  pari_sp av = avma;
+  b = subii(*a, b);
+  if (lgefint(b)<=lg(*a) && isonstack(*a)) { affii(b,*a); set_avma(av); }
+  else *a = b;
+}
+
+static void
+addzi(GEN *a, GEN b)
+{
+  pari_sp av = avma;
+  b = addii(*a, b);
+  if (lgefint(b)<=lg(*a) && isonstack(*a)) { affii(b,*a); set_avma(av); }
+  else *a = b;
+}
 
 /* x - u*y * 2^e */
 INLINE GEN
@@ -125,6 +567,19 @@ submuliu2n(GEN x, GEN y, ulong u, long e)
   y = shifti(mului(u,y), e);
   set_avma(av); return subii(x, y);
 }
+/* *x -= u*y * 2^e */
+INLINE void
+submulzu2n(GEN *x, GEN y, ulong u, long e)
+{
+  pari_sp av;
+  long ly = lgefint(y);
+  if (ly == 2) return;
+  av = avma;
+  (void)new_chunk(3+ly+lgefint(*x)+nbits2lg(e)); /* HACK */
+  y = shifti(mului(u,y), e);
+  set_avma(av); return subzi(x, y);
+}
+
 /* x + u*y * 2^e */
 INLINE GEN
 addmuliu2n(GEN x, GEN y, ulong u, long e)
@@ -138,6 +593,19 @@ addmuliu2n(GEN x, GEN y, ulong u, long e)
   set_avma(av); return addii(x, y);
 }
 
+/* *x += u*y * 2^e */
+INLINE void
+addmulzu2n(GEN *x, GEN y, ulong u, long e)
+{
+  pari_sp av;
+  long ly = lgefint(y);
+  if (ly == 2) return;
+  av = avma;
+  (void)new_chunk(3+ly+lgefint(*x)+nbits2lg(e)); /* HACK */
+  y = shifti(mului(u,y), e);
+  set_avma(av); return addzi(x, y);
+}
+
 /* n < 10; gerepileall supporting &NULL arguments. Maybe rename and export ? */
 INLINE void
 gc_lll(pari_sp av, int n, ...)
@@ -1192,6 +1660,45 @@ dpeM_diagonal_shallow(dpe_t **m, long d)
   return y;
 }
 
+static void
+affii_or_copy_gc(pari_sp av, GEN x, GEN *y)
+{
+  long l = lg(*y);
+  if (lgefint(x) <= l && isonstack(*y))
+  {
+    affii(x,*y);
+    set_avma(av);
+  }
+  else
+    *y = gerepileuptoint(av, x);
+}
+
+/* *x -= u*y */
+INLINE void
+submulziu(GEN *x, GEN y, ulong u)
+{
+  pari_sp av;
+  long ly = lgefint(y);
+  if (ly == 2) return;
+  av = avma;
+  (void)new_chunk(3+ly+lgefint(*x)); /* HACK */
+  y = mului(u,y);
+  set_avma(av); subzi(x, y);
+}
+
+/* *x += u*y */
+INLINE void
+addmulziu(GEN *x, GEN y, ulong u)
+{
+  pari_sp av;
+  long ly = lgefint(y);
+  if (ly == 2) return;
+  av = avma;
+  (void)new_chunk(3+ly+lgefint(*x)); /* HACK */
+  y = mului(u,y);
+  set_avma(av); addzi(x, y);
+}
+
 /************************** PROVED version (dpe) *************************/
 
 /* Babai's Nearest Plane algorithm (iterative).
@@ -1255,36 +1762,36 @@ Babai_dpe(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, dpe_t **mu, dpe_t *
             for (k=zeros+1; k<j; k++)
               dpe_subz(Dmael(mu,kappa,k), Dmael(mu,j,k), Dmael(mu,kappa,k));
             for (i=1; i<=n; i++)
-              gmael(B,kappa,i) = subii(gmael(B,kappa,i), gmael(B,j,i));
+              subzi(&gmael(B,kappa,i), gmael(B,j,i));
             for (i=1; i<=d; i++)
-              gmael(U,kappa,i) = subii(gmael(U,kappa,i), gmael(U,j,i));
+              subzi(&gmael(U,kappa,i), gmael(U,j,i));
             btop = avma;
             ztmp = subii(gmael(G,j,j), shifti(gmael(G,kappa,j), 1));
             ztmp = addii(gmael(G,kappa,kappa), ztmp);
-            gmael(G,kappa,kappa) = gerepileuptoint(btop, ztmp);
+            affii_or_copy_gc(btop, ztmp, &gmael(G,kappa,kappa));
             for (i=1; i<=j; i++)
-              gmael(G,kappa,i) = subii(gmael(G,kappa,i), gmael(G,j,i));
+              subzi(&gmael(G,kappa,i), gmael(G,j,i));
             for (i=j+1; i<kappa; i++)
-              gmael(G,kappa,i) = subii(gmael(G,kappa,i), gmael(G,i,j));
+              subzi(&gmael(G,kappa,i), gmael(G,i,j));
             for (i=kappa+1; i<=maxG; i++)
-              gmael(G,i,kappa) = subii(gmael(G,i,kappa), gmael(G,i,j));
+              subzi(&gmael(G,i,kappa), gmael(G,i,j));
           } else { /* otherwise X = -1 */
             for (k=zeros+1; k<j; k++)
               dpe_addz(Dmael(mu,kappa,k), Dmael(mu,j,k), Dmael(mu,kappa,k));
             for (i=1; i<=n; i++)
-              gmael(B,kappa,i) = addii(gmael(B,kappa,i),gmael(B,j,i));
+              addzi(&gmael(B,kappa,i),gmael(B,j,i));
             for (i=1; i<=d; i++)
-              gmael(U,kappa,i) = addii(gmael(U,kappa,i),gmael(U,j,i));
+              addzi(&gmael(U,kappa,i),gmael(U,j,i));
             btop = avma;
             ztmp = addii(gmael(G,j,j), shifti(gmael(G,kappa,j), 1));
             ztmp = addii(gmael(G,kappa,kappa), ztmp);
-            gmael(G,kappa,kappa) = gerepileuptoint(btop, ztmp);
+            affii_or_copy_gc(btop, ztmp, &gmael(G,kappa,kappa));
             for (i=1; i<=j; i++)
-              gmael(G,kappa,i) = addii(gmael(G,kappa,i), gmael(G,j,i));
+              addzi(&gmael(G,kappa,i), gmael(G,j,i));
             for (i=j+1; i<kappa; i++)
-              gmael(G,kappa,i) = addii(gmael(G,kappa,i), gmael(G,i,j));
+              addzi(&gmael(G,kappa,i), gmael(G,i,j));
             for (i=kappa+1; i<=maxG; i++)
-              gmael(G,i,kappa) = addii(gmael(G,i,kappa), gmael(G,i,j));
+              addzi(&gmael(G,i,kappa), gmael(G,i,j));
           }
           continue;
         }
@@ -1297,19 +1804,19 @@ Babai_dpe(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, dpe_t **mu, dpe_t *
             for (k=zeros+1; k<j; k++)
               dpe_submuluz(Dmael(mu,kappa,k), Dmael(mu,j,k), xx, Dmael(mu,kappa,k));
             for (i=1; i<=n; i++)
-              gmael(B,kappa,i) = submuliu_inplace(gmael(B,kappa,i), gmael(B,j,i), xx);
+              submulziu(&gmael(B,kappa,i), gmael(B,j,i), xx);
             for (i=1; i<=d; i++)
-              gmael(U,kappa,i) = submuliu_inplace(gmael(U,kappa,i), gmael(U,j,i), xx);
+              submulziu(&gmael(U,kappa,i), gmael(U,j,i), xx);
             btop = avma;
             ztmp = submuliu2n(mulii(gmael(G,j,j), sqru(xx)), gmael(G,kappa,j), xx, 1);
             ztmp = addii(gmael(G,kappa,kappa), ztmp);
-            gmael(G,kappa,kappa) = gerepileuptoint(btop, ztmp);
+            affii_or_copy_gc(btop, ztmp, &gmael(G,kappa,kappa));
             for (i=1; i<=j; i++)
-              gmael(G,kappa,i) = submuliu_inplace(gmael(G,kappa,i), gmael(G,j,i), xx);
+              submulziu(&gmael(G,kappa,i), gmael(G,j,i), xx);
             for (i=j+1; i<kappa; i++)
-              gmael(G,kappa,i) = submuliu_inplace(gmael(G,kappa,i), gmael(G,i,j), xx);
+              submulziu(&gmael(G,kappa,i), gmael(G,i,j), xx);
             for (i=kappa+1; i<=maxG; i++)
-              gmael(G,i,kappa) = submuliu_inplace(gmael(G,i,kappa), gmael(G,i,j), xx);
+              submulziu(&gmael(G,i,kappa), gmael(G,i,j), xx);
           }
           else
           {
@@ -1317,19 +1824,19 @@ Babai_dpe(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, dpe_t **mu, dpe_t *
             for (k=zeros+1; k<j; k++)
               dpe_addmuluz(Dmael(mu,kappa,k), Dmael(mu,j,k), xx, Dmael(mu,kappa,k));
             for (i=1; i<=n; i++)
-              gmael(B,kappa,i) = addmuliu_inplace(gmael(B,kappa,i), gmael(B,j,i), xx);
+              addmulziu(&gmael(B,kappa,i), gmael(B,j,i), xx);
             for (i=1; i<=d; i++)
-              gmael(U,kappa,i) = addmuliu_inplace(gmael(U,kappa,i), gmael(U,j,i), xx);
+              addmulziu(&gmael(U,kappa,i), gmael(U,j,i), xx);
             btop = avma;
             ztmp = addmuliu2n(mulii(gmael(G,j,j), sqru(xx)), gmael(G,kappa,j), xx, 1);
             ztmp = addii(gmael(G,kappa,kappa), ztmp);
-            gmael(G,kappa,kappa) = gerepileuptoint(btop, ztmp);
+            affii_or_copy_gc(btop, ztmp, &gmael(G,kappa,kappa));
             for (i=1; i<=j; i++)
-              gmael(G,kappa,i) = addmuliu_inplace(gmael(G,kappa,i), gmael(G,j,i), xx);
+              addmulziu(&gmael(G,kappa,i), gmael(G,j,i), xx);
             for (i=j+1; i<kappa; i++)
-              gmael(G,kappa,i) = addmuliu_inplace(gmael(G,kappa,i), gmael(G,i,j), xx);
+              addmulziu(&gmael(G,kappa,i), gmael(G,i,j), xx);
             for (i=kappa+1; i<=maxG; i++)
-              gmael(G,i,kappa) = addmuliu_inplace(gmael(G,i,kappa), gmael(G,i,j), xx);
+              addmulziu(&gmael(G,i,kappa), gmael(G,i,j), xx);
           }
         }
         else
@@ -1346,20 +1853,20 @@ Babai_dpe(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, dpe_t **mu, dpe_t *
               dpe_subz(Dmael(mu,kappa,k), &x, Dmael(mu,kappa,k));
             }
             for (i=1; i<=n; i++)
-              gmael(B,kappa,i) = submuliu2n(gmael(B,kappa,i), gmael(B,j,i), xx, e);
+              submulzu2n(&gmael(B,kappa,i), gmael(B,j,i), xx, e);
             for (i=1; i<=d; i++)
-              gmael(U,kappa,i) = submuliu2n(gmael(U,kappa,i), gmael(U,j,i), xx, e);
+              submulzu2n(&gmael(U,kappa,i), gmael(U,j,i), xx, e);
             btop = avma;
             ztmp = submuliu2n(mulshift(gmael(G,j,j), sqru(xx), 2*e),
                 gmael(G,kappa,j), xx, e+1);
             ztmp = addii(gmael(G,kappa,kappa), ztmp);
-            gmael(G,kappa,kappa) = gerepileuptoint(btop, ztmp);
+            affii_or_copy_gc(btop, ztmp, &gmael(G,kappa,kappa));
             for (i=1; i<=j; i++)
-              gmael(G,kappa,i) = submuliu2n(gmael(G,kappa,i), gmael(G,j,i), xx, e);
+              submulzu2n(&gmael(G,kappa,i), gmael(G,j,i), xx, e);
             for (   ; i<kappa; i++)
-              gmael(G,kappa,i) = submuliu2n(gmael(G,kappa,i), gmael(G,i,j), xx, e);
+              submulzu2n(&gmael(G,kappa,i), gmael(G,i,j), xx, e);
             for (i=kappa+1; i<=maxG; i++)
-              gmael(G,i,kappa) = submuliu2n(gmael(G,i,kappa), gmael(G,i,j), xx, e);
+              submulzu2n(&gmael(G,i,kappa), gmael(G,i,j), xx, e);
           } else
           {
             ulong xx = (ulong) pari_rint(ldexp(-tmp->d, BITS_IN_LONG - 1));
@@ -1371,20 +1878,20 @@ Babai_dpe(pari_sp av, long kappa, GEN *pG, GEN *pB, GEN *pU, dpe_t **mu, dpe_t *
               dpe_addz(Dmael(mu,kappa,k), &x, Dmael(mu,kappa,k));
             }
             for (i=1; i<=n; i++)
-              gmael(B,kappa,i) = addmuliu2n(gmael(B,kappa,i), gmael(B,j,i), xx, e);
+              addmulzu2n(&gmael(B,kappa,i), gmael(B,j,i), xx, e);
             for (i=1; i<=d; i++)
-              gmael(U,kappa,i) = addmuliu2n(gmael(U,kappa,i), gmael(U,j,i), xx, e);
+              addmulzu2n(&gmael(U,kappa,i), gmael(U,j,i), xx, e);
             btop = avma;
             ztmp = addmuliu2n(mulshift(gmael(G,j,j), sqru(xx), 2*e),
                 gmael(G,kappa,j), xx, e+1);
             ztmp = addii(gmael(G,kappa,kappa), ztmp);
-            gmael(G,kappa,kappa) = gerepileuptoint(btop, ztmp);
+            affii_or_copy_gc(btop, ztmp, &gmael(G,kappa,kappa));
             for (i=1; i<=j; i++)
-              gmael(G,kappa,i) = addmuliu2n(gmael(G,kappa,i), gmael(G,j,i), xx, e);
+              addmulzu2n(&gmael(G,kappa,i), gmael(G,j,i), xx, e);
             for (   ; i<kappa; i++)
-              gmael(G,kappa,i) = addmuliu2n(gmael(G,kappa,i), gmael(G,i,j), xx, e);
+              addmulzu2n(&gmael(G,kappa,i), gmael(G,i,j), xx, e);
             for (i=kappa+1; i<=maxG; i++)
-              gmael(G,i,kappa) = addmuliu2n(gmael(G,i,kappa), gmael(G,i,j), xx, e);
+              addmulzu2n(&gmael(G,i,kappa), gmael(G,i,j), xx, e);
           }
         }
       }
@@ -1798,6 +2305,72 @@ fplll(GEN *pG, GEN *pB, GEN *pU, GEN *pr, double DELTA, double ETA,
   *pG = G; *pB = B; *pU = U; return zeros; /* success */
 }
 
+/* do not support LLL_KER, LLL_ALL, LLL_KEEP_FIRST */
+static GEN
+ZM2_lll_norms(GEN x, long flag, GEN *pN)
+{
+  GEN a,b,c,d;
+  GEN G, U;
+  if (flag & LLL_GRAM)
+    G = x;
+  else
+    G = gram_matrix(x);
+  a = gcoeff(G,1,1); b = shifti(gcoeff(G,1,2),1); c = gcoeff(G,2,2);
+  d = qfb_disc3(a,b,c);
+  if (signe(d)>=0) return NULL;
+  G = redimagsl2(mkqfb(a,b,c,d),&U);
+  if (pN) (void) RgM_gram_schmidt(G, pN);
+  if (flag & LLL_INPLACE) return ZM2_mul(x,U);
+  return U;
+}
+
+static void
+fplll_flatter(GEN *pG, GEN *pB, GEN *pU, long rank, long flag)
+{
+  if (!*pG)
+  {
+    GEN T = ZM_flatter_rank(*pB, rank, flag);
+    if (*pU)
+    {
+      *pU = ZM_mul(*pU, T);
+      *pB = ZM_mul(*pB, T);
+    } else *pB = T;
+  } else
+  {
+    GEN T, G = *pG;
+    long i, j, l = lg(G);
+    for (i = 1; i < l; i++)
+      for(j = 1; j < i; j++)
+        gmael(G,j,i) = gmael(G,i,j);
+    T = ZM_flattergram_rank(G, rank, flag);
+    if (*pU) *pU = ZM_mul(*pU, T);
+    *pG = qf_ZM_apply(*pG, T);
+  }
+}
+
+static GEN
+get_gramschmidt(GEN M, long rank)
+{
+  GEN B, Q, L;
+  long r = lg(M)-1, bitprec = 3*r + 30;
+  long prec = nbits2prec64(bitprec);
+  if (rank < r) M = vconcat(gshift(M,1), matid(r));
+  if (!QR_init(RgM_gtofp(M, prec), &B, &Q, &L, prec) || !gsisinv(L)) return NULL;
+  return L;
+}
+
+static GEN
+get_gaussred(GEN M, long rank)
+{
+  pari_sp ltop = avma;
+  long r = lg(M)-1, bitprec = 3*r + 30, prec = nbits2prec64(bitprec);
+  GEN R;
+  if (rank < r) M = RgM_Rg_add(gshift(M, 1), gen_1);
+  R = RgM_Cholesky(RgM_gtofp(M, prec), prec);
+  if (!R) return NULL;
+  return gerepilecopy(ltop, R);
+}
+
 /* Assume x a ZM, if pN != NULL, set it to Gram-Schmidt (squared) norms
  * The following modes are supported:
  * - flag & LLL_INPLACE: x a lattice basis, return x*U
@@ -1811,9 +2384,21 @@ ZM_lll_norms(GEN x, double DELTA, long flag, GEN *pN)
   pari_sp av = avma;
   const double ETA = 0.51;
   const long keepfirst = flag & LLL_KEEP_FIRST;
-  long p, zeros = -1, n = lg(x)-1;
-  GEN G, B, U;
+  long p, zeros = -1, n = lg(x)-1, is_upper, is_lower, useflatter = 0, rank;
+  GEN G, B, U, L = NULL;
   pari_timer T;
+  long thre[]={31783,34393,20894,22525,13533,1928,672,671,
+                422,506,315,313,222,205,167,154,139,138,
+                110,120,98,94,81,75,74,64,74,74,
+                79,96,112,111,105,104,96,86,84,78,75,70,66,62,62,57,56,47,45,52,50,44,48,42,36,35,35,34,40,33,34,32,36,31,
+                38,38,40,38,38,37,35,31,34,36,34,32,34,32,28,27,25,31,25,27,28,26,25,21,21,25,25,22,21,24,24,22,21,23,22,22,22,22,21,24,21,22,19,20,19,20,19,19,19,18,19,18,18,20,19,20,18,19,18,21,18,20,18,18};
+  long thsn[]={23280,30486,50077,44136,78724,15690,1801,1611,
+               981,1359,978,1042,815,866,788,775,726,712,
+               626,613,548,564,474,481,504,447,453,508,
+               705,794,1008,946,767,898,886,763,842,757,
+               725,774,639,655,705,627,635,704,511,613,
+               583,595,568,640,541,640,567,540,577,584,
+               546,509,526,572,637,746,772,743,743,742,800,708,832,768,707,692,692,768,696,635,709,694,768,719,655,569,590,644,685,623,627,720,633,636,602,635,575,631,642,647,632,656,573,511,688,640,528,616,511,559,601,620,635,688,608,768,658,582,644,704,555,673,600,601,641,661,601,670};
   if (n <= 1) return lll_trivial(x, flag);
   if (nbrows(x)==0)
   {
@@ -1821,18 +2406,61 @@ ZM_lll_norms(GEN x, double DELTA, long flag, GEN *pN)
     if (flag & (LLL_INPLACE|LLL_IM)) return cgetg(1,t_MAT);
     retmkvec2(matid(n), cgetg(1,t_MAT));
   }
-  x = RgM_shallowcopy(x);
+  if (n==2 && nbrows(x)==2  && (flag&LLL_IM) && !keepfirst)
+  {
+    U = ZM2_lll_norms(x, flag, pN);
+    if (U) return U;
+  }
   if (flag & LLL_GRAM)
-  { G = x; B = NULL; U = matid(n); }
+  { G = x; B = NULL; U = matid(n); is_upper = 0; is_lower = 0; }
   else
-  { G = NULL; B = x; U = (flag & LLL_INPLACE)? NULL: matid(n); }
+  {
+    G = NULL; B = x; U = (flag & LLL_INPLACE)? NULL: matid(n);
+    is_upper = (flag & LLL_UPPER) || ZM_is_upper(B);
+    is_lower = !B || is_upper || keepfirst ? 0: ZM_is_lower(B);
+    if (is_lower) L = RgM_flip(B);
+  }
+  rank = (flag&LLL_NOFLATTER) ? 0: ZM_rank(x);
+  if (n > 2 && !(flag&LLL_NOFLATTER))
+  {
+    GEN R = B ? (is_upper ? B : (is_lower ? L : get_gramschmidt(B, rank)))
+              : get_gaussred(G, rank);
+    if (R)
+    {
+      long spr = spread(R), sz = mpexpo(gsupnorm(R, DEFAULTPREC)), thr;
+      if (DEBUGLEVEL>=5) err_printf("LLL: dim %ld, size %ld, spread %ld\n",n, sz, spr);
+      if ((is_upper && ZM_is_knapsack(B)) || (is_lower && ZM_is_knapsack(L)))
+        thr = thsn[minss(n-3,numberof(thsn)-1)];
+      else
+      {
+        thr = thre[minss(n-3,numberof(thre)-1)];
+        if (n>=10) sz = spr;
+      }
+      useflatter = sz >= thr;
+    } else
+      useflatter = 1;
+  } else useflatter = 0;
   if(DEBUGLEVEL>=4) timer_start(&T);
+  if (useflatter)
+  {
+    if (is_lower)
+    {
+      fplll_flatter(&G, &L, &U, rank, flag | LLL_UPPER);
+      B = RgM_flop(L);
+      if (U) U = RgM_flop(U);
+    }
+    else
+      fplll_flatter(&G, &B, &U, rank, flag | (is_upper? LLL_UPPER:0));
+    if (DEBUGLEVEL>=4  && !(flag & LLL_NOCERTIFY))
+      timer_printf(&T, "FLATTER");
+  }
   if (!(flag & LLL_GRAM))
   {
     long t;
+    B = gcopy(B);
     if(DEBUGLEVEL>=4)
-      err_printf("Entering L^2 (double): LLL-parameters (%.3f,%.3f)\n",
-                 DELTA,ETA);
+      err_printf("Entering L^2 (double): dim %ld, LLL-parameters (%.3f,%.3f)\n",
+                 n, DELTA,ETA);
     zeros = fplll_fast(&B, &U, DELTA, ETA, keepfirst);
     if (DEBUGLEVEL>=4) timer_printf(&T, zeros < 0? "LLL (failed)": "LLL");
     for (p = DEFAULTPREC, t = 0; zeros < 0 && t < 1 ; p += EXTRAPREC64, t++)
@@ -1843,22 +2471,26 @@ ZM_lll_norms(GEN x, double DELTA, long flag, GEN *pN)
       gc_lll(av, 2, &B, &U);
       if (DEBUGLEVEL>=4) timer_printf(&T, zeros < 0? "LLL (failed)": "LLL");
     }
+  } else
+    G = gcopy(G);
+  if (zeros < 0 || !(flag & LLL_NOCERTIFY))
+  {
+    if(DEBUGLEVEL>=4)
+      err_printf("Entering L^2 (dpe): LLL-parameters (%.3f,%.3f)\n", DELTA,ETA);
+    zeros = fplll_dpe(&G, &B, &U, pN, DELTA, ETA, keepfirst);
+    if (DEBUGLEVEL>=4) timer_printf(&T, zeros < 0? "LLL (failed)": "LLL");
+    if (zeros < 0)
+      for (p = DEFAULTPREC;; p += EXTRAPREC64)
+      {
+        if (DEBUGLEVEL>=4)
+          err_printf("Entering L^2: LLL-parameters (%.3f,%.3f), prec = %d\n",
+              DELTA,ETA, p);
+        zeros = fplll(&G, &B, &U, pN, DELTA, ETA, keepfirst, p);
+        if (DEBUGLEVEL>=4) timer_printf(&T, zeros < 0? "LLL (failed)": "LLL");
+        if (zeros >= 0) break;
+        gc_lll(av, 3, &G, &B, &U);
+      }
   }
-  if(DEBUGLEVEL>=4)
-    err_printf("Entering L^2 (dpe): LLL-parameters (%.3f,%.3f)\n", DELTA,ETA);
-  zeros = fplll_dpe(&G, &B, &U, pN, DELTA, ETA, keepfirst);
-  if (DEBUGLEVEL>=4) timer_printf(&T, zeros < 0? "LLL (failed)": "LLL");
-  if (zeros < 0)
-    for (p = DEFAULTPREC;; p += EXTRAPREC64)
-    {
-      if (DEBUGLEVEL>=4)
-        err_printf("Entering L^2: LLL-parameters (%.3f,%.3f), prec = %d\n",
-                    DELTA,ETA, p);
-      zeros = fplll(&G, &B, &U, pN, DELTA, ETA, keepfirst, p);
-      if (DEBUGLEVEL>=4) timer_printf(&T, zeros < 0? "LLL (failed)": "LLL");
-      if (zeros >= 0) break;
-      gc_lll(av, 3, &G, &B, &U);
-    }
   return lll_finish(U? U: B, zeros, flag);
 }
 
@@ -2013,14 +2645,12 @@ lllgramallgen(GEN x, long flag)
   return lll_finish(h,k-1,flag);
 }
 
-static int
-RgM_square(GEN x) { long l = lg(x); return l == 1 || l == lgcols(x); }
 static GEN
 lllallgen(GEN x, long flag)
 {
   pari_sp av = avma;
   if (!(flag & LLL_GRAM)) x = gram_matrix(x);
-  else if (!RgM_square(x)) pari_err_DIM("qflllgram");
+  else if (!RgM_is_square_mat(x)) pari_err_DIM("qflllgram");
   return gerepilecopy(av, lllgramallgen(x, flag));
 }
 GEN
@@ -2041,11 +2671,11 @@ GEN
 lllkerim(GEN x) { return lllall(x, LLL_ALL); }
 GEN
 lllgramint(GEN x)
-{ if (!RgM_square(x)) pari_err_DIM("qflllgram");
+{ if (!RgM_is_square_mat(x)) pari_err_DIM("qflllgram");
   return lllall(x, LLL_IM | LLL_GRAM); }
 GEN
 lllgramkerim(GEN x)
-{ if (!RgM_square(x)) pari_err_DIM("qflllgram");
+{ if (!RgM_is_square_mat(x)) pari_err_DIM("qflllgram");
   return lllall(x, LLL_ALL | LLL_GRAM); }
 
 GEN
@@ -2055,7 +2685,16 @@ lllfp(GEN x, double D, long flag)
   pari_sp av = avma;
   GEN h;
   if (n <= 1) return lll_trivial(x,flag);
-  if ((flag & LLL_GRAM) && !RgM_square(x)) pari_err_DIM("qflllgram");
+  if (flag & LLL_GRAM)
+  {
+    if (!RgM_is_square_mat(x)) pari_err_DIM("qflllgram");
+    if (isinexact(x))
+    {
+      x = RgM_Cholesky(x, gprecision(x));
+      if (!x) return NULL;
+      flag &= ~LLL_GRAM;
+    }
+  }
   h = ZM_lll(RgM_rescale_to_int(x), D, flag);
   return gerepilecopy(av, h);
 }
@@ -2065,6 +2704,14 @@ lllgram(GEN x) { return lllfp(x,LLLDFT,LLL_GRAM|LLL_IM); }
 GEN
 lll(GEN x) { return lllfp(x,LLLDFT,LLL_IM); }
 
+static GEN
+qflllgram(GEN x)
+{
+  GEN T = lllgram(x);
+  if (!T) pari_err_PREC("qflllgram");
+  return T;
+}
+
 GEN
 qflll0(GEN x, long flag)
 {
@@ -2072,7 +2719,7 @@ qflll0(GEN x, long flag)
   switch(flag)
   {
     case 0: return lll(x);
-    case 1: RgM_check_ZM(x,"qflll"); return lllint(x);
+    case 1: return lllfp(x, LLLDFT, LLL_IM | LLL_NOFLATTER);
     case 2: RgM_check_ZM(x,"qflll"); return lllintpartial(x);
     case 3: RgM_check_ZM(x,"qflll"); return lllall(x, LLL_INPLACE);
     case 4: RgM_check_ZM(x,"qflll"); return lllkerim(x);
@@ -2089,8 +2736,8 @@ qflllgram0(GEN x, long flag)
   if (typ(x) != t_MAT) pari_err_TYPE("qflllgram",x);
   switch(flag)
   {
-    case 0: return lllgram(x);
-    case 1: RgM_check_ZM(x,"qflllgram"); return lllgramint(x);
+    case 0: return qflllgram(x);
+    case 1: return lllfp(x, LLLDFT, LLL_IM | LLL_GRAM | LLL_NOFLATTER);
     case 4: RgM_check_ZM(x,"qflllgram"); return lllgramkerim(x);
     case 5: return lllgramkerimgen(x);
     case 8: return lllgramgen(x);
