@@ -1542,10 +1542,9 @@ FpX_FpXY_eval_resultant(GEN a, GEN b, GEN n, GEN p, GEN la, long db, long vX)
 /* assume dres := deg(Res_X(a,b), Y) <= deg(a,X) * deg(b,Y) < p */
 /* Return a Fly */
 static GEN
-Flx_FlxY_resultant_polint(GEN a, GEN b, ulong p, long dres, long sx)
+Flx_FlxY_resultant_polint(GEN a, GEN b, ulong p, ulong pi, long dres, long sx)
 {
   long i;
-  ulong pi = SMALL_ULONG(p)? 0: get_Fl_red(p);
   ulong n, la = Flx_lead(a);
   GEN  x = cgetg(dres+2, t_VECSMALL);
   GEN  y = cgetg(dres+2, t_VECSMALL);
@@ -1592,7 +1591,7 @@ FlxX_pseudorem(GEN x, GEN y, ulong p, ulong pi)
   }
   if (dx < 0) return zero_Flx(0);
   lx = dx+3; x -= 2;
-  x[0]=evaltyp(t_POL) | evallg(lx);
+  x[0]=evaltyp(t_POL) | _evallg(lx);
   x[1]=evalsigne(1) | evalvarn(vx);
   x = RgX_recip_i(x);
   if (dp)
@@ -1669,17 +1668,20 @@ FlxY_to_FlyX(GEN b, long sv)
 
 /* Return a Fly*/
 GEN
-Flx_FlxY_resultant(GEN a, GEN b, ulong pp)
+Flx_FlxY_resultant(GEN a, GEN b, ulong p)
 {
   pari_sp ltop=avma;
   long dres = degpol(a)*degpol(b);
   long sx=a[1], sy=b[1]&VARNBITS;
   GEN z;
   b = FlxY_to_FlyX(b,sx);
-  if ((ulong)dres >= pp)
-    z = FlxX_resultant(Fly_to_FlxY(a, sy), b, pp, sx);
+  if ((ulong)dres >= p)
+    z = FlxX_resultant(Fly_to_FlxY(a, sy), b, p, sx);
   else
-    z = Flx_FlxY_resultant_polint(a, b, pp, (ulong)dres, sy);
+  {
+    ulong pi = SMALL_ULONG(p)? 0: get_Fl_red(p);
+    z = Flx_FlxY_resultant_polint(a, b, p, pi, (ulong)dres, sy);
+  }
   return gerepileupto(ltop,z);
 }
 
@@ -1739,55 +1741,60 @@ FpX_FpXY_resultant(GEN a, GEN b, GEN p)
   return FpV_polint(x,y, p, vY);
 }
 
-static GEN
+GEN
 FpX_composedsum(GEN P, GEN Q, GEN p)
 {
-  long n = 1+ degpol(P)*degpol(Q);
-  GEN Pl = FpX_invLaplace(FpX_Newton(P,n,p), p);
-  GEN Ql = FpX_invLaplace(FpX_Newton(Q,n,p), p);
-  GEN L = FpX_Laplace(FpXn_mul(Pl, Ql, n, p), p);
-  GEN lead = Fp_mul(Fp_powu(leading_coeff(P),degpol(Q), p),
-                    Fp_powu(leading_coeff(Q),degpol(P), p), p);
-  GEN R = FpX_fromNewton(L, p);
-  return FpX_Fp_mul(R, lead, p);
+  pari_sp av = avma;
+  if (lgefint(p)==3)
+  {
+    ulong pp = p[2];
+    GEN z = Flx_composedsum(ZX_to_Flx(P, pp), ZX_to_Flx(Q, pp), pp);
+    return gerepileupto(av, Flx_to_ZX(z));
+  }
+  else
+  {
+    long n = 1+ degpol(P)*degpol(Q);
+    GEN Pl = FpX_invLaplace(FpX_Newton(P,n,p), p);
+    GEN Ql = FpX_invLaplace(FpX_Newton(Q,n,p), p);
+    GEN L = FpX_Laplace(FpXn_mul(Pl, Ql, n, p), p);
+    GEN lead = Fp_mul(Fp_powu(leading_coeff(P),degpol(Q), p),
+        Fp_powu(leading_coeff(Q),degpol(P), p), p);
+    GEN R = FpX_fromNewton(L, p);
+    return gerepileupto(av, FpX_Fp_mul(R, lead, p));
+  }
 }
 
-#if 0
 GEN
 FpX_composedprod(GEN P, GEN Q, GEN p)
 {
-  long n = 1+ degpol(P)*degpol(Q);
-  GEN L=FpX_convol(FpX_Newton(P,n,p), FpX_Newton(Q,n,p), p);
-  return FpX_fromNewton(L, p);
-}
-#endif
-
-GEN
-FpX_direct_compositum(GEN a, GEN b, GEN p)
-{
+  pari_sp av = avma;
   if (lgefint(p)==3)
   {
-    pari_sp av = avma;
     ulong pp = p[2];
-    GEN z = Flx_direct_compositum(ZX_to_Flx(a, pp), ZX_to_Flx(b, pp), pp);
+    GEN z = Flx_composedprod(ZX_to_Flx(P, pp), ZX_to_Flx(Q, pp), pp);
     return gerepileupto(av, Flx_to_ZX(z));
   }
-  return FpX_composedsum(a, b, p);
+  else
+  {
+    long n = 1+ degpol(P)*degpol(Q);
+    GEN L = FpX_convol(FpX_Newton(P,n,p), FpX_Newton(Q,n,p), p);
+    return gerepileupto(av,FpX_fromNewton(L, p));
+  }
 }
 
 static GEN
-_FpX_direct_compositum(void *E, GEN a, GEN b)
-{ return FpX_direct_compositum(a,b, (GEN)E); }
+_FpX_composedsum(void *E, GEN a, GEN b)
+{ return FpX_composedsum(a,b, (GEN)E); }
 
 GEN
-FpXV_direct_compositum(GEN V, GEN p)
+FpXV_composedsum(GEN V, GEN p)
 {
   if (lgefint(p)==3)
   {
     ulong pp = p[2];
-    return Flx_to_ZX(FlxV_direct_compositum(ZXV_to_FlxV(V, pp), pp));
+    return Flx_to_ZX(FlxV_composedsum(ZXV_to_FlxV(V, pp), pp));
   }
-  return gen_product(V, (void *)p, &_FpX_direct_compositum);
+  return gen_product(V, (void *)p, &_FpX_composedsum);
 }
 
 /* 0, 1, -1, 2, -2, ... */
@@ -1969,7 +1976,7 @@ ZXQ_charpoly_sqf(GEN A, GEN T, long *lambda, long v)
       return gerepileupto(av, gpowgs(gsub(pol_x(v), A), degpol(T)));
   }
   delvar = 0;
-  if (varn(T) == 0)
+  if (varncmp(varn(T), 0) <= 0)
   {
     long v0 = fetch_var(); delvar = 1;
     T = leafcopy(T); setvarn(T,v0);
@@ -1986,7 +1993,7 @@ ZXQ_charpoly_sqf(GEN A, GEN T, long *lambda, long v)
 GEN
 ZXQ_charpoly(GEN A, GEN T, long v)
 {
-  return (degpol(T) < 16) ? RgXQ_charpoly(A,T,v): ZXQ_charpoly_sqf(A,T, NULL, v);
+  return (degpol(T) < 16) ? RgXQ_charpoly_i(A,T,v): ZXQ_charpoly_sqf(A,T, NULL, v);
 }
 
 GEN
@@ -2072,12 +2079,25 @@ ZX_resultant_all(GEN A, GEN B, GEN dB, ulong bound)
   pari_sp av = avma;
   forprime_t S;
   GEN  H, worker;
+  if (!B && degpol(A)==2)
+  {
+    GEN a = gel(A,4), b = gel(A,3), c = gel(A,2);
+    H = mulii(a, subii(shifti(mulii(a, c), 2), sqri(b)));
+    if (dB) H = diviiexact(H, sqri(dB));
+    return gerepileuptoint(av, H);
+  }
   if (B)
   {
     long a = degpol(A), b = degpol(B);
     if (a < 0 || b < 0) return gen_0;
     if (!a) return powiu(gel(A,2), b);
     if (!b) return powiu(gel(B,2), a);
+    if (minss(a, b) <= 1)
+    {
+      H = RgX_resultant_all(A, B, NULL);
+      if (dB) H = diviiexact(H, powiu(dB, a));
+      return gerepileuptoint(av, H);
+    }
     if (!bound) bound = ZX_ZXY_ResBound(A, B, dB);
   }
   worker = snm_closure(is_entry("_ZX_resultant_worker"),
@@ -2303,15 +2323,18 @@ nfX_resultant(GEN nf, GEN x, GEN y)
 {
   pari_sp av = avma;
   GEN cx, cy, D, T = nf_get_pol(nf);
-  ulong bound;
-  long dx = degpol(x), dy = degpol(y), v = varn(T);
-  if (dx < 0 || dy < 0) return pol_0(v);
-  x = Q_primitive_part(x, &cx);
-  y = Q_primitive_part(y, &cy);
-  bound = ZXQX_resultant_bound(nf, x, y);
-  D = ZXQX_resultant_all(x, y, T, NULL, bound);
-  if (cx) D = gmul(D, gpowgs(cx, degpol(y)));
-  if (cy) D = gmul(D, gpowgs(cy, degpol(x)));
+  long dx = degpol(x), dy = degpol(y);
+  if (dx < 0 || dy < 0) return gen_0;
+  x = Q_primitive_part(x, &cx); if (cx) cx = gpowgs(cx, dy);
+  y = Q_primitive_part(y, &cy); if (cy) cy = gpowgs(cy, dx);
+  if (!dx)      D = _ZXQ_powu(gel(x,2), dy, T);
+  else if (!dy) D = _ZXQ_powu(gel(y,2), dx, T);
+  else
+  {
+    ulong bound = ZXQX_resultant_bound(nf, x, y);
+    D = ZXQX_resultant_all(x, y, T, NULL, bound);
+  }
+  cx = mul_content(cx, cy); if (cx) D = gmul(D, cx);
   return gerepileupto(av, D);
 }
 
@@ -2660,7 +2683,8 @@ ZX_ZXY_resultant_prime(GEN a, GEN b, ulong dp, ulong p,
                        long degA, long degB, long dres, long sX)
 {
   long dropa = degA - degpol(a), dropb = degB - degpol(b);
-  GEN Hp = Flx_FlxY_resultant_polint(a, b, p, dres, sX);
+  ulong pi = SMALL_ULONG(p)? 0: get_Fl_red(p);
+  GEN Hp = Flx_FlxY_resultant_polint(a, b, p, pi, dres, sX);
   if (dropa && dropb)
     Hp = zero_Flx(sX);
   else {
@@ -2669,18 +2693,18 @@ ZX_ZXY_resultant_prime(GEN a, GEN b, ulong dp, ulong p,
       GEN c = gel(b,degB+2); /* lc(B) */
       if (odd(degB)) c = Flx_neg(c, p);
       if (!Flx_equal1(c)) {
-        c = Flx_powu(c, dropa, p);
-        if (!Flx_equal1(c)) Hp = Flx_mul(Hp, c, p);
+        c = Flx_powu_pre(c, dropa, p, pi);
+        if (!Flx_equal1(c)) Hp = Flx_mul_pre(Hp, c, p, pi);
       }
     }
     else if (dropb)
     { /* multiply by lc(A)^(deg B - deg b) */
       ulong c = uel(a, degA+2); /* lc(A) */
       c = Fl_powu(c, dropb, p);
-      if (c != 1) Hp = Flx_Fl_mul(Hp, c, p);
+      if (c != 1) Hp = Flx_Fl_mul_pre(Hp, c, p, pi);
     }
   }
-  if (dp != 1) Hp = Flx_Fl_mul(Hp, Fl_powu(Fl_inv(dp,p), degA, p), p);
+  if (dp != 1) Hp = Flx_Fl_mul_pre(Hp, Fl_powu_pre(Fl_inv(dp,p), degA, p, pi), p, pi);
   return Hp;
 }
 
@@ -2804,7 +2828,7 @@ ZX_ZXY_rnfequation(GEN A, GEN B, long *lambda)
 }
 
 static GEN
-ZX_direct_compositum_slice(GEN A, GEN B, GEN P, GEN *mod)
+ZX_composedsum_slice(GEN A, GEN B, GEN P, GEN *mod)
 {
   pari_sp av = avma;
   long i, n = lg(P)-1;
@@ -2813,7 +2837,7 @@ ZX_direct_compositum_slice(GEN A, GEN B, GEN P, GEN *mod)
   {
     ulong p = uel(P,1);
     GEN a = ZX_to_Flx(A, p), b = ZX_to_Flx(B, p);
-    GEN Hp = Flx_direct_compositum(a, b, p);
+    GEN Hp = Flx_composedsum(a, b, p);
     H = gerepileupto(av, Flx_to_ZX(Hp));
     *mod = utoipos(p); return H;
   }
@@ -2825,34 +2849,33 @@ ZX_direct_compositum_slice(GEN A, GEN B, GEN P, GEN *mod)
   {
     ulong p = P[i];
     GEN a = gel(A,i), b = gel(B,i);
-    gel(H,i) = Flx_direct_compositum(a, b, p);
+    gel(H,i) = Flx_composedsum(a, b, p);
   }
   H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
   *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
-ZX_direct_compositum_worker(GEN P, GEN A, GEN B)
+ZX_composedsum_worker(GEN P, GEN A, GEN B)
 {
   GEN V = cgetg(3, t_VEC);
-  gel(V,1) = ZX_direct_compositum_slice(A, B, P, &gel(V,2));
+  gel(V,1) = ZX_composedsum_slice(A, B, P, &gel(V,2));
   return V;
 }
 
-/* Assume A,B irreducible (in particular squarefree) and define linearly
- * disjoint extensions: no factorisation needed */
 static GEN
-ZX_direct_compositum(GEN A, GEN B, GEN lead)
+ZX_composedsum_i(GEN A, GEN B, GEN lead)
 {
   pari_sp av = avma;
   forprime_t S;
   ulong bound;
   GEN H, worker, mod;
   if (degpol(A) < degpol(B)) swap(A, B);
+  if (!lead) lead  = mulii(leading_coeff(A),leading_coeff(B));
   bound = ZX_ZXY_ResBound_1(A, B);
-  worker = snm_closure(is_entry("_ZX_direct_compositum_worker"), mkvec2(A,B));
+  worker = snm_closure(is_entry("_ZX_composedsum_worker"), mkvec2(A,B));
   init_modular_big(&S);
-  H = gen_crt("ZX_direct_compositum", worker, &S, lead, bound, 0, &mod,
+  H = gen_crt("ZX_composedsum", worker, &S, lead, bound, 0, &mod,
               nxV_chinese_center, FpX_center);
   return gerepileupto(av, H);
 }
@@ -2871,7 +2894,7 @@ ZX_compositum_lambda(GEN A, GEN B, GEN lead, long lambda)
     if (DEBUGLEVEL>4) err_printf("Trying lambda = %ld\n", lambda);
     if (lead && dvdiu(lead,p)) { p = u_forprime_next(&S); continue; }
     a = ZX_to_Flx(ZX_rescale(A, stoi(-lambda)), p);
-    Hp = Flx_direct_compositum(a, ZX_to_Flx(B, p), p);
+    Hp = Flx_composedsum(a, ZX_to_Flx(B, p), p);
     if (!Flx_is_squarefree(Hp, p)) { lambda = next_lambda(lambda); continue; }
     if (DEBUGLEVEL>4) err_printf("Final lambda = %ld\n", lambda);
     return gc_long(av, lambda);
@@ -2887,15 +2910,15 @@ ZX_compositum(GEN A, GEN B, long *lambda)
     *lambda = ZX_compositum_lambda(A, B, lead, *lambda);
     A = ZX_rescale(A, stoi(-*lambda));
   }
-  return ZX_direct_compositum(A, B, lead);
+  return ZX_composedsum_i(A, B, lead);
 }
 
 GEN
-ZX_compositum_disjoint(GEN A, GEN B)
-{ return ZX_compositum(A, B, NULL); }
+ZX_composedsum(GEN A, GEN B)
+{ return ZX_composedsum_i(A, B, NULL); }
 
 static GEN
-ZXQX_direct_compositum_slice(GEN A, GEN B, GEN C, GEN P, GEN *mod)
+ZXQX_composedsum_slice(GEN A, GEN B, GEN C, GEN P, GEN *mod)
 {
   pari_sp av = avma;
   long i, n = lg(P)-1, dC = degpol(C), v = varn(C);
@@ -2905,7 +2928,7 @@ ZXQX_direct_compositum_slice(GEN A, GEN B, GEN C, GEN P, GEN *mod)
     ulong p = uel(P,1);
     GEN a = ZXX_to_FlxX(A, p, v), b = ZXX_to_FlxX(B, p, v);
     GEN c = ZX_to_Flx(C, p);
-    GEN Hp = FlxX_to_Flm(FlxqX_direct_compositum(a, b, c, p), dC);
+    GEN Hp = FlxX_to_Flm(FlxqX_composedsum(a, b, c, p), dC);
     H = gerepileupto(av, Flm_to_ZM(Hp));
     *mod = utoipos(p); return H;
   }
@@ -2918,31 +2941,31 @@ ZXQX_direct_compositum_slice(GEN A, GEN B, GEN C, GEN P, GEN *mod)
   {
     ulong p = P[i];
     GEN a = gel(A,i), b = gel(B,i), c = gel(C,i);
-    gel(H,i) = FlxX_to_Flm(FlxqX_direct_compositum(a, b, c, p), dC);
+    gel(H,i) = FlxX_to_Flm(FlxqX_composedsum(a, b, c, p), dC);
   }
   H = nmV_chinese_center_tree_seq(H, P, T, ZV_chinesetree(P, T));
   *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
-ZXQX_direct_compositum_worker(GEN P, GEN A, GEN B, GEN C)
+ZXQX_composedsum_worker(GEN P, GEN A, GEN B, GEN C)
 {
   GEN V = cgetg(3, t_VEC);
-  gel(V,1) = ZXQX_direct_compositum_slice(A, B, C, P, &gel(V,2));
+  gel(V,1) = ZXQX_composedsum_slice(A, B, C, P, &gel(V,2));
   return V;
 }
 
 static GEN
-ZXQX_direct_compositum(GEN A, GEN B, GEN T, ulong bound)
+ZXQX_composedsum(GEN A, GEN B, GEN T, ulong bound)
 {
   pari_sp av = avma;
   forprime_t S;
   GEN H, worker, mod;
   GEN lead = mulii(Q_content(leading_coeff(A)), Q_content(leading_coeff(B)));
-  worker = snm_closure(is_entry("_ZXQX_direct_compositum_worker")
+  worker = snm_closure(is_entry("_ZXQX_composedsum_worker")
                       , mkvec3(A,B,T));
   init_modular_big(&S);
-  H = gen_crt("ZXQX_direct_compositum", worker, &S, lead, bound, 0, &mod,
+  H = gen_crt("ZXQX_composedsum", worker, &S, lead, bound, 0, &mod,
               nmV_chinese_center, FpM_center);
   if (DEBUGLEVEL > 4)
     err_printf("nfcompositum: a priori bound: %lu, a posteriori: %lu\n",
@@ -2951,14 +2974,14 @@ ZXQX_direct_compositum(GEN A, GEN B, GEN T, ulong bound)
 }
 
 static long
-ZXQX_direct_compositum_bound(GEN nf, GEN A, GEN B)
+ZXQX_composedsum_bound(GEN nf, GEN A, GEN B)
 { return ZXQX_resultant_bound_i(nf, A, B, &RgX_RgXY_ResBound_1); }
 
 GEN
 nf_direct_compositum(GEN nf, GEN A, GEN B)
 {
-  ulong bnd = ZXQX_direct_compositum_bound(nf, A, B);
-  return ZXQX_direct_compositum(A, B, nf_get_pol(nf), bnd);
+  ulong bnd = ZXQX_composedsum_bound(nf, A, B);
+  return ZXQX_composedsum(A, B, nf_get_pol(nf), bnd);
 }
 
 /************************************************************************
@@ -3047,17 +3070,17 @@ ffinit_fact_Flx(ulong p, long n)
   GEN P, F = factoru_pow(n), Fp = gel(F,1), Fe = gel(F,2), Fm = gel(F,3);
   long i, l = lg(Fm);
   P = cgetg(l, t_VEC);
-  for (i = 1; i < l; ++i)
-    gel(P,i) = p==uel(Fp,i) ?
-                 ffinit_Artin_Schreier(uel(Fp,i), Fe[i])
-               : flinit(p, uel(Fm,i));
-  return FlxV_direct_compositum(P, p);
+  for (i = 1; i < l; i++)
+    gel(P,i) = p==uel(Fp,i) ? ffinit_Artin_Schreier(p, Fe[i])
+                            : flinit(p, uel(Fm,i));
+  return FlxV_composedsum(P, p);
 }
 
 static GEN
 init_Flxq_i(ulong p, long n, long sv)
 {
   GEN P;
+  if (!odd(p) && p != 2) pari_err_PRIME("ffinit", utoi(p));
   if (n == 1) return polx_Flx(sv);
   if (flinit_check(p, n+1, n))
   {
@@ -3109,7 +3132,7 @@ ffinit_fact(GEN p, long n)
     gel(P,i) = absequaliu(p, Fp[i]) ?
                  Flx_to_ZX(ffinit_Artin_Schreier(Fp[i], Fe[i]))
                : fpinit(p, Fm[i]);
-  return FpXV_direct_compositum(P, p);
+  return FpXV_composedsum(P, p);
 }
 
 static GEN
@@ -3123,6 +3146,7 @@ init_Fq_i(GEN p, long n, long v)
   if (n == 1) return pol_x(v);
   if (lgefint(p) == 3)
     return Flx_to_ZX(init_Flxq_i(p[2], n, evalvarn(v)));
+  if (!mpodd(p)) pari_err_PRIME("ffinit", p);
   if (fpinit_check(p, n+1, n)) return polcyclo(n+1, v);
   P = ffinit_fact(p,n);
   setvarn(P, v); return P;
