@@ -237,7 +237,7 @@ jbesselintern(GEN n, GEN z, long J, long prec)
       if (!(y = toser_i(z))) break;
       if (issmall(n,&ki)) n = utoi(labs(ki));
       y = gmul2n(gsqr(y),-2); if (J) y = gneg(y);
-      v = valp(y);
+      v = valser(y);
       if (v < 0) pari_err_DOMAIN(f, "valuation", "<", gen_0, z);
       if (v == 0) pari_err_IMPL(stack_strcat(f, " around a!=0"));
       m = lg(y) - 2;
@@ -321,7 +321,7 @@ jbesselh(GEN n, GEN z, long prec)
       long t, v;
       av = avma; if (!(y = toser_i(z))) break;
       if (gequal0(y)) return gerepileupto(av, gpowgs(y,k));
-      v = valp(y);
+      v = valser(y);
       if (v < 0) pari_err_DOMAIN("besseljh","valuation", "<", gen_0, z);
       t = lg(y)-2;
       if (v) y = sertoser(y, t + (2*k+1)*v);
@@ -369,7 +369,7 @@ kbessel1(GEN nu, GEN gx, long prec)
   shiftr_inplace(nu2,2); togglesign(nu2); /* nu2 = -4nu^2 */
   n = (long) (prec2nbits_mul(l,M_LN2) + M_PI*fabs(rtodbl(nu))) / 2;
   bit = prec2nbits(l) - 1;
-  l += EXTRAPRECWORD;
+  l += EXTRAPREC64;
   pi = mppi(l); n2 = n<<1; r = gmul2n(x,1);
   if (cmprs(x, n) < 0)
   {
@@ -440,7 +440,7 @@ static GEN
 _kbessel(long n, GEN x, long m, long prec)
 {
   GEN p1, p2, s, H;
-  long k, M = m + n, exact = (M <= bit_accuracy(prec));
+  long k, M = m + n, exact = (M <= prec2nbits(prec));
   pari_sp av;
 
   H = cgetg(M+2,t_VEC); gel(H,1) = gen_0;
@@ -560,7 +560,7 @@ kbesselintern(GEN n, GEN z, long N, long prec)
       {
         long v, mv, k = labs(ki), m = lg(y)-2;
         y = gmul2n(gsqr(y),-2); if (N) y = gneg(y);
-        v = valp(y);
+        v = valser(y);
         if (v < 0) pari_err_DOMAIN(f, "valuation", "<", gen_0, z);
         if (v == 0) pari_err_IMPL(stack_strcat(f, " around a!=0"));
         mv = m - (v >> 1);
@@ -1473,6 +1473,17 @@ cxerfc_r1(GEN x, long prec)
   return gerepileupto(av,res);
 }
 
+static GEN
+sererfc(GEN x, long prec)
+{
+  GEN u, z = invr(sqrtr_abs(Pi2n(-2,prec)));
+  setsigne(z, -1); /* -2/sqrt(Pi) */
+  z = gmul(z, integser(gmul(derivser(x), gexp(gneg(gsqr(x)), prec))));
+  u = polcoef_i(x, 0, varn(x));
+  if (!gcmp0(u)) z = gadd(z, gerfc(u, prec));
+  return z;
+}
+
 GEN
 gerfc(GEN x, long prec)
 {
@@ -1480,6 +1491,16 @@ gerfc(GEN x, long prec)
   long s;
   pari_sp av;
 
+  switch(typ(x))
+  {
+    case t_INT: case t_REAL: case t_FRAC: case t_COMPLEX:
+      break;
+    default:
+      av = avma;
+      if ((z = toser_i(x))) return gerepileupto(av, sererfc(z,prec));
+      return trans_eval("erfc",gerfc,x,prec);
+  }
+  /* x a complex scalar */
   x = trans_fix_arg(&prec,&x,&xr,&xi, &av,&res);
   s = signe(xr);
   if (s > 0 || (s == 0 && signe(xi) >= 0)) {
@@ -1530,17 +1551,18 @@ get_xinf(double beta)
 }
 /* optimize for zeta( s + it, prec ), assume |s-1| > 0.1
  * (if gexpo(u = s-1) < -5, we use the functional equation s->1-s) */
-static void
+static int
 optim_zeta(GEN S, long prec, long *pp, long *pn)
 {
   double s, t, alpha, beta, n, B;
   long p;
   if (typ(S) == t_REAL) {
-    s = rtodbl(S);
     t = 0.;
+    s = rtodbl(S);
   } else {
-    s = rtodbl(gel(S,1));
     t = fabs( rtodbl(gel(S,2)) );
+    if (t > 2500) return 0; /* lfunlarge */
+    s = rtodbl(gel(S,1));
   }
 
   B = prec2nbits_mul(prec, M_LN2);
@@ -1587,6 +1609,7 @@ optim_zeta(GEN S, long prec, long *pp, long *pn)
   *pp = p;
   *pn = (long)ceil(n);
   if (*pp < 0 || *pn < 0) pari_err_OVERFLOW("zeta");
+  return 1;
 }
 
 /* zeta(a*j+b), j=0..N-1, b>1, using sumalt. Johansonn's thesis, Algo 4.7.1 */
@@ -1745,7 +1768,7 @@ szeta(long k, long prec)
       B = bernfrac(k);
     }
     /* B = B_k */
-    z = gmul(powru(Pi2n(1, prec + EXTRAPRECWORD), k), B);
+    z = gmul(powru(Pi2n(1, prec + EXTRAPREC64), k), B);
     z = divrr(z, mpfactr(k,prec));
     setsigne(z, 1); shiftr_inplace(z, -1);
   }
@@ -1773,6 +1796,7 @@ zeta_funeq(GEN *ps)
   else
   {
     GEN sig = gel(s,1);
+    if (fabs(rtodbl(gel(s,2))) > 2500) return 0; /* lfunlarge */
     u = gsubsg(1, s);
     if (gexpo(u) >= -5
         && ((signe(sig) > 0 && expo(sig) >= -1) || gexpo(s) <= -5)) return 0;
@@ -1802,9 +1826,14 @@ czeta(GEN s0, long prec)
     if (!funeq_factor) { set_avma(av0); return real_1(prec); }
     return gerepileupto(av0, funeq_factor);
   }
-  optim_zeta(s, prec, &lim, &nn);
+  if (!optim_zeta(s, prec, &lim, &nn))
+  {
+    long bit = prec2nbits(prec);
+    y = lfun(lfuninit(gen_1, cgetg(1,t_VEC), 0, bit), s, bit);
+    if (funeq_factor) y = gmul(y, funeq_factor);
+    set_avma(av); return affc_fixlg(y,res);
+  }
   if (DEBUGLEVEL>2) err_printf("lim, nn: [%ld, %ld]\n", lim, nn);
-
   ms = gneg(s);
   if (umuluu_le(nn, prec, 10000000))
   {
@@ -1815,7 +1844,7 @@ czeta(GEN s0, long prec)
   }
   else
   {
-    Ns = dirpowerssum(nn, ms, prec);
+    Ns = dirpowerssum(nn, ms, 0, prec);
     incrprec(prec); /* one extra word of precision */
     ns = gpow(utor(nn, prec), ms, prec);
     y = gsub(Ns, gmul2n(ns, -1));
@@ -2026,7 +2055,7 @@ polylog(long m, GEN x, long prec)
   l = precision(x);
   if (!l) l = prec; else prec = l;
   res = cgetc(l); av = avma;
-  x = gtofp(x, l+EXTRAPRECWORD);
+  x = gtofp(x, l+EXTRAPREC64);
   e = gexpo(gnorm(x));
   if (!e || e == -1) {
     y = cxpolylog(m,x,prec);
@@ -2207,7 +2236,7 @@ gpolylog_i(void *E, GEN x, long prec)
       if (!m) { set_avma(av); return mkfrac(gen_m1,gen_2); }
       if (m==1) return gerepileupto(av, Li1(y, prec));
       if (gequal0(y)) return gerepilecopy(av, y);
-      v = valp(y);
+      v = valser(y);
       if (v < 0) pari_err_DOMAIN("polylog","valuation", "<", gen_0, x);
       if (v > 0) {
         n = (lg(y)-3 + v) / v;
@@ -2378,11 +2407,11 @@ qq(GEN x, long prec)
   return y;
 }
 
-/* return (y * X^d) + x. Assume d > 0, x != 0, valp(x) = 0 */
+/* return (y * X^d) + x. Assume d > 0, x != 0, valser(x) = 0 */
 static GEN
 ser_addmulXn(GEN y, GEN x, long d)
 {
-  long i, lx, ly, l = valp(y) + d; /* > 0 */
+  long i, lx, ly, l = valser(y) + d; /* > 0 */
   GEN z;
 
   lx = lg(x);
@@ -2464,7 +2493,7 @@ inteta(GEN q)
     long l = lg(q), v, n;
     pari_sp av;
 
-    v = valp(q); /* handle valuation separately to avoid overflow */
+    v = valser(q); /* handle valuation separately to avoid overflow */
     if (v <= 0) pari_err_DOMAIN("eta", "v_p(q)", "<=",gen_0,q);
     y = ser2pol_i(q, l); /* t_SER inefficient when input has low degree */
     n = degpol(y);
@@ -2474,7 +2503,7 @@ inteta(GEN q)
       setvarn(z, varn(y)); return RgX_to_ser(z, l+v);
     }
     q = leafcopy(q); av = avma;
-    setvalp(q, 0);
+    setvalser(q, 0);
     y = scalarser(gen_1, varn(q), l+v);
     vps = vqn = 0;
     for(n = 0;; n++)
@@ -2680,7 +2709,7 @@ ser_eta(long prec)
 {
   GEN e = cgetg(prec+2, t_SER), ed = e+2;
   long n, j;
-  e[1] = evalsigne(1)|_evalvalp(0)|evalvarn(0);
+  e[1] = evalsigne(1)|_evalvalser(0)|evalvarn(0);
   gel(ed,0) = gen_1;
   for (n = 1; n < prec; n++) gel(ed,n) = gen_0;
   for (n = 1, j = 0; n < prec; n++)
@@ -2709,7 +2738,7 @@ ser_E(long prec)
   GEN e = cgetg(prec+2, t_SER), ed = e+2;
   GEN F = vecfactoru_i(2, prec); /* F[n] = factoru(n+1) */
   long n;
-  e[1] = evalsigne(1)|_evalvalp(0)|evalvarn(0);
+  e[1] = evalsigne(1)|_evalvalser(0)|evalvarn(0);
   gel(ed,0) = utoipos(65520);
   for (n = 1; n < prec; n++) gel(ed,n) = coeffEu(gel(F,n));
   return e;
@@ -2721,7 +2750,7 @@ ser_j2(long prec, long v)
   pari_sp av = avma;
   GEN iD = gpowgs(ginv(ser_eta(prec)), 24); /* q/Delta */
   GEN J = gmul(ser_E(prec), iD);
-  setvalp(iD,-1); /* now 1/Delta */
+  setvalser(iD,-1); /* now 1/Delta */
   J = gadd(gdivgu(J, 691), iD);
   J = gerepileupto(av, J);
   if (prec > 1) gel(J,3) = utoipos(744);
@@ -2747,7 +2776,7 @@ ser_j(long prec, long v)
     gel(S5,n) = mului(21, usumdivk_fact(fa,5));
   }
   J = cgetg(prec+2, t_SER),
-  J[1] = evalvarn(v)|evalsigne(1)|evalvalp(-1);
+  J[1] = evalvarn(v)|evalsigne(1)|evalvalser(-1);
   j = J+3;
   gel(j,-1) = gen_1;
   gel(j,0) = utoipos(744);
@@ -2819,8 +2848,8 @@ jell(GEN x, long prec)
 }
 
 static GEN
-to_form(GEN a, GEN w, GEN C)
-{ return mkvec3(a, w, diviiexact(C, a)); }
+to_form(GEN a, GEN w, GEN C, GEN D)
+{ return mkqfb(a, w, diviiexact(C, a), D); }
 static GEN
 form_to_quad(GEN f, GEN sqrtD)
 {
@@ -2844,11 +2873,11 @@ double_eta_quotient(GEN a, GEN w, GEN D, long p, long q, GEN pq, GEN sqrtD)
   GEN d, t, z, zp, zq, zpq, s_t, s_tp, s_tpq, s, sp, spq;
   long prec = realprec(sqrtD);
 
-  z = eta_form(to_form(a, w, C), sqrtD, &s_t, prec);
+  z = eta_form(to_form(a, w, C, D), sqrtD, &s_t, prec);
   s = gel(s_t, 1);
-  zp = eta_form(to_form(mului(p, a), w, C), sqrtD, &s_tp, prec);
+  zp = eta_form(to_form(mului(p, a), w, C, D), sqrtD, &s_tp, prec);
   sp = gel(s_tp, 1);
-  zpq = eta_form(to_form(mulii(pq, a), w, C), sqrtD, &s_tpq, prec);
+  zpq = eta_form(to_form(mulii(pq, a), w, C, D), sqrtD, &s_tpq, prec);
   spq = gel(s_tpq, 1);
   if (p == q) {
     z = gdiv(gsqr(zp), gmul(z, zpq));
@@ -2857,7 +2886,7 @@ double_eta_quotient(GEN a, GEN w, GEN D, long p, long q, GEN pq, GEN sqrtD)
     if (sp != gen_1) z = gmul(z, sp);
   } else {
     GEN s_tq, sq;
-    zq = eta_form(to_form(mului(q, a), w, C), sqrtD, &s_tq, prec);
+    zq = eta_form(to_form(mului(q, a), w, C, D), sqrtD, &s_tq, prec);
     sq = gel(s_tq, 1);
     z = gdiv(gmul(zp, zq), gmul(z, zpq));
     t = gsub(gadd(gel(s_tp,2), gel(s_tq,2)),
@@ -2886,6 +2915,7 @@ cxanalyze(cxanalyze_t *T, GEN z)
   GEN a, b;
   long ta, tb;
 
+  T->u = z;
   T->v = 0;
   if (is_intreal_t(typ(z)))
   {
@@ -2909,7 +2939,7 @@ cxanalyze(cxanalyze_t *T, GEN z)
     T->t = gsigne(a) < 0? 4: 0;
     return 1;
   }
-  if (ta != tb || ta == t_REAL) { T->u = z; return 0; }
+  if (ta != tb || ta == t_REAL) return 0;
   /* a,b both non zero, both t_INT or t_FRAC */
   if (ta == t_INT)
   {
@@ -3216,7 +3246,7 @@ cxEk(GEN tau, long k, long prec)
   long n, b, l = precision(tau);
 
   if (l) prec = l;
-  b = bit_accuracy(prec);
+  b = prec2nbits(prec);
   /* sum n^(k-1) x^n <= x(1 + (k!-1)x) / (1-x)^k (cf Eulerian polynomials)
    * S = \sum_{n > 0} n^(k-1) |q^n/(1-q^n)| <= x(1+(k!-1)x) / (1-x)^(k+1),
    * where x = |q| = exp(-2Pi Im(tau)) < 1. Neglegt 2/zeta(1-k) * S if
