@@ -33,7 +33,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 static jmp_buf *env;
 static pari_stack s_env;
-void (*cb_gp_output)(GEN z) = NULL;
 void (*cb_pari_end_output)(void) = NULL;
 
 static void
@@ -114,23 +113,24 @@ usage(char *s)
 static void
 gp_head(void)
 {
+  ulong p, f;
   pari_print_version();
   pari_putc('\n');
-  pari_center("Copyright (C) 2000-2022 The PARI Group");
+  pari_center("Copyright (C) 2000-2024 The PARI Group");
   pari_putc('\n');
   print_text("PARI/GP is free software, covered by the GNU General Public \
 License, and comes WITHOUT ANY WARRANTY WHATSOEVER.");
   pari_puts("\nType ? for help, \\q to quit.\n");
   pari_printf("Type ?%d for how to get moral"
               " (and possibly technical) support.\n", pari_community());
+  p = GP_DATA->primelimit;
+  f = GP_DATA->factorlimit;
   if (pari_mainstack->vsize)
-    pari_printf("\nparisizemax = %lu, primelimit = %lu",
-                pari_mainstack->vsize,GP_DATA->primelimit);
+    pari_printf("\nparisizemax = %lu, primelimit = %lu, factorlimit = %lu",
+                pari_mainstack->vsize, p, f);
   else
-    pari_printf("\nparisize = %lu, primelimit = %lu",
-                pari_mainstack->rsize,GP_DATA->primelimit);
-  if (pari_mt_nbthreads > 1)
-    pari_printf(", nbthreads = %lu", pari_mt_nbthreads);
+    pari_printf("\nparisize = %lu, primelimit = %lu, factorlimit = %lu",
+                pari_mainstack->rsize, p, f);
   pari_putc('\n');
 }
 
@@ -138,7 +138,7 @@ static char *
 read_arg(long *nread, char *t, long argc, char **argv)
 {
   long i = *nread;
-  if (isdigit((int)*t)) return t;
+  if (isdigit((unsigned char)*t)) return t;
   if (*t || i==argc) usage(argv[0]);
   *nread = i+1; return argv[i];
 }
@@ -147,7 +147,7 @@ static char *
 read_arg_equal(long *nread, char *t, long argc, char **argv)
 {
   long i = *nread;
-  if (*t=='=' && isdigit((int)t[1])) return t+1;
+  if (*t=='=' && isdigit((unsigned char)t[1])) return t+1;
   if (*t || i==argc) usage(argv[0]);
   *nread = i+1; return argv[i];
 }
@@ -275,61 +275,17 @@ static int
 test_is_interactive(void) { return 0; }
 
 static void
-test_output(GEN z) { init_linewrap(76); gen_output(z); }
+test_output(long n) { init_linewrap(76); gen_output(pari_get_hist(n)); }
 void
 init_test(void)
 {
   disable_color = 1;
   init_linewrap(76);
   pari_errfile = stdout;
-  cb_gp_output = test_output;
+  cb_pari_display_hist = test_output;
   cb_pari_is_interactive = test_is_interactive;
 }
 
-/*******************************************************************/
-/**                                                               **/
-/**                   FORMAT GP OUTPUT                            **/
-/**                                                               **/
-/*******************************************************************/
-    /* REGULAR */
-static void
-normal_output(GEN z, long n)
-{
-  long l = 0;
-  char *s;
-  /* history number */
-  if (n)
-  {
-    char buf[64];
-    if (!(GP_DATA->flags & gpd_QUIET))
-    {
-      term_color(c_HIST);
-      sprintf(buf, "%%%ld = ", n);
-      pari_puts(buf);
-      l = strlen(buf);
-    }
-  }
-  /* output */
-  term_color(c_OUTPUT);
-  s = GENtostr(z);
-  if (GP_DATA->lim_lines)
-    lim_lines_output(s, l, GP_DATA->lim_lines);
-  else
-    pari_puts(s);
-  pari_free(s);
-  term_color(c_NONE); pari_putc('\n');
-}
-
-static void
-gp_output(GEN z)
-{
-  if (cb_gp_output) { cb_gp_output(z); return; }
-  if (GP_DATA->fmt->prettyp == f_PRETTY)
-  { if (tex2mail_output(z, GP_DATA->hist->total)) return; }
-  normal_output(z, GP_DATA->hist->total);
-  pari_flush();
-}
-
 static GEN
 gp_main_loop(long ismain)
 {
@@ -399,7 +355,8 @@ gp_main_loop(long ismain)
     }
     if (GP_DATA->simplify) z = simplify_shallow(z);
     pari_add_hist(z, t, r);
-    if (z != gnil && ! is_silent(b->buf) ) gp_output(z);
+    if (z != gnil && ! is_silent(b->buf) )
+      gp_display_hist(GP_DATA->hist->total);
     set_avma(av);
     parivstack_reset();
   }
@@ -427,8 +384,7 @@ read_main(const char *s)
 static const char *
 break_loop_prompt(long n)
 {
-  const char *s = (n==1)? "break> ": stack_sprintf("break[%ld]> ", n);
-  return gp_format_prompt(s);
+  return n==0 ? "GP prompt" : n==1? "break> ": stack_sprintf("break[%ld]> ", n);
 }
 
 static long frame_level=0, dbg_level = 0;
@@ -450,16 +406,17 @@ break_loop(int numerr)
 
   b = filtered_buffer(&F);
   nenv=pari_stack_new(&s_env);
-  prompt = break_loop_prompt(s_env.n-1);
+  prompt = gp_format_prompt(break_loop_prompt(nenv));
   iferr_env = NULL;
   dbg_level = 0;
   frame_level = closure_context(oldframe_level, dbg_level);
   pari_infile = newfile(stdin, "stdin", mf_IN)->file;
   term_color(c_ERR); pari_putc('\n');
   if (sigint)
-    msg = "Break loop: <Return> to continue; 'break' to go back to GP prompt";
+    msg = "Break loop: <Return> to continue; 'break' to go back to %s";
   else
-    msg = "Break loop: type 'break' to go back to GP prompt";
+    msg = "Break loop: type 'break' to go back to %s";
+  msg = stack_sprintf(msg, break_loop_prompt(nenv-1));
   print_errcontext(pariOut, msg, NULL, NULL);
   term_color(c_NONE);
   av = avma;
@@ -579,7 +536,7 @@ main(int argc, char **argv)
   /* must be defined here in case an error is raised in pari_init_opts, e.g.
    * when parsing function prototypes */
   cb_pari_err_recover = gp_err_recover;
-  pari_init_opts(1000000 * sizeof(long), 0, INIT_SIGm | INIT_noPRIMEm | INIT_noIMTm);
+  pari_init_opts(8000000, 0, INIT_SIGm | INIT_noPRIMEm | INIT_noIMTm);
   cb_pari_pre_recover = gp_pre_recover;
   cb_pari_break_loop = break_loop;
   cb_pari_is_interactive = is_interactive;
@@ -628,15 +585,6 @@ main(int argc, char **argv)
   return 0; /* LCOV_EXCL_LINE */
 }
 
-void
-pari_breakpoint(void)
-{
-  if (!pari_last_was_newline()) pari_putc('\n');
-  closure_err(0);
-  if (cb_pari_break_loop && cb_pari_break_loop(-1)) return;
-  cb_pari_err_recover(e_MISC);
-}
-
 void
 dbg_down(long k)
 {
