@@ -12,6 +12,7 @@ Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
+
 /*********************************************************************/
 /*                       MALLOC/FREE WRAPPERS                        */
 /*********************************************************************/
@@ -157,7 +158,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
        gel(_v,3) = (z);\
        gel(_v,4) = (t);\
        gel(_v,5) = (u); return _v; } while(0)
-
+#define retmkmat22(a,b,c,d)\
+  do { GEN _v = cgetg(3, t_MAT);\
+       gel(_v,1) = mkcol2(a,c);\
+       gel(_v,2) = mkcol2(b,d); return _v; } while(0)
 INLINE GEN
 mkintmod(GEN x, GEN y) { retmkintmod(x,y); }
 INLINE GEN
@@ -422,7 +426,7 @@ INLINE GEN
 zeroser(long v, long e)
 {
   GEN x = cgetg(2, t_SER);
-  x[1] = evalvalp(e) | evalvarn(v); return x;
+  x[1] = evalvalser(e) | evalvarn(v); return x;
 }
 INLINE int
 ser_isexactzero(GEN x)
@@ -1083,6 +1087,9 @@ INLINE GEN
 vecpermute(GEN A, GEN x)
 { pari_APPLY_type(typ(A), gel(A, x[i])) }
 
+INLINE GEN
+veclast(GEN A) { return gel(A, lg(A)-1); }
+
 INLINE GEN
 rowpermute(GEN x, GEN p)
 { pari_APPLY_same(typ(gel(x,i)) == t_VECSMALL ? vecsmallpermute(gel(x, i), p)
@@ -1204,7 +1211,7 @@ pari_calloc(size_t size)
   memset(t, 0, size); return t;
 }
 INLINE GEN
-cgetalloc(long t, size_t l)
+cgetalloc(size_t l, long t)
 { /* evallg may raise e_OVERFLOW, which would leak x */
   ulong x0 = evaltyp(t) | evallg(l);
   GEN x = (GEN)pari_malloc(l * sizeof(long));
@@ -1492,7 +1499,7 @@ gtofp(GEN z, long prec)
       if (isintzero(a)) {
         GEN y = cgetg(3, t_COMPLEX);
         b = cxcompotor(b, prec);
-        gel(y,1) = real_0_bit(expo(b) - prec2nbits(prec));
+        gel(y,1) = real_0_bit(expo(b) - prec);
         gel(y,2) = b; return y;
       }
       return cxtofp(z, prec);
@@ -1607,7 +1614,7 @@ trunc_safe(GEN x) { long e; return gcvtoi(x,&e); }
 INLINE long
 ndec2nlong(long x) { return 1 + (long)((x)*(LOG2_10/BITS_IN_LONG)); }
 INLINE long
-ndec2prec(long x) { return 2 + ndec2nlong(x); }
+ndec2prec(long x) { return ndec2nlong(x) << TWOPOTBITS_IN_LONG; }
 INLINE long
 ndec2nbits(long x) { return ndec2nlong(x) << TWOPOTBITS_IN_LONG; }
 /* Fast implementation of ceil(x / (8*sizeof(long))); typecast to (ulong)
@@ -1620,16 +1627,16 @@ nbits2nlong(long x) {
 
 INLINE long
 nbits2extraprec(long x) {
-  return (long)(((ulong)x+BITS_IN_LONG-1) >> TWOPOTBITS_IN_LONG);
+  return (((ulong)x+BITS_IN_LONG-1)>>TWOPOTBITS_IN_LONG) << TWOPOTBITS_IN_LONG;
 }
 
-/* Fast implementation of 2 + nbits2nlong(x) */
 INLINE long
 nbits2prec(long x) {
-  return (long)(((ulong)x+3*BITS_IN_LONG-1) >> TWOPOTBITS_IN_LONG);
+  return (((ulong)x+BITS_IN_LONG-1)>>TWOPOTBITS_IN_LONG) << TWOPOTBITS_IN_LONG;
 }
+
 INLINE long
-nbits2lg(long x) {
+prec2lg(long x) {
   return (long)(((ulong)x+3*BITS_IN_LONG-1) >> TWOPOTBITS_IN_LONG);
 }
 /* ceil(x / sizeof(long)) */
@@ -1638,21 +1645,21 @@ nchar2nlong(long x) {
   return (long)(((ulong)x+sizeof(long)-1) >> (TWOPOTBITS_IN_LONG-3L));
 }
 INLINE long
-prec2nbits(long x) { return (x-2) * BITS_IN_LONG; }
+prec2nbits(long x) { return x; }
 INLINE double
 bit_accuracy_mul(long x, double y) { return (x-2) * (BITS_IN_LONG*y); }
 INLINE double
-prec2nbits_mul(long x, double y) { return (x-2) * (BITS_IN_LONG*y); }
+prec2nbits_mul(long x, double y) { return x * y; }
 INLINE long
-bit_prec(GEN x) { return prec2nbits(realprec(x)); }
+bit_prec(GEN x) { return realprec(x); }
 INLINE long
-bit_accuracy(long x) { return prec2nbits(x); }
+bit_accuracy(long x) { return (x-2) * BITS_IN_LONG; }
 INLINE long
-prec2ndec(long x) { return (long)prec2nbits_mul(x, LOG10_2); }
+prec2ndec(long x) { return (long)(x * LOG10_2); }
 INLINE long
-nbits2ndec(long x) { return (long)(x * LOG10_2); }
+nbits2ndec(long x) { return prec2ndec(x); }
 INLINE long
-precdbl(long x) {return (x - 1) << 1;}
+precdbl(long x) {return x << 1;}
 INLINE long
 divsBIL(long n) { return n >> TWOPOTBITS_IN_LONG; }
 INLINE long
@@ -1688,6 +1695,12 @@ Fp_add(GEN a, GEN b, GEN m)
   return gerepileuptoint(av, p);
 }
 INLINE GEN
+Fp_double(GEN x, GEN N)
+{
+  GEN z = shifti(x, 1);
+  return cmpii(z, N) >= 0? subii(z, N): z;
+}
+INLINE GEN
 Fp_sub(GEN a, GEN b, GEN m)
 {
   pari_sp av=avma;
@@ -2261,6 +2274,10 @@ INLINE GEN
 FpXQX_renormalize(GEN x, long lx) { return ZXX_renormalize(x,lx); }
 INLINE GEN
 F2x_renormalize(GEN x, long lx)   { return Flx_renormalize(x,lx); }
+
+INLINE GEN
+F2xX_shift(GEN a, long n, long vs) { return FlxX_shift(a, n, vs); }
+
 INLINE GEN
 F2v_to_F2x(GEN x, long sv) {
   GEN y = leafcopy(x);
@@ -2389,6 +2406,10 @@ gener_Fq_local(GEN T, GEN p, GEN L)
 { return T? gener_FpXQ_local(T,p, L)
           : pgener_Fp_local(p, L); }
 
+INLINE GEN
+random_Fq(GEN T, GEN p)
+{ return T ? random_FpX(get_FpX_degree(T), get_FpX_var(T), p): randomi(p); }
+
 /* FpXQX */
 INLINE GEN
 FpXQX_div(GEN x, GEN y, GEN T, GEN p) { return FpXQX_divrem(x, y, T, p, NULL); }
@@ -3003,6 +3024,12 @@ gchar_set_evalprec(GEN gc, long prec) { gmael(gc, 8, 1)[1] = prec; }
 INLINE void
 gchar_set_prec(GEN gc, long prec) { gmael(gc, 8, 1)[2] = prec; }
 INLINE void
+gchar_copy_precs(GEN gc, GEN gc2)
+{
+  gel(gc2, 8) = shallowcopy(gel(gc,8));
+  gmael(gc2, 8, 1) = shallowcopy(gmael(gc, 8, 1));
+}
+INLINE void
 gchar_set_nfprec(GEN gc, long prec) { gmael(gc, 8, 1)[3] = prec; }
 INLINE long
 gchar_get_ntors(GEN gc)   { return  gmael(gc, 8, 2)[1]; }
@@ -3042,6 +3069,8 @@ bid_get_fact(GEN bid) { return gmael(bid,3,1); }
 INLINE GEN
 bid_get_fact2(GEN bid) { return gmael(bid,3,2); }
 INLINE GEN
+bid_get_MOD(GEN bid) { GEN y = gel(bid,4); return lg(y) == 4? gel(y,3): NULL; }
+INLINE GEN
 bid_get_sprk(GEN bid) { return gmael(bid,4,1); }
 INLINE GEN
 bid_get_sarch(GEN bid) { return gmael(bid,4,2); }
@@ -3226,3 +3255,4 @@ pari_err_VAR(const char *f, GEN x, GEN y) { pari_err(e_VAR, f,x,y); }
 INLINE void
 pari_err_PRIORITY(const char *f, GEN x, const char *op, long v)
 { pari_err(e_PRIORITY, f,x,op,v); }
+
