@@ -13,22 +13,39 @@ Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
+GEN
+ZM2_sqr(GEN A)
+{
+  GEN a = gcoeff(A,1,1), b = gcoeff(A,1,2), c = gcoeff(A,2,1), d = gcoeff(A,2,2);
+  if (equalii(b, c)) /* symetric, 3S + 1M */
+  {
+    GEN b2 = sqri(b), t = mulii(b, addii(a, d));
+    retmkmat2(mkcol2(addii(b2, sqri(a)), t), mkcol2(t, addii(b2, sqri(d))));
+  }
+  else
+  { /* general, 2S + 3M */
+    GEN bc = mulii(b, c), t = addii(a, d);
+    retmkmat2(mkcol2(addii(bc, sqri(a)), mulii(c, t)),
+              mkcol2(mulii(b, t), addii(bc, sqri(d))));
+  }
+}
+
 GEN
 ZM2_mul(GEN A, GEN B)
 {
-  const long t = 52;
+  const long t = ZM2_MUL_LIMIT+2;
   GEN A11=gcoeff(A,1,1),A12=gcoeff(A,1,2), B11=gcoeff(B,1,1),B12=gcoeff(B,1,2);
   GEN A21=gcoeff(A,2,1),A22=gcoeff(A,2,2), B21=gcoeff(B,2,1),B22=gcoeff(B,2,2);
   if (lgefint(A11) < t || lgefint(B11) < t || lgefint(A22) < t || lgefint(B22) < t
    || lgefint(A12) < t || lgefint(B12) < t || lgefint(A21) < t || lgefint(B21) < t)
-  {
+  { /* 8M */
     GEN a = mulii(A11, B11), b = mulii(A12, B21);
     GEN c = mulii(A11, B12), d = mulii(A12, B22);
     GEN e = mulii(A21, B11), f = mulii(A22, B21);
     GEN g = mulii(A21, B12), h = mulii(A22, B22);
     retmkmat2(mkcol2(addii(a,b), addii(e,f)), mkcol2(addii(c,d), addii(g,h)));
   } else
-  {
+  { /* Strassen: 7M */
     GEN M1 = mulii(addii(A11,A22), addii(B11,B22));
     GEN M2 = mulii(addii(A21,A22), B11);
     GEN M3 = mulii(A11, subii(B12,B22));
@@ -85,9 +102,25 @@ mulqi(GEN M, GEN q, GEN *ap, GEN *bp)
   return res;
 }
 
+/* test whether n is a power of 2 */
+static long
+isint2n(GEN n)
+{
+  GEN x;
+  long lx = lgefint(n), i;
+  if (lx == 2) return 0;
+  x = int_MSW(n);
+  if (*(ulong*)x != 1UL<<expu(*(ulong*)x) ) return 0;
+  for (i = 3; i < lx; i++)
+  {
+    x = int_precW(x); if (*x) return 0;
+  }
+  return 1;
+}
+
 static long
 uexpi(GEN a)
-{ return expi(subiu(shifti(a,1),1)); }
+{ return expi(a)+!isint2n(a); }
 
 static GEN
 FIXUP0(GEN M, GEN *a, GEN *b, long m)
@@ -182,26 +215,85 @@ magic_threshold(GEN a)
 { return (3+uexpi(a))>>1; }
 
 static GEN
-HGCD_basecase(GEN a, GEN b)
+HGCD_basecase(GEN y, GEN x)
 {
-  pari_sp av=avma;
-  long m = magic_threshold(a);
-  GEN u,u1,v,v1;
-  u1 = v = gen_0;
-  u = v1 = gen_1;
-  while (expi(b) >= m)
+  pari_sp av = avma;
+  GEN d, d1, q, r;
+  GEN u, u1, v, v1;
+  ulong xu, xu1, xv, xv1; /* Lehmer stage recurrence matrix */
+  int lhmres;             /* Lehmer stage return value */
+
+  long m = magic_threshold(y);
+
+  /* There is no special case for single-word numbers since this is
+   * mainly meant to be used with large moduli. */
+  if (cmpii(y,x) <= 0)
   {
-    GEN r, q = dvmdii(a,b, &r);
-    a = b; b = r; swap(u,u1); swap(v,v1);
-    u = addii(mulii(u1, q), u);
-    v = addii(mulii(v1, q), v);
-   if (gc_needed(av,2))
+    d = x; d1 = y;
+    u = gen_1; u1 = gen_0;
+    v = gen_0; v1 = gen_1;
+  } else
+  {
+    d = y; d1 = x;
+    u = gen_0; u1 = gen_1;
+    v = gen_1; v1 = gen_0;
+  }
+  while (lgefint(d) > 3 &&  expi(d1) >= m + BITS_IN_LONG + 1)
+  {
+    /* do a Lehmer-Jebelean round */
+    lhmres = lgcdii((ulong *)d, (ulong *)d1, &xu, &xu1, &xv, &xv1, 0);
+
+    if (lhmres)
     {
-      if (DEBUGMEM>1) pari_warn(warnmem,"halfgcd (d = %ld)",lgefint(b));
-      gerepileall(av,6, &a,&b,&u1,&v1,&u,&v);
+      if (lhmres == 1 || lhmres == -1)
+      {
+        if (xv1 == 1)
+        {
+          r = subii(d,d1); d = d1; d1 = r;
+          r = addii(u,u1); u = u1; u1 = r;
+          r = addii(v,v1); v = v1; v1 = r;
+        }
+        else
+        {
+          r = subii(d, mului(xv1,d1)); d = d1; d1 = r;
+          r = addii(u, mului(xv1,u1)); u = u1; u1 = r;
+          r = addii(v, mului(xv1,v1)); v = v1; v1 = r;
+        }
+      }
+      else
+      {
+        r  = subii(muliu(d,xu),  muliu(d1,xv));
+        d1 = subii(muliu(d,xu1), muliu(d1,xv1)); d = r;
+        r  = addii(muliu(u,xu),  muliu(u1,xv));
+        u1 = addii(muliu(u,xu1), muliu(u1,xv1)); u = r;
+        r  = addii(muliu(v,xu),  muliu(v1,xv));
+        v1 = addii(muliu(v,xu1), muliu(v1,xv1)); v = r;
+        if (lhmres&1) togglesign(d); else togglesign(d1);
+      }
+    } /* lhmres != 0 */
+    if (expi(d1) < m) break;
+
+    if (lhmres <= 0 && signe(d1))
+    {
+      q = dvmdii(d,d1,&r);
+      d = d1; d1 = r;
+      r = addii(u, mulii(q,u1)); u = u1; u1 = r;
+      r = addii(v, mulii(q,v1)); v = v1; v1 = r;
+    }
+    if (gc_needed(av,1))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"ratlift");
+      gerepileall(av, 6, &d, &d1, &u, &u1, &v, &v1);
     }
   }
-  return gerepilecopy(av, mkvec3(mkmat22(u,u1,v,v1), a, b));
+  while (expi(d1) >= m)
+  {
+    GEN r, q = dvmdii(d,d1, &r);
+    d = d1; d1 = r; swap(u,u1); swap(v,v1);
+    u1 = addii(mulii(u, q), u1);
+    v1 = addii(mulii(v, q), v1);
+  }
+  return gerepilecopy(av, mkvec3(mkmat22(u1,u,v1,v), d, d1));
 }
 
 static GEN HGCD(GEN x, GEN y);
@@ -291,9 +383,9 @@ GEN
 halfgcdii(GEN A, GEN B)
 {
   pari_sp av = avma;
-  GEN M, Q, a, b, m = absi_shallow(abscmpii(A, B)>0 ? A: B);
+  GEN M, Q, a, b, m = abscmpii(A, B)>0 ? A: B;
   M = HGCD0(A,B); Q = gel(M,1); a = gel(M,2); b = gel(M,3);
-  while (signe(b) && cmpii(sqri(b), m) >= 0)
+  while (signe(b) && abscmpii(sqri(b), m) >= 0)
   {
     GEN r, q = dvmdii(a, b, &r);
     a = b; b = r;
