@@ -139,7 +139,7 @@ static GEN
 mulur_2(ulong x, GEN y, long s)
 {
   long m, sh, i, lx = lg(y), e = expo(y);
-  GEN z = cgetr(lx);
+  GEN z = cgetg(lx, t_REAL);
   ulong garde;
   LOCAL_HIREMAINDER;
 
@@ -156,15 +156,15 @@ mulur_2(ulong x, GEN y, long s)
 INLINE GEN
 mul0r(GEN x)
 {
-  long l = lg(x), e = expo(x);
-  e = (l > 2)? -prec2nbits(l) + e: (e < 0? 2*e: 0);
+  long l = realprec(x), e = expo(x);
+  e = (l > 0)? e - l: (e < 0? 2*e: 0);
   return real_0_bit(e);
 }
 /* lg(x) > 2 */
 INLINE GEN
 div0r(GEN x) {
-  long l = lg(x), e = expo(x);
-  return real_0_bit(-prec2nbits(l) - e);
+  long l = realprec(x), e = expo(x);
+  return real_0_bit(-l - e);
 }
 
 GEN
@@ -254,7 +254,7 @@ sqrz_i(GEN z, GEN x, long lz)
   LOCAL_HIREMAINDER;
   LOCAL_OVERFLOW;
 
-  if (lz > SQRR_SQRI_LIMIT)
+  if (lz > prec2lg(SQRR_SQRI_LIMIT))
   {
     pari_sp av = avma;
     GEN hi = sqrispec_mirror(x+2, lz-2);
@@ -348,7 +348,7 @@ mulrrz_i(GEN z, GEN x, GEN y, long lz, long flag, long sz)
   LOCAL_OVERFLOW;
 
   if (x == y) { sqrz_i(z,x,lz); return; }
-  if (lz > MULRR_MULII_LIMIT) { mulrrz_int(z,x,y,lz,flag,sz); return; }
+  if (lz > prec2lg(MULRR_MULII_LIMIT)) { mulrrz_int(z,x,y,lz,flag,sz); return; }
   if (lz == 3) { mulrrz_3(z,x,y,flag,sz); return; }
   ez = expo(x) + expo(y);
   if (flag) { (void)mulll(x[2],y[lz]); garde = hiremainder; } else garde = 0;
@@ -400,7 +400,7 @@ mulrr(GEN x, GEN y)
   lz = lg(x);
   ly = lg(y);
   if (lz > ly) { lz = ly; swap(x, y); flag = 1; } else flag = (lz != ly);
-  z = cgetr(lz);
+  z = cgetg(lz, t_REAL);
   mulrrz_i(z, x,y, lz,flag, sx);
   return z;
 }
@@ -412,7 +412,7 @@ sqrr(GEN x)
   GEN z;
 
   if (!sx) return real_0_bit(2*expo(x));
-  lz = lg(x); z = cgetr(lz);
+  lz = lg(x); z = cgetg(lz, t_REAL);
   sqrz_i(z, x, lz);
   return z;
 }
@@ -432,17 +432,17 @@ mulir(GEN x, GEN y)
   if (sy < 0) sx = -sx;
   {
     long lz = lg(y), lx = lgefint(x);
-    GEN hi, z = cgetr(lz);
+    GEN hi, z = cgetg(lz, t_REAL);
     pari_sp av = avma;
-    if (lx < (lz>>1) || (lx < lz && lz > MULRR_MULII_LIMIT))
+    if (lx < (lz>>1) || (lx < lz && lz > prec2lg(MULRR_MULII_LIMIT)))
     { /* size mantissa of x < half size of mantissa z, or lx < lz so large
        * that mulrr will call mulii anyway: mulii */
-      x = itor(x, lx);
+      x = itor(x, lg2prec(lx));
       hi = muliispec_mirror(y+2, x+2, lz-2, lx-2);
       mulrrz_end(z, hi, lz, sx, expo(x)+expo(y), hi[lz]);
     }
     else /* dubious: complete x with 0s and call mulrr */
-      mulrrz_i(z, itor(x,lz), y, lz, 0, sx);
+      mulrrz_i(z, itor(x, lg2prec(lz)), y, lz, 0, sx);
     set_avma(av); return z;
   }
 }
@@ -565,8 +565,8 @@ divir(GEN x, GEN y)
     if (signe(x) < 0) togglesign(z);
     return z;
   }
-  z = cgetr(ly); av = avma;
-  affrr(divrr(itor(x, ly+1), y), z);
+  z = cgetg(ly, t_REAL); av = avma;
+  affrr(divrr(itor(x, lg2prec(ly+1)), y), z);
   set_avma(av); return z;
 }
 
@@ -574,18 +574,18 @@ GEN
 divur(ulong x, GEN y)
 {
   pari_sp av;
-  long ly = lg(y);
+  long p = realprec(y);
   GEN z;
 
-  if (ly == 2) pari_err_INV("divur",y);
+  if (p == 0) pari_err_INV("divur",y);
   if (!x) return div0r(y);
-  if (ly > INVNEWTON_LIMIT) {
+  if (p > INVNEWTON_LIMIT) {
     av = avma; z = invr(y);
     if (x == 1) return z;
     return gerepileuptoleaf(av, mulur(x, z));
   }
-  z = cgetr(ly); av = avma;
-  affrr(divrr(utor(x,ly+1), y), z);
+  z = cgetr(p); av = avma;
+  affrr(divrr(utor(x, p + BITS_IN_LONG), y), z);
   set_avma(av); return z;
 }
 
@@ -593,19 +593,19 @@ GEN
 divsr(long x, GEN y)
 {
   pari_sp av;
-  long ly = lg(y);
+  long p = realprec(y);
   GEN z;
 
-  if (ly == 2) pari_err_INV("divsr",y);
+  if (p == 0) pari_err_INV("divsr",y);
   if (!x) return div0r(y);
-  if (ly > INVNEWTON_LIMIT) {
+  if (p > INVNEWTON_LIMIT) {
     av = avma; z = invr(y);
     if (x == 1) return z;
     if (x ==-1) { togglesign(z); return z; }
     return gerepileuptoleaf(av, mulsr(x, z));
   }
-  z = cgetr(ly); av = avma;
-  affrr(divrr(stor(x,ly+1), y), z);
+  z = cgetr(p); av = avma;
+  affrr(divrr(stor(x,p + BITS_IN_LONG), y), z);
   set_avma(av); return z;
 }
 
@@ -613,10 +613,10 @@ divsr(long x, GEN y)
 static GEN
 invr_basecase(GEN y)
 {
-  long ly = lg(y);
-  GEN z = cgetr(ly);
+  long p = realprec(y);
+  GEN z = cgetr(p);
   pari_sp av = avma;
-  affrr(divrr(real_1(ly+1), y), z);
+  affrr(divrr(real_1(p + BITS_IN_LONG), y), z);
   set_avma(av); return z;
 }
 /* returns 1/b, Newton iteration */
@@ -628,15 +628,15 @@ invr(GEN b)
   GEN x, a;
   ulong mask;
 
-  if (l <= maxss(INVNEWTON_LIMIT, (1L<<s) + 2)) {
+  if (l <= maxss(prec2lg(INVNEWTON_LIMIT), (1L<<s) + 2)) {
     if (l == 2) pari_err_INV("invr",b);
     return invr_basecase(b);
   }
   mask = quadratic_prec_mask(l-2);
   for(i=0, p=1; i<s; i++) { p <<= 1; if (mask & 1) p--; mask >>= 1; }
-  x = cgetr(l);
+  x = cgetg(l, t_REAL);
   a = rcopy(b); a[1] = _evalexpo(0) | evalsigne(1);
-  affrr(invr_basecase(rtor(a, p+2)), x);
+  affrr(invr_basecase(rtor(a, lg2prec(p+2))), x);
   while (mask > 1)
   {
     p <<= 1; if (mask & 1) p--;
@@ -708,7 +708,7 @@ divru(GEN x, ulong y)
   }
   e = expo(x);
   lx = lg(x);
-  z = cgetr(lx);
+  z = cgetg(lx, t_REAL);
   if (lx == 3)
   {
     if (y <= uel(x,2))
