@@ -21,35 +21,36 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /***                                                                ***/
 /**********************************************************************/
 
-static ulong _maxprime = 0;
-static ulong diffptrlen;
+static ulong _maxprimelim = 0;
+static GEN _prodprimes,_prodprimes_addr;
+typedef unsigned char *byteptr;
 
-/* Building/Rebuilding the diffptr table. The actual work is done by the
+/* Build/Rebuild table of prime differences. The actual work is done by the
  * following two subroutines;  the user entry point is the function
- * initprimes() below.  initprimes1() is the old algorithm, called when
+ * initprimes() below; initprimes1() is the basecase, called when
  * maxnum (size) is moderate. Must be called after pari_init_stack() )*/
 static void
-initprimes1(ulong size, long *lenp, ulong *lastp, byteptr p1)
+initprimes1(ulong size, long *lenp, pari_prime *p1)
 {
   pari_sp av = avma;
   long k;
   byteptr q, r, s, p = (byteptr)stack_calloc(size+2), fin = p + size;
+  pari_prime *re;
 
   for (r=q=p,k=1; r<=fin; )
   {
     do { r+=k; k+=2; r+=k; } while (*++q);
     for (s=r; s<=fin; s+=k) *s = 1;
   }
-  r = p1; *r++ = 2; *r++ = 1; /* 2 and 3 */
+  re = p1; *re++ = 2; *re++ = 3; /* 2 and 3 */
   for (s=q=p+1; ; s=q)
   {
     do q++; while (*q);
     if (q > fin) break;
-    *r++ = (unsigned char) ((q-s) << 1);
+    *re++ = (pari_prime) 2*(q-p)+1;
   }
-  *r++ = 0;
-  *lenp = r - p1;
-  *lastp = ((s - p) << 1) + 1;
+  *re++ = 0;
+  *lenp = re - p1;
   set_avma(av);
 }
 
@@ -172,9 +173,10 @@ typedef struct {
     ulong arena;
     double power;
     double cutoff;
+    ulong bigarena;
 } cache_model_t;
 
-static cache_model_t cache_model = { CACHE_ARENA, CACHE_ALPHA, CACHE_CUTOFF };
+static cache_model_t cache_model = { CACHE_ARENA, CACHE_ALPHA, CACHE_CUTOFF, 0 };
 
 /* Assume that some calculation requires a chunk of memory to be
    accessed often in more or less random fashion (as in sieving).
@@ -281,7 +283,8 @@ good_arena_size(ulong slow2_size, ulong total, ulong fixed_to_cache,
 /* Use as in
     install(set_optimize,lLDG)          \\ Through some M too?
     set_optimize(2,1) \\ disable dependence on limit
-    \\ 1: how much cache usable, 2: slowdown of setup, 3: alpha, 4: cutoff
+    \\ 1: how much cache usable, 2: slowdown of setup, 3: alpha, 4: cutoff,
+    \\ 5: cache size (typically whole L2 or L3) in bytes to use in forprime()
     \\ 2,3,4 are in units of 0.001
 
     { time_primes_arena(ar,limit) =     \\ ar = arena size in K
@@ -311,6 +314,9 @@ set_optimize(long what, GEN g)
   case 4:
     ret = (long)(cache_model.cutoff * 1000);
     break;
+  case 5:
+    ret = (long)(cache_model.bigarena);
+    break;
   default:
     pari_err_BUG("set_optimize");
     break;
@@ -323,25 +329,26 @@ set_optimize(long what, GEN g)
     case 2: slow2_in_roots     = (double)val / 1000.; break;
     case 3: cache_model.power  = (double)val / 1000.; break;
     case 4: cache_model.cutoff = (double)val / 1000.; break;
+    case 5: cache_model.bigarena = val; break;
     }
   }
   return ret;
 }
 
-/* s is odd; prime differences (starting from 5-3=2) start at known_primes[2],
-  terminated by a 0 byte. Checks n odd numbers starting at 'start', setting
-  bytes starting at data to 0 (composite) or 1 (prime) */
+/* s is odd; prime (starting from 3 = known_primes[2]), terminated by a 0 byte.
+ * Checks n odd numbers starting at 'start', setting bytes to 0 (composite)
+ * or 1 (prime), starting at data */
 static void
-sieve_chunk(byteptr known_primes, ulong s, byteptr data, ulong n)
+sieve_chunk(pari_prime *known_primes, ulong s, byteptr data, ulong n)
 {
-  ulong p, cnt = n-1, start = s, delta = 1;
-  byteptr q;
+  ulong p, cnt = n-1, start = s;
+  pari_prime *q;
 
   memset(data, 0, n);
   start >>= 1;  /* (start - 1)/2 */
   start += n; /* Corresponds to the end */
   /* data corresponds to start, q runs over primediffs */
-  for (q = known_primes + 1, p = 3; delta; delta = *++q, p += delta)
+  for (q = known_primes + 1, p = 3; p; p = *++q)
   { /* first odd number >= start > p and divisible by p
        = last odd number <= start + 2p - 2 and 0 (mod p)
        = p + last number <= start + p - 2 and 0 (mod 2p)
@@ -352,150 +359,135 @@ sieve_chunk(byteptr known_primes, ulong s, byteptr data, ulong n)
   }
 }
 
-/* assume maxnum <= 436273289 < 2^29 */
 static void
-initprimes0(ulong maxnum, long *lenp, ulong *lastp, byteptr p1)
+set_prodprimes(void)
+{
+  pari_sp ltop = avma, av;
+  ulong b = 1UL << 8, M = minuu(maxprime(), GP_DATA->factorlimit);
+  GEN W, w, v = primes_interval_zv(3, M);
+  long s, u, j, jold, l = lg(v);
+
+  W = cgetg(64+1, t_VEC);
+  for (jold = j = u = 1; j < l; j++)
+    if (j==l-1 || uel(v,j) >= b)
+    {
+      long lw = (j == l-1? l: j) - jold + 1;
+      w = v+jold-1; w[0] = evaltyp(t_VECSMALL) | _evallg(lw);
+      gel(W,u++) = zv_prod_Z(w); /* p_jold ... p_{j-1} */
+      jold = j; b *= 2;
+      if (b > M) b = M; /* truncate last run */
+    }
+  setlg(W, u);
+  for (j = 2; j < u; j++) gel(W,j) = mulii(gel(W,j-1), gel(W,j));
+  s = gsizeword(W);
+  w = (GEN)pari_malloc(s*sizeof(long));
+  av = (pari_sp)(w + s);
+  _prodprimes_addr = w;
+  _prodprimes = gcopy_avma(W, &av);
+  set_avma(ltop);
+}
+
+static void
+initprimes0(ulong maxnum, long *lenp, pari_prime *p1)
 {
   pari_sp av = avma, bot = pari_mainstack->bot;
   long alloced, psize;
-  byteptr q, end, p, end1, plast, curdiff;
-  ulong last, remains, curlow, rootnum, asize;
-  ulong prime_above;
-  byteptr p_prime_above;
+  byteptr q, end, p;
+  ulong remains, curlow, rootnum, asize, prime_above, last;
+  pari_prime *end1, *curdiff, *p_prime_above;
 
-  maxnum |= 1; /* make it odd. */
+  if (!odd(maxnum)) maxnum--; /* make it odd. */
   /* base case */
-  if (maxnum < 1ul<<17) { initprimes1(maxnum>>1, lenp, lastp, p1); return; }
+  if (maxnum < 1ul<<17) { initprimes1(maxnum>>1, lenp, p1); return; }
 
   /* Checked to be enough up to 40e6, attained at 155893 */
   rootnum = usqrt(maxnum) | 1;
-  initprimes1(rootnum>>1, &psize, &last, p1);
+  initprimes1(rootnum>>1, &psize, p1);
+  last = rootnum;
   end1 = p1 + psize - 1;
   remains = (maxnum - last) >> 1; /* number of odd numbers to check */
-
   /* we access primes array of psize too; but we access it consecutively,
    * thus we do not include it in fixed_to_cache */
   asize = good_arena_size((ulong)(rootnum * slow2_in_roots), remains+1, 0,
                           &cache_model) - 1;
   /* enough room on the stack ? */
   alloced = (((byteptr)avma) <= ((byteptr)bot) + asize);
-  if (alloced)
-    p = (byteptr)pari_malloc(asize+1);
-  else
-    p = (byteptr)stack_malloc(asize+1);
+  p = (byteptr)(alloced? pari_malloc(asize+1): stack_malloc(asize+1));
   end = p + asize; /* the 0 sentinel goes at end. */
   curlow = last + 2; /* First candidate: know primes up to last (odd). */
   curdiff = end1;
 
   /* During each iteration p..end-1 represents a range of odd
-     numbers.  plast is a pointer which represents the last prime seen,
-     it may point before p..end-1. */
-  plast = p - 1;
+     numbers.   */
   p_prime_above = p1 + 2;
   prime_above = 3;
   while (remains)
   { /* cycle over arenas; performance not crucial */
-    unsigned char was_delta;
+    pari_prime was_delta;
     if (asize > remains) { asize = remains; end = p + asize; }
     /* Fake the upper limit appropriate for the given arena */
     while (prime_above*prime_above <= curlow + (asize << 1) && *p_prime_above)
-      prime_above += *p_prime_above++;
+      prime_above = *p_prime_above++;
     was_delta = *p_prime_above;
     *p_prime_above = 0; /* sentinel for sieve_chunk */
     sieve_chunk(p1, curlow, p, asize);
     *p_prime_above = was_delta; /* restore */
 
     p[asize] = 0; /* sentinel */
-    for (q = p; ; plast = q++)
+    for (q = p; ; q++)
     { /* q runs over addresses corresponding to primes */
       while (*q) q++; /* use sentinel at end */
       if (q >= end) break;
-      *curdiff++ = (unsigned char)(q-plast) << 1; /* < 255 for q < 436273291 */
+      *curdiff++ = (pari_prime) 2*(q-p) + curlow;
     }
-    plast -= asize;
     remains -= asize;
     curlow += (asize<<1);
   }
-  last = curlow - ((p - plast) << 1);
   *curdiff++ = 0; /* sentinel */
   *lenp = curdiff - p1;
-  *lastp = last;
   if (alloced) pari_free(p); else set_avma(av);
 }
 
 ulong
-maxprime(void) { return diffptr ? _maxprime : 0; }
+maxprime(void) { return pari_PRIMES? pari_PRIMES[pari_PRIMES[0]]: 0; }
 ulong
-maxprimeN(void) { return diffptr ? diffptrlen-1: 0; }
-
+maxprimelim(void) { return pari_PRIMES? _maxprimelim: 0; }
+ulong
+maxprimeN(void) { return pari_PRIMES? pari_PRIMES[0]: 0; }
+GEN
+prodprimes(void) { return pari_PRIMES? _prodprimes: NULL; }
 void
-maxprime_check(ulong c) { if (_maxprime < c) pari_err_MAXPRIME(c); }
+maxprime_check(ulong c) { if (maxprime() < c) pari_err_MAXPRIME(c); }
 
-/* We ensure 65302 <= maxnum <= 436273289: the LHS ensures modular function
- * have enough fast primes to work, the RHS ensures that p_{n+1} - p_n < 255
- * (N.B. RHS would be incorrect since initprimes0 would make it odd, thereby
- * increasing it by 1) */
-byteptr
-initprimes(ulong maxnum, long *lenp, ulong *lastp)
+static pari_prime*
+initprimes(ulong maxnum)
 {
-  byteptr t;
+  pari_prime *t;
+  long len;
+  ulong N;
   if (maxnum < 65537)
+  {
     maxnum = 65537;
-  else if (maxnum > 436273289)
-    maxnum = 436273289;
-  t = (byteptr)pari_malloc((size_t) (1.09 * maxnum/log((double)maxnum)) + 146);
-  initprimes0(maxnum, lenp, lastp, t);
-  return (byteptr)pari_realloc(t, *lenp);
+    N = 6543;
+  }
+  else
+    N = (long) ceil(primepi_upper_bound((double)maxnum));
+  t = (pari_prime*) pari_malloc(sizeof(*t) * (N+2));
+  initprimes0(maxnum, &len, t+1); t[0] = (pari_prime)(len-1);
+  _maxprimelim = maxnum;
+  return (pari_prime*) pari_realloc(t, sizeof(*t) * (len+1));
 }
 
 void
 initprimetable(ulong maxnum)
 {
-  long len;
-  ulong last;
-  byteptr p = initprimes(maxnum, &len, &last), old = diffptr;
-  diffptrlen = minss(diffptrlen, len);
-  _maxprime  = minss(_maxprime,last); /*Protect against ^C*/
-  diffptr = p; diffptrlen = len; _maxprime = last;
+  pari_prime *old = pari_PRIMES;
+#ifdef LONG_IS_64BIT
+  maxnum = minuu(maxnum, 4294967295);
+#endif
+  pari_PRIMES = initprimes(maxnum);
   if (old) free(old);
-}
-
-/* all init_primepointer_xx routines set *ptr to the corresponding place
- * in prime table */
-/* smallest p >= a */
-ulong
-init_primepointer_geq(ulong a, byteptr *pd)
-{
-  ulong n, p;
-  prime_table_next_p(a, pd, &p, &n);
-  return p;
-}
-/* largest p < a */
-ulong
-init_primepointer_lt(ulong a, byteptr *pd)
-{
-  ulong n, p;
-  prime_table_next_p(a, pd, &p, &n);
-  PREC_PRIME_VIADIFF(p, *pd);
-  return p;
-}
-/* largest p <= a */
-ulong
-init_primepointer_leq(ulong a, byteptr *pd)
-{
-  ulong n, p;
-  prime_table_next_p(a, pd, &p, &n);
-  if (p != a) PREC_PRIME_VIADIFF(p, *pd);
-  return p;
-}
-/* smallest p > a */
-ulong
-init_primepointer_gt(ulong a, byteptr *pd)
-{
-  ulong n, p;
-  prime_table_next_p(a, pd, &p, &n);
-  if (p == a) NEXT_PRIME_VIADIFF(p, *pd);
-  return p;
+  set_prodprimes();
 }
 
 /**********************************************************************/
@@ -510,7 +502,9 @@ optimize_chunk(ulong a, ulong b)
 {
   /* TODO: Optimize size (surely < 512k to stay in L2 cache, but not so large
    * as to force recalculating too often). */
-  ulong chunk = 0x80000UL;
+  /* bigarena is in bytes, we use bits, and only odds */
+  ulong defchunk = (a>>31) > 1 ? 0x80000UL : 0x8000;
+  ulong chunk = (cache_model.bigarena ? cache_model.bigarena : defchunk)<<4;
   ulong tmp = (b - a) / chunk + 1;
 
   if (tmp == 1)
@@ -542,47 +536,70 @@ u_forprime_set_prime_table(forprime_t *T, ulong a)
   if (a < 3)
   {
     T->p = 0;
-    T->d = diffptr;
+    T->n = 0;
   }
   else
-    T->p = init_primepointer_lt(a, &T->d);
+  {
+    long n = PRIMES_search(a - 1);
+    if (n < 0) n = - n - 1;
+    T->n = n;
+    T->p = pari_PRIMES[n];
+  }
 }
 
 /* Set p so that p + q the smallest integer = c (mod q) and > original p.
- * Assume 0 < c < q. Set p = 0 on overflow */
+ * Assume 0 < c < q. */
 static void
 arith_set(forprime_t *T)
 {
   ulong r = T->p % T->q; /* 0 <= r <= min(p, q-1) */
   pari_sp av = avma;
-  GEN d = adduu(T->p - r, T->c);
+  GEN d = adduu(T->p - r, T->c); /* = c mod q */
   if (T->c > r) d = subiu(d, T->q);
   /* d = c mod q,  d = c > r? p-r+c-q: p-r+c, so that
    *  d <= p  and  d+q = c>r? p-r+c  : p-r+c+q > p */
-  if (signe(d) <= 0) d = addiu(d, T->q); /* and now d > 0 */
-  T->p = itou_or_0(d); set_avma(av);
+  if (signe(d) <= 0)
+  {
+    T->p = 0;
+    T->strategy = PRST_nextprime;
+    affii(d, T->pp);
+  }
+  else
+    T->p = itou_or_0(d);
+  set_avma(av);
 }
 
-/* run through primes in arithmetic progression = c (mod q) */
+/* Run through primes in arithmetic progression = c (mod q).
+ * Warning: b = ULONG_MAX may signal that we are called by higher level
+ * function handling a continuation for larger b; this sentinel value
+ * must not be modified */
 static int
 u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
                             ulong a, ulong b, ulong c, ulong q)
 {
-  ulong maxp, maxp2;
+#ifdef LONG_IS_64BIT
+  const ulong UPRIME_MAX = 18446744073709551557UL;
+#else
+  const ulong UPRIME_MAX = 4294967291UL;
+#endif
+  ulong P, P2, Y, sieveb;
+
   if (!odd(b) && b > 2) b--;
   if (a > b || b < 2)
   {
     T->strategy = PRST_diffptr; /* paranoia */
     T->p = 0; /* empty */
     T->b = 0; /* empty */
-    T->d = diffptr;
+    T->n = 0;
     return 0;
   }
-  maxp = maxprime();
+  P = maxprime();
+  if (b != ULONG_MAX && b > UPRIME_MAX) b = UPRIME_MAX;
   if (q != 1)
   {
-    c %= q;
-    if (ugcd(c,q) != 1) { a = maxuu(a,c); b = minuu(b,c); }
+    ulong D;
+    c %= q; D = ugcd(c, q);
+    if (D != 1) { a = maxuu(a,D); if (b != ULONG_MAX) b = minuu(b,D); }
     if (odd(q) && (a > 2 || c != 2))
     { /* only *odd* primes. If a <= c = 2, then p = 2 must be included :-( */
       if (!odd(c)) c += q;
@@ -595,37 +612,37 @@ u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
   T->psieve = psieve; /* unused for now */
   T->isieve = NULL; /* unused for now */
   T->b = b;
-  if (maxp >= b) { /* [a,b] \subset prime table */
+  if (P >= b) { /* [a,b] \subset prime table */
     u_forprime_set_prime_table(T, a);
     return 1;
   }
-  /* b > maxp */
-  if (a >= maxp)
+  /* b > P */
+  if (a >= P)
   {
     T->p = a - 1;
-    if (T->q > 1) arith_set(T);
+    if (T->q != 1) arith_set(T);
   }
   else
     u_forprime_set_prime_table(T, a);
 
-  maxp2 = (maxp & HIGHMASK)? 0 : maxp*maxp;
+  P2 = (P & HIGHMASK)? 0 : P*P;
+  sieveb = b; if (P2 && P2 < b) sieveb = P2;
+  /* maxprime^2 >= sieveb */
+  Y = sieveb - maxprimelim(); /* length of sievable interval */
+  P = usqrt(sieveb); /* largest sieving prime */
   /* FIXME: should sieve as well if q != 1, adapt sieve code */
-  if (q != 1 || (maxp2 && maxp2 <= a)
-             || T->b - maxuu(a,maxp) < maxp / expu(b))
+  if (q != 1 || (P2 && P2 <= a) || 3/M_LN2 * Y < uprimepi(P))
+  /* Sieve unimplemented/impossible/too costly. Cost model is
+   * - nextprime: about Y / log(b) primes to test [neglect cost for composites]
+   *   individual cost average = 3 log2(b) mulmod, total = 3 Y / log(2) mulmod
+   * - sieve: pi(P) mod + Y loglog(b) add
+   * Since loglog(b) < 4, and add < 10*mulmod, we neglect the Y loglog(b) term.
+   * We have mod < mulmod < 2*mod; for now, assume mulmod ~ mod. */
   { if (T->strategy==PRST_none) T->strategy = PRST_unextprime; }
   else
   { /* worth sieving */
-#ifdef LONG_IS_64BIT
-    const ulong UPRIME_MAX = 18446744073709551557UL;
-#else
-    const ulong UPRIME_MAX = 4294967291UL;
-#endif
-    ulong sieveb;
-    if (b > UPRIME_MAX) b = UPRIME_MAX;
-    sieveb = b;
-    if (maxp2 && maxp2 < b) sieveb = maxp2;
     if (T->strategy==PRST_none) T->strategy = PRST_sieve;
-    sieve_init(T, maxuu(maxp+2, a), sieveb);
+    sieve_init(T, maxuu(maxprimelim()+1, a), sieveb);
   }
   return 1;
 }
@@ -652,9 +669,40 @@ u_forprime_restrict(forprime_t *T, ulong c) { T->b = c; }
 int
 forprimestep_init(forprime_t *T, GEN a, GEN b, GEN q)
 {
+  GEN c = NULL;
   long lb;
+
   a = gceil(a); if (typ(a) != t_INT) pari_err_TYPE("forprime_init",a);
-  if (signe(a) <= 0) a = gen_1;
+  T->qq = NULL; T->q = 1; T->c = 0;
+  if (q)
+  {
+    switch(typ(q))
+    {
+      case t_INT:
+        c = a; break;
+      case t_INTMOD:
+        c = gel(q,2); q = gel(q,1);
+        /* first int >= initial a which is = c (mod q) */
+        a = addii(a, modii(subii(c,a), q)); break;
+      default: pari_err_TYPE("forprimestep_init",q);
+    }
+    if (signe(q) <= 0) pari_err_TYPE("forprimestep_init (q <= 0)",q);
+    if (equali1(q)) c = q = NULL;
+    else
+    {
+      GEN D = gcdii(c, q);
+      if (!is_pm1(D))
+      { /* at most one prime: c */
+        if (cmpii(a, D) < 0) a = D;
+        if (gcmp(b, D) > 0) b = D;
+      }
+      if ((T->q = itou_or_0(q)))
+        T->c = umodiu(c, T->q);
+      else
+        T->qq = q;
+    }
+  }
+  if (signe(a) <= 0) a = q? modii(a, q): gen_1;
   if (b && typ(b) != t_INFINITY)
   {
     b = gfloor(b);
@@ -677,33 +725,13 @@ forprimestep_init(forprime_t *T, GEN a, GEN b, GEN q)
     T->strategy = PRST_nextprime; /* paranoia */
     T->bb = T->pp = gen_0; return 0;
   }
-  T->pp = cgeti(lb);
-  T->c = 0;
-  T->q = 1;
+  T->pp = cgeti(T->qq? maxuu(lb, lgefint(T->qq)): lb);
   /* a, b are positive integers, a <= b */
-  if (q)
-  {
-    switch(typ(q))
-    {
-      case t_INT: break;
-      case t_INTMOD: a = addii(a, modii(subii(gel(q,2),a), gel(q,1)));
-                     q = gel(q,1); break;
-      default: pari_err_TYPE("forprimestep_init",q);
-    }
-    if (signe(q) <= 0) pari_err_TYPE("forprimestep_init (q <= 0)",q);
-    if (equali1(q)) q = NULL;
-    else
-    {
-      T->q = itou(q);
-      T->c = umodiu(a, T->q);
-    }
-  }
-  if (lgefint(a) == 3) /* lb == 3 implies b != NULL */
+  if (!T->qq && lgefint(a) == 3) /* lb == 3 implies b != NULL */
     return u_forprime_arith_init(T, uel(a,2), lb == 3? uel(b,2): ULONG_MAX,
                                     T->c, T->q);
   T->strategy = PRST_nextprime;
-  affii(subiu(a,T->q), T->pp);
-  return 1;
+  affii(T->qq? subii(a,T->qq): subiu(a,T->q), T->pp); return 1;
 }
 int
 forprime_init(forprime_t *T, GEN a, GEN b)
@@ -716,13 +744,11 @@ forprime_init(forprime_t *T, GEN a, GEN b)
 static void
 sieve_block(ulong a, ulong b, ulong maxpos, unsigned char* sieve)
 {
-  ulong p = 2, lim = usqrt(b), sz = (b-a) >> 1;
-  byteptr d = diffptr+1;
+  ulong i, lim = usqrt(b), sz = (b-a) >> 1;
   (void)memset(sieve, 0, maxpos+1);
-  for (;;)
+  for (i = 2;; i++)
   { /* p is odd */
-    ulong k, r;
-    NEXT_PRIME_VIADIFF(p, d); /* starts at p = 3 */
+    ulong k, r, p = pari_PRIMES[i]; /* starts at p = 3 */
     if (p > lim) break;
 
     /* solve a + 2k = 0 (mod p) */
@@ -771,7 +797,11 @@ pari_init_primes(ulong maxprime)
 void
 pari_close_primes(void)
 {
-  pari_free(diffptr);
+  if (pari_PRIMES)
+  {
+    pari_free(pari_PRIMES);
+    pari_free(_prodprimes_addr);
+  }
   pari_free(pari_sieve_modular.sieve);
 }
 
@@ -815,19 +845,19 @@ u_forprime_next(forprime_t *T)
   {
     for(;;)
     {
-      if (!*(T->d))
+      if (++T->n <= pari_PRIMES[0])
       {
+        T->p = pari_PRIMES[T->n];
+        if (T->p > T->b) return 0;
+        if (T->q == 1 || T->p % T->q == T->c) return T->p;
+      }
+      else
+      { /* beyond the table */
         T->strategy = T->isieve? PRST_sieve: PRST_unextprime;
         if (T->q != 1) { arith_set(T); if (!T->p) return 0; }
         /* T->p possibly not a prime if q != 1 */
         break;
       }
-      else
-      {
-        NEXT_PRIME_VIADIFF(T->p, T->d);
-        if (T->p > T->b) return 0;
-        if (T->q == 1 || T->p % T->q == T->c) return T->p;
-      }
     }
   }
   if (T->strategy == PRST_sieve)
@@ -942,6 +972,7 @@ NEXT_CHUNK:
       }
 #endif
       T->p = unextprime(T->p + 1);
+      if (T->p > T->b) return 0;
     }
     else do {
       T->p += T->q;
@@ -976,7 +1007,7 @@ forprime_next(forprime_t *T)
     p = nextprime(addiu(p, 1));
     if (T->bb && abscmpii(p, T->bb) > 0) return gc_NULL(av);
   } else do {
-    p = addiu(p, T->q);
+    p = T->qq? addii(p, T->qq): addiu(p, T->q);
     if (T->bb && abscmpii(p, T->bb) > 0) return gc_NULL(av);
   } while (!BPSW_psp(p));
   affii(p, T->pp); return gc_const(av, T->pp);
