@@ -86,13 +86,13 @@ static const ulong readonly_constants[] = {
   (ulong)(readonly_constants+7)
 };
 THREAD GEN zetazone, bernzone, eulerzone, primetab;
-byteptr diffptr;
+pari_prime *pari_PRIMES;
 FILE    *pari_outfile, *pari_errfile, *pari_logfile, *pari_infile;
 char    *current_logfile, *current_psfile, *pari_datadir;
 long    gp_colors[c_LAST];
 int     disable_color;
 ulong   DEBUGLEVEL, DEBUGMEM;
-long    DEBUGVAR;
+THREAD  long    DEBUGVAR;
 ulong   pari_mt_nbthreads;
 long    precreal;
 ulong   precdl, pari_logstyle;
@@ -109,6 +109,7 @@ static pari_stack s_MODULES;
 const long functions_tblsz = 135; /* size of functions_hash */
 entree **functions_hash, **defaults_hash;
 
+void (*cb_pari_display_hist)(long n);
 char *(*cb_pari_fgets_interactive)(char *s, int n, FILE *f);
 int (*cb_pari_get_line_interactive)(const char*, const char*, filtre_t *F);
 void (*cb_pari_quit)(long);
@@ -123,6 +124,7 @@ void (*cb_pari_err_recover)(long);
 int (*cb_pari_break_loop)(int);
 int (*cb_pari_is_interactive)(void);
 void (*cb_pari_start_output)(void);
+void (*cb_pari_long_help)(const char *s, long num);
 
 const char * pari_library_path = NULL;
 
@@ -141,17 +143,11 @@ static void pari_init_timer(void);
 static THREAD long next_block;
 static THREAD GEN cur_block; /* current block in block list */
 static THREAD GEN root_block; /* current block in block list */
-#ifdef DEBUG
-static THREAD long NUM;
-#endif
 
 static void
 pari_init_blocks(void)
 {
   next_block = 0; cur_block = NULL; root_block = NULL;
-#ifdef DEBUG
-  NUM = 0;
-#endif
 }
 
 static void
@@ -350,9 +346,6 @@ newblock(size_t n)
   bl_num(x)  = next_block++;
   if (cur_block) bl_next(cur_block) = x;
   root_block = blockinsert(x, root_block, &d);
-#ifdef DEBUG
-  err_printf("+ %ld\n", ++NUM);
-#endif
   if (DEBUGMEM > 2)
     err_printf("new block, size %6lu (no %ld): %08lx\n", n, next_block-1, x);
   return cur_block = x;
@@ -385,30 +378,29 @@ gunclone(GEN x)
     err_printf("killing block (no %ld): %08lx\n", bl_num(x), x);
   free((void*)bl_base(x)); /* pari_free not needed: we already block */
   BLOCK_SIGINT_END;
-#ifdef DEBUG
-  err_printf("- %ld\n", NUM--);
-#endif
 }
 
+static void
+vec_gunclone_deep(GEN x)
+{
+  long i, l = lg(x);
+  for (i = 1; i < l; i++) gunclone_deep(gel(x,i));
+}
 /* Recursively look for clones in the container and kill them. Then kill
- * container if clone. SIGINT could be blocked until it returns */
+ * container if clone. */
 void
 gunclone_deep(GEN x)
 {
-  long i, lx;
   GEN v;
   if (isclone(x) && bl_refc(x) > 1) { --bl_refc(x); return; }
   BLOCK_SIGINT_START;
   switch(typ(x))
   {
     case t_VEC: case t_COL: case t_MAT:
-      lx = lg(x);
-      for (i=1;i<lx;i++) gunclone_deep(gel(x,i));
+      vec_gunclone_deep(x);
       break;
     case t_LIST:
-      v = list_data(x); lx = v? lg(v): 1;
-      for (i=1;i<lx;i++) gunclone_deep(gel(v,i));
-      if (v) killblock(v);
+      if ((v = list_data(x))) { vec_gunclone_deep(v); gunclone(v); }
       break;
   }
   if (isclone(x)) gunclone(x);
@@ -804,12 +796,7 @@ pari_init_defaults(void)
   long i;
   initout(1);
 
-#ifdef LONG_IS_64BIT
   precreal = 128;
-#else
-  precreal = 96;
-#endif
-
   precdl = 16;
   DEBUGLEVEL = 0;
   setalldebug(0);
@@ -1148,6 +1135,7 @@ new_chunk_resize(size_t x)
 static void
 pari_thread_set_global(struct pari_global_state *gs)
 {
+  setdebugvar(gs->debugvar);
   push_localbitprec(gs->bitprec);
   pari_set_primetab(gs->primetab);
   pari_set_seadata(gs->seadata);
@@ -1157,6 +1145,7 @@ pari_thread_set_global(struct pari_global_state *gs)
 static void
 pari_thread_get_global(struct pari_global_state *gs)
 {
+  gs->debugvar = getdebugvar();
   gs->bitprec = get_localbitprec();
   gs->primetab = primetab;
   gs->seadata = pari_get_seadata();
@@ -1245,8 +1234,6 @@ dflt_err_recover(long errnum) { (void) errnum; pari_exit(); }
 static void
 dflt_pari_quit(long err) { (void)err; /*do nothing*/; }
 
-static int pari_err_display(GEN err);
-
 /* initialize PARI data. Initialize [new|old]fun to NULL for default set. */
 void
 pari_init_opts(size_t parisize, ulong maxprime, ulong init_opts)
@@ -1266,6 +1253,7 @@ pari_init_opts(size_t parisize, ulong maxprime, ulong init_opts)
   cb_pari_is_interactive = NULL;
   cb_pari_start_output = NULL;
   cb_pari_sigint = dflt_sigint_fun;
+  cb_pari_long_help = NULL;
   if (init_opts&INIT_JMPm) cb_pari_err_recover = dflt_err_recover;
 
   pari_stackcheck_init(&u);
@@ -1279,7 +1267,7 @@ pari_init_opts(size_t parisize, ulong maxprime, ulong init_opts)
   pari_mainstack = (struct pari_mainstack *) malloc(sizeof(*pari_mainstack));
   paristack_alloc(parisize, 0);
   init_universal_constants();
-  diffptr = NULL;
+  pari_PRIMES = NULL;
   if (!(init_opts&INIT_noPRIMEm))
   {
     GP_DATA->primelimit = maxprime;
@@ -1333,7 +1321,7 @@ pari_close_opts(ulong init_opts)
 
   free((void*)functions_hash);
   free((void*)defaults_hash);
-  if (diffptr) pari_close_primes();
+  if (pari_PRIMES) pari_close_primes();
   free(current_logfile);
   free(current_psfile);
   pari_mainstack_free(pari_mainstack);
@@ -1463,7 +1451,7 @@ pari_warn(int numerr, ...)
   {
     case warnuser:
       out_puts(pariErr, "user warning: ");
-      out_print0(pariErr, NULL, va_arg(ap, GEN), f_RAW);
+      out_print1(pariErr, NULL, va_arg(ap, GEN), f_RAW);
       break;
 
     case warnmem:
@@ -1510,6 +1498,7 @@ pari_sigint(const char *time_s)
   int recover=0;
   BLOCK_SIGALRM_START
   err_init();
+  mt_break_recover();
   closure_err(0);
   err_init_msg(0);
   out_puts(pariErr, "user interrupt after ");
@@ -1807,30 +1796,25 @@ pari_err2str(GEN e)
                         gel(e,2), gdisplay(gel(e,3)));
   case e_STACK:
   case e_STACKTHREAD:
+  {
+    const char *what = numerr == e_STACK? "PARI": "thread";
+    const char *var = numerr == e_STACK? "parisizemax": "threadsizemax";
+    size_t s = numerr == e_STACK? pari_mainstack->vsize: GP_DATA->threadsizemax;
+    char *hint = (char *) pari_malloc(512*sizeof(char));
+    char *buf = (char *) pari_malloc(512*sizeof(char));
+    if (s)
+      sprintf(hint,"you can increase '%s' using default()", var);
+    else
     {
-      const char *stack = numerr == e_STACK? "PARI": "thread";
-      const char *var = numerr == e_STACK? "parisizemax": "threadsizemax";
-      size_t rsize = numerr == e_STACKTHREAD && GP_DATA->threadsize ?
-                                GP_DATA->threadsize: pari_mainstack->rsize;
-      size_t vsize = numerr == e_STACK? pari_mainstack->vsize:
-                                        GP_DATA->threadsizemax;
-      char *buf = (char *) pari_malloc(512*sizeof(char));
-      if (vsize)
-      {
-        sprintf(buf, "the %s stack overflows !\n"
-            "  current stack size: %lu (%.3f Mbytes)\n"
-            "  [hint] you can increase '%s' using default()\n",
-            stack, (ulong)vsize, (double)vsize/1048576., var);
-      }
-      else
-      {
-        sprintf(buf, "the %s stack overflows !\n"
-            "  current stack size: %lu (%.3f Mbytes)\n"
-            "  [hint] set '%s' to a nonzero value in your GPRC\n",
-            stack, (ulong)rsize, (double)rsize/1048576., var);
-      }
-      return buf;
+      s = (numerr != e_STACK || !GP_DATA->threadsize)? pari_mainstack->rsize
+                                                     : GP_DATA->threadsize;
+      sprintf(hint,"set '%s' to a nonzero value in your GPRC", var);
     }
+    sprintf(buf, "the %s stack overflows !\n"
+                 "  current stack size: %lu (%.3f Mbytes)\n  [hint] %s\n",
+                 what, (ulong)s, (double)s/1048576., hint);
+    return buf;
+  }
   case e_SYNTAX:
     return pari_strdup(GSTR(gel(e,2)));
   case e_TYPE:
@@ -1848,7 +1832,7 @@ pari_err2str(GEN e)
   return NULL; /*LCOV_EXCL_LINE*/
 }
 
-static int
+int
 pari_err_display(GEN err)
 {
   long numerr=err_get_num(err);
@@ -2190,7 +2174,7 @@ icopy_avma_canon(GEN x, pari_sp AVMA)
 {
   long i, lx = lgefint(x);
   GEN y = ((GEN)AVMA) - lx;
-  y[0] = evaltyp(t_INT)|evallg(lx); /* kills isclone */
+  y[0] = evaltyp(t_INT)|_evallg(lx); /* kills isclone */
   y[1] = x[1]; x = int_MSW(x);
   for (i=2; i<lx; i++, x = int_precW(x)) y[i] = *x;
   return y;
@@ -2221,7 +2205,7 @@ gcopy_av0_canon(GEN x, pari_sp *AVMA)
       GEN y = cgetlist_avma(AVMA), z = list_data(x);
       if (z) {
         list_data(y) = gcopy_av0_canon(z, AVMA);
-        y[1] = evaltyp(t)|evallg(lg(z)-1);
+        y[1] = evaltyp(t)|_evallg(lg(z)-1);
       } else {
         list_data(y) = NULL;
         y[1] = evaltyp(t);
@@ -2369,7 +2353,7 @@ gclone(GEN x)
   { /* non recursive types */
     case t_INT:
       lx = lgefint(x);
-      y[0] = evaltyp(t_INT)|evallg(lx);
+      y[0] = evaltyp(t_INT)|_evallg(lx);
       for (i=1; i<lx; i++) y[i] = x[i];
       break;
     case t_REAL:
@@ -2438,8 +2422,15 @@ shiftaddress_canon(GEN x, long dec)
       break;
 
     /* one more special case */
-    case t_LIST:
-      if (!list_data(x)) break;
+    case t_LIST: {
+      GEN Lx = list_data(x);
+      if (Lx) {
+        GEN L = (GEN)((long)Lx+dec);
+        shiftaddress_canon(L, dec);
+        list_data(x) = gcopy(L);
+      }
+      break;
+    }
     default: /* Fall through */
       lx = lg(x);
       for (i=lontyp[tx]; i<lx; i++) {
@@ -2481,7 +2472,7 @@ obj_init(long d, long n)
 GEN
 obj_insert(GEN S, long K, GEN O)
 {
-  GEN o, v = gel(S, lg(S)-1);
+  GEN o, v = veclast(S);
   if (typ(v) != t_VEC) pari_err_TYPE("obj_insert", S);
   if (!check_clone(v))
   {
@@ -2496,7 +2487,7 @@ obj_insert(GEN S, long K, GEN O)
 GEN
 obj_insert_shallow(GEN S, long K, GEN O)
 {
-  GEN v = gel(S, lg(S)-1);
+  GEN v = veclast(S);
   if (typ(v) != t_VEC) pari_err_TYPE("obj_insert", S);
   gel(v,K) = O;
   return gel(v,K);
@@ -2506,7 +2497,7 @@ obj_insert_shallow(GEN S, long K, GEN O)
 GEN
 obj_check(GEN S, long K)
 {
-  GEN O, v = gel(S, lg(S)-1);
+  GEN O, v = veclast(S);
   if (typ(v) != t_VEC || K >= lg(v)) pari_err_TYPE("obj_check", S);
   O = gel(v,K); return isintzero(O)? NULL: O;
 }
@@ -2540,7 +2531,7 @@ obj_checkbuild_padicprec(GEN S, long tag, GEN (*build)(GEN,long), long prec)
 void
 obj_free(GEN S)
 {
-  GEN v = gel(S, lg(S)-1);
+  GEN v = veclast(S);
   long i;
   if (typ(v) != t_VEC) pari_err_TYPE("obj_free", S);
   for (i = 1; i < lg(v); i++)
@@ -2700,10 +2691,15 @@ gerepile(pari_sp av, pari_sp tetpil, GEN q)
 }
 
 void
-fill_stack(void)
+dbg_fill_stack(void)
 {
+#ifdef LONG_IS_64BIT
+  const long JUNK = 0xBADC0FFEE0DDF00D;
+#else
+  const long JUNK = 0xDEADBEEF;
+#endif
   GEN x = ((GEN)pari_mainstack->bot);
-  while (x < (GEN)avma) *x++ = 0xfefefefeUL;
+  while (x < (GEN)avma) *x++ = JUNK;
 }
 
 void
