@@ -310,7 +310,7 @@ IsGoodSubgroup(GEN H, GEN bnr, GEN map)
     GEN pr = gel(P, j), e;
     /* if pr divides modH, it is ramified, so it's good */
     if (tablesearch(PH, pr, cmp_prime_ideal)) continue;
-    /* inertia degree of pr in bnr(modH)/H is charorder(e, cycH) */
+    /* inertia degree of pr in bnr(modH)/H is charorder(cycH, e) */
     e = ZM_ZC_mul(gel(qH,3), isprincipalray(bnrH, pr));
     e = ZV_ZV_mod(e, gel(qH,2));
     if (ZV_equal0(e)) return gc_long(av,0); /* f = 1 */
@@ -361,7 +361,7 @@ CplxModulus(GEN data, long *newprec)
   pari_sp av = avma;
   for (;;)
   {
-    GEN cpl, pol = AllStark(data, -1, dprec);
+    GEN cpl, pol = AllStark(data, 1, dprec);
     cpl = RgX_fpnorml2(pol, LOWDEFAULTPREC);
     dprec = maxss(dprec, nbits2extraprec(gexpo(pol))) + EXTRAPREC64;
     if (!gequal0(cpl)) { *newprec = dprec; return gexpo(cpl); }
@@ -777,11 +777,12 @@ _data3(GEN arch, long r2)
 static void
 ch_get3(GEN dtcr, long *a, long *b, long *c)
 { GEN v = ch_3(dtcr); *a = v[1]; *b = v[2]; *c = v[3]; }
+/* 2^(2*r2) pi^N */
 static GEN
-get_C(GEN nf, long prec)
+get_P(GEN nf, long prec)
 {
   long r2 = nf_get_r2(nf), N = nf_get_degree(nf);
-  return gmul2n(sqrtr_abs(divir(nf_get_disc(nf), powru(mppi(prec),N))), -r2);
+  GEN P = powru(mppi(prec), N); shiftr_inplace(P, 2*r2); return P;
 }
 /* sort chars according to conductor */
 static GEN
@@ -809,22 +810,21 @@ static GEN
 InitChar(GEN bnr, GEN ch, long all, long prec)
 {
   GEN bnf = checkbnf(bnr), nf = bnf_get_nf(bnf), mod = bnr_get_mod(bnr);
-  GEN C, dataCR, ncyc, vChar = sortChars(ch);
+  GEN P, dataCR, ncyc, dk = nf_get_disc(nf), vChar = sortChars(ch);
   long n, l, r2 = nf_get_r2(nf), prec2 = precdbl(prec) + EXTRAPREC64;
   long lv = lg(vChar);
 
-  C = get_C(nf, prec2);
+  P = get_P(nf, prec2);
   ncyc = cyc_normalize(bnr_get_cyc(bnr));
 
   dataCR = cgetg_copy(ch, &l);
   for (n = 1; n < lv; n++)
   {
-    GEN D, bnrc, v = gel(vChar, n); /* indices of chars of given conductor */
+    GEN D, bnrc, v = gel(vChar, n); /* indices of chars of given cond */
     long a, i = v[1], lc = lg(v);
     GEN F = gmael(ch,i,2);
-
     gel(dataCR, i) = D = cgetg(8, t_VEC);
-    ch_C(D) = mulrr(C, gsqrt(ZM_det_triangular(gel(F,1)), prec2));
+    ch_C(D) = sqrtr_abs(divir(mulii(dk, ZM_det_triangular(gel(F,1))), P));
     ch_3(D) = _data3(gel(F,2), r2);
     if (gequal(F, mod))
     {
@@ -859,23 +859,25 @@ InitChar(GEN bnr, GEN ch, long all, long prec)
   return mkvec2(vChar, dataCR);
 }
 
-/* recompute dataCR with the new precision, modify bnr components in place */
+/* recompute dataCR with the new precision, in place. Don't modify bnr
+ * components in place  */
 static void
 CharNewPrec(GEN data, long prec)
 {
   long j, l, prec2 = precdbl(prec) + EXTRAPREC64;
-  GEN C, nf, dataCR = gmael(data,4,2), D = gel(dataCR,1);
+  GEN P, nf, dataCR = gmael(data,4,2), D = gel(dataCR,1);
 
   if (ch_prec(D) >= prec2) return;
   nf = bnr_get_nf(ch_bnr(D));
   if (nf_get_prec(nf) < prec) nf = nfnewprec_shallow(nf, prec); /* not prec2 */
-  C = get_C(nf, prec2); l = lg(dataCR);
+  P = get_P(nf, prec2); l = lg(dataCR);
   for (j = 1; j < l; j++)
   {
-    GEN f0;
+    GEN f0, bnr;
     D = gel(dataCR,j); f0 = gel(bnr_get_mod(ch_bnr(D)), 1);
-    ch_C(D) = mulrr(C, gsqrt(ZM_det_triangular(f0), prec2));
-    gmael(ch_bnr(D), 1, 7) = nf;
+    ch_C(D) = sqrtr_abs(divir(mulii(nf_get_disc(nf),ZM_det_triangular(f0)), P));
+    ch_bnr(D) = bnr = shallowcopy(ch_bnr(D));
+    gel(bnr,1) = shallowcopy(gel(bnr,1)); gmael(bnr,1,7) = nf;
     ch_CHI(D) = get_Char(gel(ch_CHI(D),1), prec2);
     ch_CHI0(D)= get_Char(gel(ch_CHI0(D),1), prec2);
   }
@@ -1048,9 +1050,8 @@ static GEN
 EvalCoeff(GEN z, int* c, long deg)
 {
   long i,j;
-  GEN e, r;
+  GEN r, e = NULL;
 
-  if (!c) return gen_0;
 #if 0
   /* standard Horner */
   e = stoi(c[deg - 1]);
@@ -1058,7 +1059,6 @@ EvalCoeff(GEN z, int* c, long deg)
     e = gadd(stoi(c[i]), gmul(z, e));
 #else
   /* specific attention to sparse polynomials */
-  e = NULL;
   for (i = deg-1; i >=0; i=j-1)
   {
     for (j=i; c[j] == 0; j--)
@@ -1296,7 +1296,7 @@ InitPrimes(GEN bnr, ulong N0, LISTray *R)
 static GEN /* cf polcoef */
 _sercoeff(GEN x, long n)
 {
-  long i = n - valp(x);
+  long i = n - valser(x);
   return (i < 0)? gen_0: gel(x,i+2);
 }
 
@@ -1341,11 +1341,11 @@ ppgamma(ST_t *T, long prec)
   sqpi = sqrtr_abs(mppi(prec)); /* Gamma(1/2) */
 
   G1 = gexp(integser(psi1series(r-1, 0, prec)), prec); /* Gamma(1 + x) */
-  G = shallowcopy(G1); setvalp(G,-1); /* Gamma(x) */
+  G = shallowcopy(G1); setvalser(G,-1); /* Gamma(x) */
 
   /* expansion of log(Gamma(u) / Gamma(1/2)) at u = 1/2 */
   G2 = cgetg(r+2, t_SER);
-  G2[1] = evalsigne(1) | _evalvalp(1) | evalvarn(0);
+  G2[1] = evalsigne(1) | _evalvalser(1) | evalvarn(0);
   gel(G2,2) = gneg(gadd(gmul2n(mplog2(prec), 1), mpeuler(prec)));
   for (i = 1; i < r; i++) gel(G2,i+2) = mulri(gel(G1,i+2), int2um1(i));
   G2 = gmul(sqpi, gexp(G2, prec)); /* Gamma(1/2 + x) */
@@ -1437,9 +1437,9 @@ GetValue1(GEN bnr, long flag, long prec)
   if (flag)
   {
     GEN diff = divcond(bnr);
-    long i, l;
-    l = lg(diff) - 1; r += l;
-    for (i = 1; i <= l; i++) c = gmul(c, glog(pr_norm(gel(diff,i)), prec));
+    long i, l = lg(diff);
+    r += l - 1;
+    for (i = 1; i < l; i++) c = gmul(c, glog(pr_norm(gel(diff,i)), prec));
   }
   return mkvec2(utoi(r), c);
 }
@@ -1472,7 +1472,7 @@ chk_reccoeff(void *data, GEN x)
   GEN v = gmul(d->U, x), z = gel(v,1);
 
   if (!gequal1(z)) return NULL;
-  *++v = evaltyp(t_COL) | evallg( lg(d->M) );
+  *++v = evaltyp(t_COL) | _evallg( lg(d->M) ); /* pop 1st elt */
   if (TestOne(gmul(d->M, v), d)) return v;
   return NULL;
 }
@@ -1488,7 +1488,7 @@ RecCoeff3(GEN nf, RC_data *d, long prec)
   FP_chk_fun chk = { &chk_reccoeff, &chk_reccoeff_init, NULL, NULL, 0 };
   chk.data = (void*)d;
 
-  d->G = minss(-10, -prec2nbits(prec) >> 4);
+  d->G = minss(-10, -prec >> 4);
   BIG = maxss(32, -2*d->G);
   tB  = sqrtnr(real2n(BIG-N,DEFAULTPREC), N-1);
   Bd  = grndtoi(gmin_shallow(B, tB), &e);
@@ -1553,18 +1553,18 @@ RecCoeff2(GEN nf,  RC_data *d,  long prec)
   GEN vec, M = nf_get_M(nf), beta = d->beta;
   long bit, min, max, lM = lg(M);
 
-  d->G = minss(-20, -prec2nbits(prec) >> 4);
+  d->G = minss(-20, -prec >> 4);
 
   vec  = vec_prepend(row(M, d->v), gneg(beta));
-  min = (long)prec2nbits_mul(prec, 0.75);
-  max = (long)prec2nbits_mul(prec, 0.98);
+  min = (long)prec * 0.75;
+  max = (long)prec * 0.98;
   av = avma;
   for (bit = max; bit >= min; bit-=32, set_avma(av))
   {
     long e;
     GEN v = lindep_bit(vec, bit), z = gel(v,1);
     if (!signe(z)) continue;
-    *++v = evaltyp(t_COL) | evallg(lM);
+    *++v = evaltyp(t_COL) | _evallg(lM); /* pop 1st elt */
     v = grndtoi(gdiv(v, z), &e);
     if (e > 0) break;
     if (TestOne(RgM_RgC_mul(M, v), d)) return v;
@@ -1575,9 +1575,11 @@ RecCoeff2(GEN nf,  RC_data *d,  long prec)
 
 /* Attempts to find a polynomial with coefficients in nf such that
    its coefficients are close to those of pol at the place v and
-   less than B at all the other places */
+   less than a certain bound at all the other places. This bound is obtained by assuming
+   that the roots at the other places are, in absolute values, less than 2 if flag == 0,
+   and less than 1 if flag = 1. */
 static GEN
-RecCoeff(GEN nf,  GEN pol,  long v, long prec)
+RecCoeff(GEN nf,  GEN pol,  long v, long flag, long prec)
 {
   long j, md, cl = degpol(pol);
   pari_sp av = avma;
@@ -1587,7 +1589,7 @@ RecCoeff(GEN nf,  GEN pol,  long v, long prec)
   for (j = 2; j <= cl+1; j++)
   {
     GEN t = gel(pol, j);
-    if (prec2nbits(precision(t)) - gexpo(t) < 34) return NULL;
+    if (precision(t) - gexpo(t) < 34) return NULL;
   }
 
   md = cl/2;
@@ -1600,8 +1602,9 @@ RecCoeff(GEN nf,  GEN pol,  long v, long prec)
   { /* start with the coefficients in the middle,
        since they are the harder to recognize! */
     long cf = md + (j%2? j/2: -j/2);
-    GEN t, bound = shifti(binomial(utoipos(cl), cf), cl-cf);
+    GEN t, bound = binomial(utoipos(cl), cf);
 
+    if (flag == 0) bound = shifti(bound, cl-cf);
     if (DEBUGLEVEL>1) err_printf("RecCoeff (cf = %ld, B = %Ps)\n", cf, bound);
     d.beta = real_i( gel(pol,cf+2) );
     d.B    = bound;
@@ -1793,12 +1796,12 @@ QuadGetST(GEN bnr, GEN *pS, GEN *pT, GEN CR, long prec)
         return;
       }
       /* FIXME: is this value of N0 correct for the general case ? */
-      N0[j] = (long)prec2nbits_mul(prec, 0.35 * gtodouble(c));
+      N0[j] = (long)prec * 0.35 * gtodouble(c);
     }
     else /* complex quadratic */
     {
       cs[j] = 1;
-      N0[j] = (long)prec2nbits_mul(prec, 0.7 * gtodouble(c));
+      N0[j] = (long)prec * 0.7 * gtodouble(c);
     }
     if (n0 < N0[j]) n0 = N0[j];
   }
@@ -2055,7 +2058,7 @@ GetST0(GEN bnr, GEN *pS, GEN *pT, GEN CR, long prec)
   C  = cgetg(ncond+1, t_VEC);
   N0 = cgetg(ncond+1, t_VECSMALL);
   n0 = 0;
-  limx = zeta_get_limx(r1, r2, prec2nbits(prec));
+  limx = zeta_get_limx(r1, r2, prec);
   for (j = 1; j <= ncond; j++)
   {
     GEN dtcr = gel(dataCR, mael(vChar,j,1)), c = ch_C(dtcr);
@@ -2063,7 +2066,7 @@ GetST0(GEN bnr, GEN *pS, GEN *pT, GEN CR, long prec)
     N0[j] = zeta_get_N0(c, limx);
     if (n0 < N0[j]) n0  = N0[j];
   }
-  cScT.i0 = i0 = zeta_get_i0(r1, r2, prec2nbits(prec), limx);
+  cScT.i0 = i0 = zeta_get_i0(r1, r2, prec, limx);
   if (DEBUGLEVEL>1) err_printf("i0 = %ld, N0 = %ld\n",i0, n0);
   InitPrimes(bnr, n0, &LIST);
   prec2 = precdbl(prec) + EXTRAPREC64;
@@ -2136,12 +2139,9 @@ GetST(GEN bnr, GEN *pS, GEN *pT, GEN CR, long prec)
 static GEN
 GenusFieldQuadReal(GEN disc)
 {
-  long i, i0 = 0, l;
-  pari_sp av = avma;
-  GEN T = NULL, p0 = NULL, P;
+  GEN T = NULL, p0 = NULL, P = gel(Z_factor(disc), 1);
+  long i, i0 = 0, l = lg(P);
 
-  P = gel(Z_factor(disc), 1);
-  l = lg(P);
   for (i = 1; i < l; i++)
   {
     GEN p = gel(P,i);
@@ -2163,21 +2163,16 @@ GenusFieldQuadReal(GEN disc)
     else
       d = (mod4(p) == 1)? p: mulii(p0, p);
     t = mkpoln(3, gen_1, gen_0, negi(d)); /* x^2 - d */
-    T = T? ZX_compositum_disjoint(T, t): t;
+    T = T? ZX_composedsum(T, t): t;
   }
-  return gerepileupto(av, polredbest(T, 0));
+  return polredbest(T, 0);
 }
 static GEN
 GenusFieldQuadImag(GEN disc)
 {
-  long i, l;
-  pari_sp av = avma;
-  GEN T = NULL, P;
-
-  P = gel(absZ_factor(disc), 1);
-  l = lg(P);
-  l--; /* remove last prime */
-  for (i = 1; i < l; i++)
+  GEN T = NULL, P = gel(absZ_factor(disc), 1);
+  long i, n = lg(P)-1;
+  for (i = 1; i < n; i++) /* remove last prime */
   {
     GEN p = gel(P,i), d, t;
     if (absequaliu(p, 2))
@@ -2190,12 +2185,13 @@ GenusFieldQuadImag(GEN disc)
     else
       d = (mod4(p) == 1)? p: negi(p);
     t = mkpoln(3, gen_1, gen_0, negi(d)); /* x^2 - d */
-    T = T? ZX_compositum_disjoint(T, t): t;
+    T = T? ZX_composedsum(T, t): t;
   }
-  return gerepileupto(av, polredbest(T, 0));
+  return polredbest(T, 0);
 }
 
-/* if flag != 0, computes a fast and crude approximation of the result */
+/* if flag = 1, computes a fast and crude approximation of the trace of the Stark unit
+   if flag = 2, computes the Stark unit */
 static GEN
 AllStark(GEN data, long flag, long newprec)
 {
@@ -2224,23 +2220,23 @@ LABDOUB:
   W = AllArtinNumbers(CR, newprec);
   if (DEBUGLEVEL) timer_printf(&ti,"Compute W");
   Lp = cgetg(cl + 1, t_VEC);
-  if (!flag)
+  if (flag != 1)
   {
     GetST(bnr, &S, &T, CR, newprec);
     if (DEBUGLEVEL) timer_printf(&ti, "S&T");
     for (i = 1; i <= cl; i++)
     {
-      GEN chi = gel(dataCR, i), v = gen_0;
+      GEN chi = gel(dataCR, i), vv = gen_0;
       if (ch_comp(chi))
-        v = gel(GetValue(chi, gel(W,i), gel(S,i), gel(T,i), 2, newprec), 2);
-      gel(Lp, i) = v;
+        vv = gel(GetValue(chi, gel(W,i), gel(S,i), gel(T,i), 2, newprec), 2);
+      gel(Lp, i) = vv;
     }
   }
   else
   { /* compute a crude approximation of the result */
     C = cgetg(cl + 1, t_VEC);
     for (i = 1; i <= cl; i++) gel(C,i) = ch_C(gel(dataCR, i));
-    n = zeta_get_N0(vecmax(C), zeta_get_limx(r1, r2, prec2nbits(newprec)));
+    n = zeta_get_N0(vecmax(C), zeta_get_limx(r1, r2, newprec));
     if (n > BND) n = BND;
     if (DEBUGLEVEL) err_printf("N0 in QuickPol: %ld \n", n);
     InitPrimes(bnr, n, &LIST);
@@ -2272,8 +2268,11 @@ LABDOUB:
   }
 
   p1 = gel(data,3);
-  den = flag ? h: 2*h;
-  vzeta = cgetg(h + 1, t_VEC);
+  den = (flag == 0) ? 2*h: h;
+  if (flag == 2)
+    vzeta = cgetg(2*h + 1, t_VEC);
+  else
+    vzeta = cgetg(h + 1,t_VEC);
   for (i = 1; i <= h; i++)
   {
     GEN z = gen_0, sig = gel(p1,i);
@@ -2284,7 +2283,14 @@ LABDOUB:
       if (chi_get_deg(CHI) != 2) t = gmul2n(t, 1); /* character not real */
       z = gadd(z, t);
     }
-    gel(vzeta,i) = gmul2n(gcosh(gdivgu(z,den), newprec), 1);
+    z = gdivgu(z, den);
+    if (flag == 2)
+    {
+      gel(vzeta, i) = gexp(gmul2n(z,1), newprec);
+      gel(vzeta, i+h) = ginv(gel(vzeta,i));
+    }
+    /* if flag == 0, we first try with the square-root of the Stark unit */
+    else gel(vzeta,i) = gmul2n(gcosh(z, newprec), 1);
   }
   polrelnum = roots_to_pol(vzeta, 0);
   if (DEBUGLEVEL)
@@ -2292,20 +2298,18 @@ LABDOUB:
     if (DEBUGLEVEL>1) {
       err_printf("polrelnum = %Ps\n", polrelnum);
       err_printf("zetavalues = %Ps\n", vzeta);
-      if (!flag)
+      if (flag == 0)
         err_printf("Checking the square-root of the Stark unit...\n");
     }
     timer_printf(&ti, "Compute %s", flag? "quickpol": "polrelnum");
   }
-  if (flag) return gerepilecopy(av, polrelnum);
-
+  if (flag == 1) return gerepilecopy(av, polrelnum);
   /* try to recognize this polynomial */
-  polrel = RecCoeff(nf, polrelnum, v, newprec);
-  if (!polrel)
+  polrel = RecCoeff(nf, polrelnum, v, flag, newprec);
+
+  /* if this doesn't work, maybe the Stark unit is not a square */
+  if (!polrel && flag == 0)
   {
-    for (j = 1; j <= h; j++)
-      gel(vzeta,j) = gsubgs(gsqr(gel(vzeta,j)), 2);
-    polrelnum = roots_to_pol(vzeta, 0);
     if (DEBUGLEVEL)
     {
       if (DEBUGLEVEL>1) {
@@ -2314,7 +2318,11 @@ LABDOUB:
       }
       timer_printf(&ti, "Compute polrelnum");
     }
-    polrel = RecCoeff(nf, polrelnum, v, newprec);
+    for (j = 1; j <= h; j++)
+      gel(vzeta,j) = gsubgs(gsqr(gel(vzeta,j)), 2);
+    polrelnum = roots_to_pol(vzeta, 0);
+
+    polrel = RecCoeff(nf, polrelnum, v, 0, newprec);
   }
   if (!polrel) /* FAILED */
   {
@@ -2322,7 +2330,7 @@ LABDOUB:
     long incr_pr;
     if (++cpt >= 3) pari_err_PREC( "stark (computation impossible)");
     /* estimate needed precision */
-    incr_pr = prec2nbits(gprecision(polrelnum))- gexpo(polrelnum);
+    incr_pr = gprecision(polrelnum) - gexpo(polrelnum);
     if (incr_pr < 0) incr_pr = -incr_pr + EXTRA_BITS;
     newprec += nbits2extraprec(maxss(3*EXTRA_BITS, cpt*incr_pr));
     if (DEBUGLEVEL) pari_warn(warnprec, "AllStark", newprec);
@@ -2389,6 +2397,37 @@ bnrstark(GEN bnr, GEN subgrp, long prec)
   return gerepileupto(av, AllStark(data, 0, newprec));
 }
 
+GEN
+bnrstarkunit(GEN bnr, GEN subgrp)
+{
+  long newprec, deg;
+  pari_sp av = avma;
+  GEN nf, data, dtQ, bnf, bnrf, Cm, candD, D, QD, CR;
+
+  /* check the input */
+  checkbnr(bnr); bnf = bnr_get_bnf(bnr); nf = bnf_get_nf(bnf);
+  if (!nf_get_varn(nf))
+    pari_err_PRIORITY("bnrstarkunit", nf_get_pol(nf), "=", 0);
+  deg = nf_get_degree(nf);
+  if (deg == 1) pari_err_IMPL("bnrstarkunit for basefield Q");
+  if (nf_get_r2(nf)) pari_err_DOMAIN("bnrstarkunit", "r2", "!=", gen_0, nf);
+  bnr_subgroup_sanitize(&bnr, &subgrp);
+  if (lg(bid_get_archp(bnr_get_bid(bnr)))-1 != deg-1)
+    pari_err_DOMAIN("bnrstarkunit", "# unramified places", "!=", gen_1, bnr);
+  bnrf = Buchray(bnf, gel(bnr_get_mod(bnr), 1), nf_INIT);
+  subgrp = abmap_subgroup_image(bnrsurjection(bnr, bnrf), subgrp);
+  dtQ  = InitQuotient(subgrp);
+  Cm   = ComputeKernel(bnr, bnrf, dtQ);
+  candD = subgrouplist_cond_sub(bnr, Cm, mkvec(gen_2));
+  if (lg(candD) != 2) pari_err(e_MISC, "incorrect modulus in bnrstark");
+  D    = gel(candD, 1);
+  QD   = InitQuotient(D);
+  CR   = InitChar(bnr, AllChars(bnr, QD, 1), 0, DEFAULTPREC);
+  data = mkvec4(bnr, D, subgroup_classes(Cm), CR);
+  CplxModulus(data, &newprec);
+  return gerepileupto(av, AllStark(data, 2, newprec));
+}
+
 /* For each character of Cl(bnr)/subgp, compute L(1, chi) (or equivalently
  * the first nonzero term c(chi) of the expansion at s = 0).
  * If flag & 1: compute the value at s = 1 (for nontrivial characters),
@@ -2828,7 +2867,7 @@ quadhilbertimag(GEN D)
     }
   L = qfbforms(D);
   h = lg(L)-1;
-  if ((1L << vals(h)) == h) /* power of 2 */
+  if (! (h & (h - 1))) /* power of 2 */
   { /* check whether > |Cl|/2 elements have order <= 2 ==> 2-elementary */
     long lim = (h>>1) + 1;
     for (i=1; i <= lim; i++)
@@ -2879,7 +2918,7 @@ quadhilbertimag(GEN D)
     P = grndtoi(P,&exmax);
     if (DEBUGLEVEL>1) timer_printf(&ti,"product, error bits = %ld",exmax);
     if (exmax <= -10) break;
-    set_avma(av0); prec += nbits2extraprec(prec2nbits(DEFAULTPREC)+exmax);
+    set_avma(av0); prec += nbits2extraprec(DEFAULTPREC + exmax);
     if (DEBUGLEVEL) pari_warn(warnprec,"quadhilbertimag",prec);
   }
   return P;
@@ -3083,7 +3122,7 @@ computeth2(GEN om, GEN la, long prec)
 
   p1 = gsub(ellphist(om,res,la,prec), ellphist(om,res,gen_1,prec));
   p2 = imag_i(p1);
-  if (gexpo(real_i(p1))>20 || gexpo(p2)> prec2nbits(minss(prec,realprec(p2)))-10)
+  if (gexpo(real_i(p1)) > 20 || gexpo(p2) > minss(prec,realprec(p2)) - 10)
     return NULL;
   return gexp(p1,prec);
 }
